/******************************************************************************
 * $Header: /CommonBe/agmsmith/Programming/ScriptOCron/RCS/CronBitset.cpp,v 1.15 2015/12/26 18:27:01 agmsmith Exp $
 *
 * This utility class converts from Cron style strings to a set of dates and
 * times, which can be used more easily to test if a particular time is in the
 * specified set.  It can also convert the set back into a string.
 *
 * This CronBitset software is licensed under the MIT license.
 *
 * Copyright (c) 2015 by Alexander G. M. Smith.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#include "CronBitset.h"

/* Standard C Library. */

#include <stdio.h>


const char * g_DayNames [] =
{
  "Sunday",
  "Monday",
  "Tuesday",
  "Wednesday",
  "Thursday",
  "Friday",
  "Saturday"
};

const char * g_MonthNames [] =
{
  "January",
  "February",
  "March",
  "April",
  "May",
  "June",
  "July",
  "August",
  "September",
  "October",
  "November",
  "December"
};

const char * g_CronElementNames [CronBitset::CET_NUMBER_OF_ELEMENTS] =
{
  "Day of Week",
  "Month",
  "Day of Month",
  "Hour",
  "Minute"
};



/******************************************************************************
 * The CronElementSet support class.
 */

CronElementSet::CronElementSet (
 int LowestValue, int HighestValue, const char **ValueNames, bool LeadingZero)
{
  m_Bits.reset ();
  m_LowestValue = LowestValue;
  m_Size = HighestValue - LowestValue + 1;
  if (m_Size < 0)
    m_Size = 0;
  m_ValueNames = ValueNames;
  m_LeadingZeroes = LeadingZero;
}


bool CronElementSet::GetAllSet ()
{
  for (int iValue = 0; iValue < m_Size; iValue++)
  {
    if (!m_Bits [iValue])
      return false;
  }
  return true;
}


void CronElementSet::SetAll (bool NewState)
{
  if (!NewState)
    m_Bits.reset ();
  else
    for (int iValue = 0; iValue < m_Size; iValue++)
      m_Bits [iValue] = true;
}


const char * CronElementSet::FromString (const char *Buffer)
{
  m_Bits.reset (); // Start out empty, string specifies only set bits.

  // Skip leading space.

  const char *pSource = Buffer;
  while (*pSource != 0 && isspace (*pSource))
    pSource++;

  // Look for the next word, which could be the name of a value, a numerical
  // value, * for all values, or separators "-,".  Check easy ones first.

  int iPreviousValue = -1; // Previously read in value, zero based.
  bool RangeMarkerEncountered = false; // Did we see a "-" between values?
  while (*pSource != 0 && !isspace (*pSource))
  {
    if (*pSource == '?')
    { // Question mark is used for an empty set.  Just skip over it.
      RangeMarkerEncountered = false;
      pSource++;
      continue;
    }

    if (*pSource == '-')
    { // Hyphen marks a range specified by previous and following values.
      RangeMarkerEncountered = true;
      pSource++;
      continue;
    }

    if (*pSource == ',')
    { // Comma separates individual values.
      RangeMarkerEncountered = false;
      pSource++;
      continue;
    }

    if (*pSource == '*')
    { // Asterisk means all values are set.
      for (int iValue = 0; iValue < m_Size; iValue++)
        m_Bits [iValue] = true;
      RangeMarkerEncountered = false;
      pSource++;
      continue;
    }

    int iDecodedValue = -1; // Decoded value of the current word, zero based.

    // See if it matches one of the names first, before trying numbers.

    if (m_ValueNames != NULL)
    {
      const char *pTest = NULL;
      int iName;
      for (iName = 0; iName < m_Size; iName++)
      {
        const char *pName = m_ValueNames [iName];
        if (pName == NULL)
          continue;

        // Count how many characters of the name the current word matches, case
        // insensitive.

        int MatchCount = 0;
        pTest = pSource;
        while (true)
        {
          char SourceLetter = *pTest;
          char NameLetter = *pName;
          if (NameLetter == 0 || SourceLetter == 0)
            break;
          if (toupper (SourceLetter) != toupper (NameLetter))
            break;
          MatchCount++;
          pTest++;
          pName++;
        }
        if (MatchCount >= 3)
          break; // Just need first 3 letters to match, more are optional.
      }

      if (iName < m_Size)
      { // Found a name.
        pSource = pTest;
        iDecodedValue = iName;
      }
    }

    // If not a name, see if the word can be converted to a number.

    if (iDecodedValue < 0)
    {
      char *pEnd = NULL;
      int UserValue = (int) strtol (pSource, &pEnd, 10 /* base 10 number */);
      if (pEnd > pSource)
      {
        // Decoded a few characters, hope we got a useable number.

        iDecodedValue = UserValue - m_LowestValue;

        if (iDecodedValue < 0 || iDecodedValue >= m_Size)
          iDecodedValue = -1; // Was out of range, ignore it.
        else
          pSource = pEnd;
      }
    }

    if (iDecodedValue < 0)
      return pSource; // Wasn't able to read a value, return bad word location.

    m_Bits[iDecodedValue] = true;

    if (RangeMarkerEncountered)
    {
      // Have a range, fill in the whole series.

      if (iPreviousValue >= 0)
      {
        for (int iMiddle = iPreviousValue; iMiddle < iDecodedValue; iMiddle++)
        {
          m_Bits[iMiddle] = true;
        }
      }
      RangeMarkerEncountered = false;
    }

    iPreviousValue = iDecodedValue;
  }
  return pSource;
}


bool CronElementSet::ToString (char *Buffer, int BufferSize) const
{
  int SizeRemaining = 0;

  char *pDest = Buffer;
  if (BufferSize < 1)
    return false;
  *pDest = 0;

  for (int iValue = 0; iValue < m_Size; )
  {
    if (!m_Bits[iValue])
    {
      iValue++; // Skip over unset values.
      continue;
    }

    SizeRemaining = BufferSize - (pDest - Buffer);

    // Look for a sequence of values.  Ends up with iEndSeq being one past the
    // last value in the sequence.

    int iEndSeq = iValue + 1;
    while (iEndSeq < m_Size && m_Bits[iEndSeq])
      iEndSeq++;
    int SequenceCount = iEndSeq - iValue;

    if (SequenceCount >= m_Size) // All of them are set, use "*".
    {
      if (SizeRemaining < 1)
        return false;
      *pDest++ = '*';
      iValue = iEndSeq;
    }
    else if (SequenceCount >= 3) // 3 or more, use a range "A-B".
    {
      if (!ValueToString (pDest, SizeRemaining, iValue))
        return false;
      pDest += strlen (pDest);
      SizeRemaining = BufferSize - (pDest - Buffer);
      if (SizeRemaining < 1)
        return false;
      *pDest++ = '-';
      SizeRemaining--;
      if (!ValueToString (pDest, SizeRemaining, iEndSeq - 1))
        return false;
      pDest += strlen (pDest);
      iValue = iEndSeq;
    }
    else // Just print one element.
    {
      if (!ValueToString (pDest, SizeRemaining, iValue))
        return false;
      pDest += strlen (pDest);
      iValue++;
    }

    // Add a comma between values.  Don't know if there is a next one, so
    // always add it now.

    SizeRemaining = BufferSize - (pDest - Buffer);
    if (SizeRemaining < 1)
      return false;
    *pDest++ = ',';
  }

  // Remove excess trailing comma.

  if (pDest > Buffer && pDest[-1] == ',')
    pDest--;

  // If nothing was output, write a single question mark.  Otherwise an empty
  // set would get lost in the leading white space.  No size test needed,
  // alread know buffer is at least 1 byte long.

  if (pDest == Buffer)
    *pDest++ = '?';

  SizeRemaining = BufferSize - (pDest - Buffer);
  if (SizeRemaining < 1)
    return false;
  *pDest = 0;
  return true;
}


bool CronElementSet::ValueToString (char *Buffer, int BufferSize,
  int iValue) const
{
  // Is it a named value?

  if (m_ValueNames != NULL && m_ValueNames[iValue] != NULL)
  {
    if ((int) strlen (m_ValueNames[iValue]) + 1 < BufferSize)
    {
      strcpy (Buffer, m_ValueNames[iValue]);
      return true;
    }
    else
      return false;
  }

  // Just use the number then.  Maybe with a leading zero.

  int AmountWritten = snprintf (Buffer, BufferSize,
    m_LeadingZeroes ? "%02d" : "%d",
    iValue + m_LowestValue);
  if (AmountWritten >= BufferSize)
    return false;
  return true;
}



/******************************************************************************
 * The main CronBitset class.
 */

CronBitset::CronBitset ()
: m_Minute (0, 59, NULL, true /* want leading zeroes */),
  m_Hour (0, 23, NULL, true),
  m_DayOfMonth (1, 31),
  m_Month (1, 12, g_MonthNames),
  m_DayOfWeek (0, 6, g_DayNames)
{
  a_Elements[CET_MINUTE] = &m_Minute;
  a_Elements[CET_HOUR] = &m_Hour;
  a_Elements[CET_DAY_OF_MONTH] = &m_DayOfMonth;
  a_Elements[CET_MONTH] = &m_Month;
  a_Elements[CET_DAY_OF_WEEK] = &m_DayOfWeek;
}


void CronBitset::ClearAll ()
{
  for (int iElement = 0; iElement < CET_NUMBER_OF_ELEMENTS; iElement++)
    a_Elements[iElement]->SetAll (false);
}


const char * CronBitset::FromString (const char *Buffer)
{
  if (Buffer == NULL)
    return false;
  const char *pSource = Buffer;

  for (int iElement = 0; iElement < CET_NUMBER_OF_ELEMENTS; iElement++)
  {
    pSource = a_Elements[iElement]->FromString (pSource);
    if (*pSource == 0 && iElement < CET_NUMBER_OF_ELEMENTS - 1)
      return pSource; // Ran out of data before the end of string.
    if (*pSource != 0 && !isspace (*pSource))
      return pSource; // Hit a bad word.
  }
  return NULL; // Success.
}


bool CronBitset::GetAllSet (CronElementTypeEnum Element)
{
  if (Element < 0 || Element >= CET_NUMBER_OF_ELEMENTS)
    return false;

  return a_Elements[Element]->GetAllSet ();
}


bool CronBitset::GetValue (CronElementTypeEnum Element, int Value)
{
  if (Element < 0 || Element >= CET_NUMBER_OF_ELEMENTS)
    return false;

  return a_Elements[Element]->GetValue (Value);
}


void CronBitset::SetAll (CronElementTypeEnum Element, bool NewState)
{
  if (Element < 0 || Element >= CET_NUMBER_OF_ELEMENTS)
    return;

  a_Elements[Element]->SetAll (NewState);
}


void CronBitset::SetValue (CronElementTypeEnum Element, int Value,
  bool NewState)
{
  if (Element < 0 || Element >= CET_NUMBER_OF_ELEMENTS)
    return;

  a_Elements[Element]->SetValue (Value, NewState);
}


bool CronBitset::TimeMatches (const struct tm *pTime)
{
  if (!m_Minute.GetValue (pTime->tm_min))
    return false;
  if (!m_Hour.GetValue (pTime->tm_hour))
    return false;
  if (!m_DayOfMonth.GetValue (pTime->tm_mday))
    return false;
  if (!m_Month.GetValue (pTime->tm_mon + 1))
    return false;
  if (!m_DayOfWeek.GetValue (pTime->tm_wday))
    return false;
  return true;
}


bool CronBitset::ToString (char *Buffer, int BufferSize) const
{
  char *pDest = Buffer;
  if (BufferSize < 1)
    return false;
  *pDest = 0;

  for (int iElement = 0; iElement < CET_NUMBER_OF_ELEMENTS; iElement++)
  {
    int SizeRemaining = BufferSize - (pDest - Buffer);
    if (!a_Elements[iElement]->ToString (pDest, SizeRemaining))
      return false;
    pDest = pDest + strlen (pDest);

    // Use a space to separate the elements.

    if (iElement < CET_NUMBER_OF_ELEMENTS - 1)
    {
      SizeRemaining = BufferSize - (pDest - Buffer);
      if (SizeRemaining < 2)
        return false;
      *pDest++ = ' ';
      *pDest = 0;
    }
  }
  return true;
}



/******************************************************************************
 * Self test code.
 */

#if TEST_CRON_BITSET
void TestCronBitset ()
{
  char Buffer [1024];
  static const char * TestStrings [] =
  {
    "2 4 5\t6 5,14-17 With a tab and trailing stuff",
    "4 2 32 23 59",
    "1 2 3 4",
    "? ? ? ? ? The Empty Set",
    ", , , , ,",
    "- - - - -",
    "-?, -? -?* ,,, ??",
    "? 2?-6 12?25 ? 6-?12 Almost empty",
    "1 2 3,----,,-1 4 5",
    "   5 4      3   \t   2   \n   1 Leading spaces.",
    "0 * 1-12 * *",
    "W 1 1 1 1",
    "We 1 1 1 1",
    "Wed 1 1 1 1",
    "Wedn 1 1 1 1",
    "Wednesday 1 1 1 1",
    "Wednesdayish 1 1 1 1",
    "Sun-Sat Jan-Dec, 1-31, 0-12,23-2 0-59"
  };

  printf ("Testing $Header: /CommonBe/agmsmith/Programming/ScriptOCron/RCS/CronBitset.cpp,v 1.15 2015/12/26 18:27:01 agmsmith Exp $\n");
  CronBitset TestSet;
  TestSet.ToString (Buffer, sizeof (Buffer));
  printf ("Empty CronBitset: \"%s\"\n", Buffer);

  TestSet.SetValue (CronBitset::CET_MINUTE, 15);
  TestSet.ToString (Buffer, sizeof (Buffer));
  printf ("Just 15 minute: \"%s\"\n", Buffer);

  TestSet.SetValue (CronBitset::CET_HOUR, 14);
  TestSet.SetValue (CronBitset::CET_DAY_OF_MONTH, 20);
  TestSet.SetValue (CronBitset::CET_MONTH, 8);
  TestSet.SetValue (CronBitset::CET_DAY_OF_WEEK, 2);
  TestSet.ToString (Buffer, sizeof (Buffer));
  printf ("One particular time Tuesday, August 20, 14:15: \"%s\"\n", Buffer);

  TestSet.SetValue (CronBitset::CET_MONTH, 9);
  TestSet.ToString (Buffer, sizeof (Buffer));
  printf ("Two months (add Sept): \"%s\"\n", Buffer);

  TestSet.SetValue (CronBitset::CET_MONTH, 10);
  TestSet.ToString (Buffer, sizeof (Buffer));
  printf ("Three months (add Oct): \"%s\"\n", Buffer);

  TestSet.SetValue (CronBitset::CET_MONTH, 1);
  TestSet.ToString (Buffer, sizeof (Buffer));
  printf ("Three months and January: \"%s\"\n", Buffer);

  TestSet.SetValue (CronBitset::CET_MONTH, 2);
  TestSet.ToString (Buffer, sizeof (Buffer));
  printf ("Three months and J&F: \"%s\"\n", Buffer);

  TestSet.SetValue (CronBitset::CET_MONTH, 12);
  TestSet.ToString (Buffer, sizeof (Buffer));
  printf ("Last month too: \"%s\"\n", Buffer);

  for (int i = 0; i < 60; i++)
    TestSet.SetValue (CronBitset::CET_MINUTE, i);
  bool Result = TestSet.ToString (Buffer, sizeof (Buffer));
  printf ("All minutes, Three months and J&F with boolean result: \"%s\", %d\n",
    Buffer, Result);

  Result = TestSet.ToString (Buffer, 10);
  printf ("Buffer size of 10: \"%s\", %d\n", Buffer, Result);

  for (int i = 0; i < CronBitset::CET_NUMBER_OF_ELEMENTS; i++)
    for (int j = 0; j < 70; j++)
      TestSet.SetValue ((CronBitset::CronElementTypeEnum) i, j, rand () & 1);
  TestSet.ToString (Buffer, sizeof (Buffer));
  printf ("Random: \"%s\"\n", Buffer);

  for (int i = 0; i < CronBitset::CET_NUMBER_OF_ELEMENTS; i++)
    for (int j = 0; j < 70; j++)
      TestSet.SetValue ((CronBitset::CronElementTypeEnum) i, j, (j % 3) != 2);
  TestSet.ToString (Buffer, sizeof (Buffer));
  printf ("Worst case: \"%s\" string length was %d.\n", Buffer, (int) strlen (Buffer));

  for (int i = 0; i < (int) (sizeof (TestStrings) / sizeof (TestStrings[0]));
  i++)
  {
    TestSet.ClearAll ();
    const char * FailedWord = TestSet.FromString (TestStrings[i]);
    TestSet.ToString (Buffer, sizeof (Buffer));
    if (FailedWord == NULL)
      printf ("\"%s\" successfully converted to \"%s\"\n",
        TestStrings[i], Buffer);
    else
      printf ("\"%s\" failed conversion starting at \"%s\"\n",
        TestStrings[i], FailedWord);
  }
}
#endif


/******************************************************************************
 * $Log: CronBitset.cpp,v $
 * Revision 1.15  2015/12/26 18:27:01  agmsmith
 * Force base 10 numbers, leading zero was triggering octal recognition.
 *
 * Revision 1.14  2015/11/24 20:11:48  agmsmith
 * Added MIT licensing.
 *
 * Revision 1.13  2015/10/29 20:56:56  agmsmith
 * Added leading zero for hours and minutes less than 10.
 *
 * Revision 1.12  2015/10/28 19:08:20  agmsmith
 * Added functions to test if all bits set and to set/clear all bits, also
 * worst case conversion to string seems to take more bytes.
 *
 * Revision 1.11  2015/10/28 18:39:36  agmsmith
 * Print empty sets as "?" and parse them when reading too.
 *
 * Revision 1.10  2015/10/28 18:13:40  agmsmith
 * Narrowed down worst case string length when converting to string.
 *
 * Revision 1.9  2015/10/25 21:41:13  agmsmith
 * Reverse order of time fields to be more intuitive to ordinary users, so
 * minutes come last, not first.
 *
 * Revision 1.8  2015/10/25 20:36:37  agmsmith
 * Make useful strings (names of months) globally available.
 *
 * Revision 1.7  2015/10/24 17:28:43  agmsmith
 * Better documentation of what State is in setting values.
 *
 * Revision 1.6  2015/10/04 20:08:01  agmsmith
 * Implemented the TimeMatches function, now it's fully functioning!
 *
 * Revision 1.5  2015/09/30 21:55:39  agmsmith
 * Finished FromString, CronBitset is now fully functional.
 *
 * Revision 1.4  2015/09/30 14:14:29  agmsmith
 * Working on reading from strings, mostly done but still have to implement the
 * "*" type of value and do more testing.
 *
 * Revision 1.3  2015/09/28 23:31:17  agmsmith
 * Working on reading in a Cron string...
 *
 * Revision 1.2  2015/09/28 22:05:55  agmsmith
 * Work in progress on CronBitset.
 *
 * Revision 1.1  2015/09/25 16:37:48  agmsmith
 * Initial revision
 * A utility class for converting Cron style strings into bitsets, and back to
 * strings if needed.  Does parsing and storage and testing against current
 * time.
 */
