/******************************************************************************
 * $Header: /CommonBe/agmsmith/Programming/ScriptOCron/RCS/CronBitset.h,v 1.14 2015/11/24 20:11:48 agmsmith Exp $
 *
 * This utility class converts from Cron style strings to a set of dates and
 * times, which can be used more easily to test if a particular time is in the
 * specified set.  It can also convert the set back into a string.
 *
 * A set of date and time values is expressed as a Cron style string, which has
 * 5 elements, separated by white space.  The order is the reverse of the usual
 * crontab one, which seems unintuitive for end users (minute first, then
 * hour?).  First is the the day of week from 0 (Sunday) to 6 or the names of
 * days, second is the month from 1 (January) to 12 or the names of months,
 * third is the day of month from 1 to 31, fourth is the hour from 0 to 23,
 * fifth is minute from 0 to 59.
 *
 * Only the first three letters of day or month names are needed.  Ranges can
 * be shown with "-", sets of values are separated by commas, or you can use
 * "*" to specify the entire range of values.  A "?" is used for empty sets.
 * All five elements need match for a particular time and date to be considered
 * as matching the set.
 *
 * This CronBitset software is licensed under the MIT license.
 *
 * Copyright (c) 2015 by Alexander G. M. Smith.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#include <time.h>
#include <bitset>


/******************************************************************************
 * A support class for storing one kind of value (minutes, days, etc).  Has the
 * bitset, range limits for the value, and readable equivalents for the numbers
 * (days of the week for example).  Usually you'd use class CronBitset rather
 * than this class.
 */

typedef bitset<64> CronBitSet;
  /* Due to the lack of polymorphism in the bitset template (you can't pass
  different sized bitsets to a common function; there is no generic bitset
  pointer type or even a common base class), we always use a 64 bit one since
  that's big enough for the 60 minutes worst case. */

class CronElementSet
{
public: // Constructor and destructor.
  CronElementSet (int LowestValue, int HighestValue,
    const char **ValueNames = NULL, bool LeadingZero = false);

public: // Member functions.
  bool GetAllSet (); // TRUE if all bits are set, else false.
  inline int GetLowestValue () const { return m_LowestValue; };
  inline int GetHighestValue () const { return m_LowestValue + m_Size - 1; };
  inline int GetSize () const { return m_Size; };

  inline bool GetValue (int Value) const
  {
    if (Value < m_LowestValue || Value > m_LowestValue + m_Size)
      return false;
    return m_Bits [Value - m_LowestValue];
  };

  inline void SetValue (int Value, bool NewState = true)
  {
    if (Value < m_LowestValue || Value > m_LowestValue + m_Size)
      return;
    m_Bits [Value - m_LowestValue] = NewState;
  };

  void SetAll (bool NewState = true);

  const char * FromString (const char *Buffer);
    /* Convert a Cron style string into the bitset, see ToString for format
    details.  Leading white space is skipped.  Returns the point inside the
    Buffer where it finished parsing, so when successful that point is either
    white space after the element or the end of the string NUL.  If something
    goes wrong (like finding a word which isn't a known name or a number thats
    too big), returns with the pointer at the start of that bad word.  Only the
    first three letters of a value name need to be specified (case
    insensitive), but if there are more then they must match the corresponding
    portion of the name. */

  bool ToString (char *Buffer, int BufferSize) const;
    /* Converts the set of values to a Cron type string.  Numbers are used to
    list the values which are set, or their names if the names exist.  Ranges
    of sequential set values are output as Start-Finish (a hyphen) and
    individual values are comma separated.  If all are set, then "*" is output.
    If none are set, a single question mark is output.  Adds a NUL at the end
    of the string.  Returns TRUE if successful, FALSE if it ran out of string
    space.  Worst case should take under 200 bytes for English month names. */

private: // Member variables.
  CronBitSet m_Bits;
    /* Contains bits marking whether the given time value is selected (true) or
    not (false).  Bit at index zero corresponds to the m_LowestValue of the
    range.  Bit at index m_Size-1 corresponds to highest value of the range,
    any excess bits after that should be left unused and set to false. */

  bool m_LeadingZeroes;
    /* TRUE if leading zeroes should be printed, useful for two digit numeric
    values like minutes or hours. */

  int m_LowestValue;
    /* Lowest value in the range.  Quite often will be zero (minutes from 0 to
    59, etc), except for days of the month and months which both start at 1. */

  int m_Size;
    /* Number of bits used in m_Bits and the number of distinct values in the
    range or in other numbers, equal to HighestValue - LowestValue + 1.  For
    minutes, that would be 59 - 0 + 1 = 60. */

  const char **m_ValueNames;
    /* Alternative names for the different values, so you can say "Tuesday"
    rather than "2".  Points to an array (size is m_Size) of pointers to
    strings, or NULL if none specified.  The strings passed in the constructor
    are assumed to remain allocated by the user while this class is in use
    (they're not copied).  If you want to be silly, this can be used to spell
    out the numbers, so it would say July First instead of July 1. */

private: // Internal functions.
  bool ValueToString (char *Buffer, int BufferSize, int iValue) const;
};


/******************************************************************************
 * The main class for storing a Cron style date & time set and converting it to
 * or from strings and testing it against the given time value.
 */

class CronBitset
{
public: // Constructor and destructor.
  CronBitset ();

public: // Public types.
  enum CronElementTypeEnum // In the reverse of standard crontab order.
  {
    CET_DAY_OF_WEEK = 0,
    CET_MONTH,
    CET_DAY_OF_MONTH,
    CET_HOUR,
    CET_MINUTE,
    CET_NUMBER_OF_ELEMENTS
  };

public: // Member functions.
  void ClearAll ();

  bool GetValue (CronElementTypeEnum Element, int Value);
  void SetValue (CronElementTypeEnum Element, int Value, bool NewState = true);
    /* Get or set the bit of a particular single minute, hour, etc, to
    NewState. */

  void SetAll (CronElementTypeEnum Element, bool NewState = true);
    /* Set or clear all the bits in an element group. */

  bool GetAllSet (CronElementTypeEnum Element);
    /* Returns TRUE if all the bits are set in a particular element group. */

  const char * FromString (const char *Buffer);
    /* Convert a Cron style string into the collection of bitsets.  Returns
    NULL if successful, else a pointer to the word in the string which made
    parsing fail (like a number out of range). */

  bool ToString (char *Buffer, int BufferSize) const;
    /* Converts the set of values to a Cron type string.  Each element is
    separated by white space from the other elements.  Returns TRUE if
    successful, FALSE if it ran out of string space.  Worst case (every third
    value missing) takes 310 bytes, so use a 400 byte buffer to be safe or
    check the return value. */

  bool TimeMatches (const struct tm *pTime);
    /* See if a given time matches the set. */

private: // Member variables.
  CronElementSet m_Minute;
  CronElementSet m_Hour;
  CronElementSet m_DayOfMonth;
  CronElementSet m_Month;
  CronElementSet m_DayOfWeek;

  CronElementSet *a_Elements[CET_NUMBER_OF_ELEMENTS];
    /* A convenient way of accessing the elements as a sequential group, rather
    than using m_Minute etc.  [4] is minutes, [3] hours, [2] day of month, [1]
    month and [0] is day of week, which is the opposite of crontab order. */
};



/******************************************************************************
 * Useful global string constants.
 */

extern const char * g_DayNames [7];
extern const char * g_MonthNames [12];
extern const char * g_CronElementNames [CronBitset::CET_NUMBER_OF_ELEMENTS];



#if TEST_CRON_BITSET
/******************************************************************************
 * Stuff for testing this class, normally not compiled.
 */

void TestCronBitset ();
#endif


/******************************************************************************
 * $Log: CronBitset.h,v $
 * Revision 1.14  2015/11/24 20:11:48  agmsmith
 * Added MIT licensing.
 *
 * Revision 1.13  2015/10/29 20:56:56  agmsmith
 * Added leading zero for hours and minutes less than 10.
 *
 * Revision 1.12  2015/10/28 19:08:20  agmsmith
 * Added functions to test if all bits set and to set/clear all bits, also
 * worst case conversion to string seems to take more bytes.
 *
 * Revision 1.11  2015/10/28 18:39:36  agmsmith
 * Print empty sets as "?" and parse them when reading too.
 *
 * Revision 1.10  2015/10/28 18:13:40  agmsmith
 * Narrowed down worst case string length when converting to string.
 *
 * Revision 1.9  2015/10/25 21:41:13  agmsmith
 * Reverse order of time fields to be more intuitive to ordinary users, so
 * minutes come last, not first.
 *
 * Revision 1.8  2015/10/25 20:36:37  agmsmith
 * Make useful strings (names of months) globally available.
 *
 * Revision 1.7  2015/10/24 17:28:43  agmsmith
 * Better documentation of what State is in setting values.
 *
 * Revision 1.6  2015/10/04 20:08:01  agmsmith
 * Implemented the TimeMatches function, now it's fully functioning!
 *
 * Revision 1.5  2015/09/30 21:55:39  agmsmith
 * Finished FromString, CronBitset is now fully functional.
 *
 * Revision 1.4  2015/09/30 14:14:29  agmsmith
 * Working on reading from strings, mostly done but still have to implement the
 * "*" type of value and do more testing.
 *
 * Revision 1.3  2015/09/28 23:31:17  agmsmith
 * Working on reading in a Cron string...
 *
 * Revision 1.2  2015/09/28 22:05:55  agmsmith
 * Work in progress on CronBitset.
 *
 * Revision 1.1  2015/09/25 16:37:48  agmsmith
 * Initial revision
 * A utility class for converting Cron style strings into bitsets, and back to
 * strings if needed.  Does parsing and storage and testing against current
 * time.
 */
