/******************************************************************************
 * $Header: /CommonBe/agmsmith/Programming/ScriptOCron/RCS/ScriptOCron.cpp,v 1.133 2015/12/30 22:23:12 agmsmith Exp $
 *
 * This file implements the command line version of AGMSScriptOCron.
 *
 * This is a BeOS / Haiku OS program for editing and running commands without
 * needing to know about using the command line.  There's a graphical user
 * interface for editing the commands, letting you specify each argument
 * separately.  Running of commands can be triggered by "cron" style time based
 * events, the BeOS scripting system (see the Hey command), as well as a GUI
 * start button.
 *
 * Note that this uses the AGMS coding style.  That means no tabs, indents are
 * two spaces, m_ is the prefix for member variables, g_ is the prefix for
 * global names, C style comments, constants are in all capital letters and
 * most other things are mixed case, it's word wrapped to fit in 79 characters
 * per line to make proofreading on paper easier, and functions are listed in
 * reverse dependency order so that forward declarations (function prototypes
 * with no code) aren't needed.
 *
 * The command line only (no graphical user interface) AGMSScriptOCron software
 * is licensed under the MIT license.  The commercial Fetchit! software (the
 * graphical user interface layer on top of AGMSSCriptOCron) is sold under the
 * TuneTracker Systems Sales Terms and Conditions (summary: licensed to use a
 * single copy on a single computer, no reselling, no source code).  The source
 * code for the commercial version is proprietary, owned by TuneTracker Systems
 * LLC under a royalty agreement with Alexander G. M. Smith.
 *
 * The contents of this source code file are MIT licensed.
 *
 * Copyright (c) 2015 by Alexander G. M. Smith.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */


/* Standard C Library. */

#include <stdio.h>
#include <errno.h>
#include <signal.h> // For kill().
#include <unistd.h> // For fork, pipe, dup2, execve etc.

/* Standard C++ library. */

#include <iostream>
#include <sstream>

/* STL (Standard Template Library) headers. */

#include <map>
#include <vector>

/* BeOS (Be Operating System) headers. */

#include <Alert.h>
#include <Application.h>
#include <Archivable.h>
#include <ClassInfo.h> // For cast_as, probably obsolete.
#include <Directory.h>
#include <FindDirectory.h>
#include <Message.h>
#include <Path.h>
#include <PropertyInfo.h>
#include <Resources.h>
#include <String.h>

/* Our other stuff. */

#include "CronBitset.h"
#include "ScriptOCron.h"
#include "Trash.h"

/******************************************************************************
 * Global variables, and not-so-variable things too.  Grouped by functionality.
 */

static const char *g_AppSignature = "application/scriptocron";
  /* The application signature is the same for both the command line and
  commercial graphical user interface versions of the program, so that third
  party scripts will work with both.  It's also shorter than the usual
  x-vnd.blah so that it can be more easily used in "hey" commands. */

const char *g_DataName = "data"; // "data" field in a BMessage.
static const char *g_CommandListName = "CommandList"; // Used in BArchive.
const char *g_ErrorName = "error"; // Error code stored in a BMessage.
static const char *g_FieldListName = "FieldList"; // Used in archived Command.
static const char *g_LogDirName = "Logs"; // Log files go in this directory.
static const char *g_InternalName = "_name"; // Archived BHandler name field.
static const char *g_KeyOriginalModel = "Original Model"; // UserData key.
static const char *g_LogFileEnabledName = "LogFileEnabled"; // For archiving.
const char *g_MessageName = "message"; // Describes error in a BMessage.
static const char *g_ModelListName = "ModelList"; // Used for BArchived Models.
const char *g_ModelName = "model"; // Scripting field for Model's name.
static const char *g_NameBuiltIn = "builtin"; // Identifies built-in Models.
static const char *g_NameKey = "key"; // Name of UserData archived keys.
static const char *g_NameIndex = "index"; // BMessage specifier for index #.
const char *g_NameName = "name"; // Names BMessage field with new name.
const char *g_NameReboot = "@Launch"; // For the trigger string.
const char *g_NameUnnamed = "Unnamed"; // Default name for new things.
const char *g_PathName = "pathname"; // Usually log path in a BMessage.
const char *g_ResultName = "result"; // Results returned in a BMessage.
static const char *g_SettingsFileName = PROGRAM_NAME_WORD " Settings";
static const uint32 g_SettingsWhatCode = 'ASOC';
static const char *g_TemplateName = "Template"; // Used in archived Command.
static const char *g_TriggerName = "Trigger"; // Used in archived Command.
static const char *g_UserDataName = "UserData"; // Used in archiving it.
static const char *g_ValueName = "Value"; // Used in archived Fields.

static const char * g_CommandStateNames [CS_MAX] =
{ // Readable versions of the states, mostly for error messages.
  "Ready to Run",
  "Running",
  "Editing"
};

static class ScriptOCronApp *g_App = NULL;
  /* A global reference to the main BApplication, non-null once the application
  has been initialised. */

bool g_FirstTimeAppRun = false;
  /* TRUE if this is the very first time this program has been run (detected by
  the lack of a settings file).  FALSE for later times.  Useful for doing very
  first time initialisation, like setting up auto-run on system boot. */

static int g_QuitCountdown = -1;
  /* Set to the number of pulse timing events (about one every half second) to
  count down before the program quits.  Negative means stop counting.  Zero
  means quit at the next pulse event.  This is used to keep the program alive
  for a short while after someone requests that it quit, so that it has a
  chance to shut down running commands. */

bool g_SettingsNeedSaving = false;
  /* Set to TRUE by anything that changes something which may affect the
  settings file.  Once a minute the main application will check this flag and
  save the settings if needed, clearing the flag after it's done. */


#if INCLUDE_GUI_CODE
volatile bool g_AdvancedMode = false;
  /* When TRUE the GUI is running in advanced mode, otherwise basic mode. */

BRect g_InitialWindowFrame;
  /* A saved setting that specifies the initial window size and position. */

bool g_MetaMode = false;
  /* When TRUE the GUI has some of the meta things shown in advanced mode,
  currently just the Template text entry and expanded command line. */

volatile bool g_ServerMode = false;
  /* When TRUE the program runs in server mode - error messages don't result in
  pop-up dialog boxes, but you can still see them in stderr.  Also the window
  is minimized, if it exists.  Don't need to lock the Application or go through
  scripting since this is simple enough to be read directly from multiple
  threads. */
#endif

/* not const due to warnings */ char * g_PropertyNames [PN_MAX] =
{
  "Command",
  "Edit",
  "Run",
  "Serial",
  "Template",
  "CommandLine",
  "Trigger",
  "Log",
  "Model",
  "Field",
  "Name",
  "Value",
  "UserData",
  "LogFileEnable",
  "Duplicate"
};


/* This array lists the scripting commands the top level Application can
handle, in a format that the scripting system can understand too. */

static struct property_info g_AppScriptingPropertyList [] =
{
  /* *name; commands[10]; specifiers[10]; *usage; extra_data; ... */

  {g_PropertyNames[PN_SERIAL],
    {B_GET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Returns the change serial number for the Application overall.  It gets "
    "increased by some small amount (may wrap around to negative values "
    "eventually) when this object or any of its children have been changed.  "
    "Useful for knowing when to update a GUI displaying this object.",
    PN_SERIAL, {}, {}, {}},

  {g_PropertyNames[PN_COMMAND],
    {B_COUNT_PROPERTIES, 0}, {B_DIRECT_SPECIFIER, 0},
    "Counts the number of Commands currently in the list of commands.",
    PN_COMMAND, {}, {}, {}},

  {g_PropertyNames[PN_COMMAND],
    {B_CREATE_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Create a new Command.  The Command's default name is \"Unnamed\".  You "
    "can set a name for the new Command by adding a string named \"name\" "
    "to the BMessage used for scripting.  If you want the new Command to be "
    "a copy of a Model Command, add a string named \"model\" with the name "
    "of the Model to use.  If no Model is specified, you get an empty Command "
    "(no Fields, no command line, etc).  The new Command starts in Edit mode, "
    "so it doesn't accidentally start running.  The Hey command would look "
    "like: hey " PROGRAM_NAME_WORD " create Command with "
    "name=NewCommandNameHere and model=ModelNameHere",
    PN_COMMAND, {}, {}, {}},

  {g_PropertyNames[PN_COMMAND],
    {0}, {B_NAME_SPECIFIER, B_INDEX_SPECIFIER, 0},
    "Specify a particular Command sub-object for further scripting.  "
    "It's better to refer to Commands by name than by index, since the array "
    "is kept in alphabetical order and adding or removing Commands will "
    "affect the indices of several other Commands.  The Hey command syntax "
    "would be: \"hey " PROGRAM_NAME_WORD " get something of Command "
    "NameOfCommandHere\"  For indexing, it would be: \"hey "
    PROGRAM_NAME_WORD " get something of Command [0]\" (note the space "
    "between Command and [0]).  If those don't work, try \"hey "
    PROGRAM_NAME_WORD " let Command [0] do get something\" which goes through "
    "a different code path to find the target object.",
    PN_COMMAND, {}, {}, {}},

  {g_PropertyNames[PN_MODEL],
    {B_GET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Gets a list of the names of the Models, returned as an array of "
    "strings, in alphabetical order.  A Model is a saved Command that can be "
    "copied to make a new Command.  There are built-in ones for common tasks "
    "and ones which the user can create from existing Commands.",
    PN_MODEL, {}, {}, {}},

  {g_PropertyNames[PN_MODEL],
    {B_CREATE_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Creates a Model from a Command.  An extra string called \"name\" "
    "specifies the name of the Command to be copied into the new Model (which "
    "will have the same name).  Fails if the Command doesn't exist or the "
    "name is already in use by another Model.",
    PN_MODEL, {}, {}, {}},

  {g_PropertyNames[PN_MODEL],
    {B_DELETE_PROPERTY, 0}, {B_NAME_SPECIFIER, 0},
    "Deletes the Model specified by the name specifier.  Try: \"hey "
    PROGRAM_NAME_WORD " delete Model NameOfModelHere\"",
    PN_MODEL, {}, {}, {}},

  {0, {0}, {0}, 0, 0, {}, {}, {}} /* End of list of property commands. */
};


/* This array lists the scripting commands the Command object can handle.  Note
that the Name property is stored in the InternalName of the BHandler parent
class. */

static struct property_info g_CommandScriptingPropertyList [] =
{
  /* *name; commands[10]; specifiers[10]; *usage; extra_data; ... */
  {g_PropertyNames[PN_NAME],
    {B_GET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Gets the name of this Command.",
    PN_NAME, {}, {}, {}},

  {g_PropertyNames[PN_NAME],
    {B_SET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Changes the name of this Command.  This will also affect the order of "
    "the Commands array (which is always in alphabetical order), so the index "
    "to this Command and others may change.  Fails if the new name is already "
    "in use.  You can change the name at any time.",
    PN_NAME, {}, {}, {}},

  {g_PropertyNames[PN_SERIAL],
    {B_GET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Returns the change serial number for this Command, increased when this "
    "object or any of its children change in any way (including new log text, "
    "settings modifications, etc).",
    PN_SERIAL, {}, {}, {}},

  {g_PropertyNames[PN_COMMAND],
    {B_DELETE_PROPERTY, 0},
    {B_DIRECT_SPECIFIER, B_NAME_SPECIFIER, B_INDEX_SPECIFIER, 0},
    "Delete this Command.  Fails if the command is not in edit mode (running "
    "or idle).  Hey syntax is a bit odd, use \"hey " PROGRAM_NAME_WORD " let "
    "Command [3] do delete\".  A bit of a hack has been put in so you can use "
    "the more obvious \"delete Command [3]\" syntax, also \"hey "
    PROGRAM_NAME_WORD " delete Command of Command [3]\" does work.",
    PN_COMMAND, {}, {}, {}},

  {g_PropertyNames[PN_DUPLICATE],
    {B_CREATE_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Make a copy of this Command.  The new name of the command is \"Unnamed\" "
    "unless you specify it with a string called \"name\".  The new name must "
    "not exist, otherwise this operation will fail.  Can make a copy of a "
    "Command in any state; the new one will always start in the Edit state.",
    PN_DUPLICATE, {}, {}, {}},

  {g_PropertyNames[PN_EDIT],
    {B_CREATE_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Start editing this Command.  This creates a backup copy of the Command "
    "so that it can be restored to its original state if you cancel the "
    "editing.  Can only start editing if the Command isn't running or "
    "otherwise in use.  Changes the Command state to being edited, which will "
    "prevent other uses of the Command (like running it) while it is being "
    "edited.  Use \"hey " PROGRAM_NAME_WORD " let Command [3] do create "
    "edit\"",
    PN_EDIT, {}, {}, {}},

  {g_PropertyNames[PN_EDIT],
    {B_DELETE_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Cancel the editing of this Command.  Restores it to its original state, "
    "unless you've somehow created another command with the old name.  In "
    "that rare case, you'll get an error and the Command and its backup copy "
    "are deleted rather than being restored.",
    PN_EDIT, {}, {}, {}},

  {g_PropertyNames[PN_EDIT],
    {B_EXECUTE_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Finish editing the Command.  Gets rid of the backup copy of the original "
    "state of the Command, and changes the Command state to being ready to "
    "run.  The Hey command awkwardly uses \"do\" to mean execute, so it would "
    "have a double do like: hey " PROGRAM_NAME_WORD " let Command [0] do do "
    "Edit",
    PN_EDIT, {}, {}, {}},

  {g_PropertyNames[PN_EDIT],
    {B_GET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Returns true if the Command is being edited, false otherwise.",
    PN_EDIT, {}, {}, {}},

  {g_PropertyNames[PN_RUN],
    {B_EXECUTE_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Start running the Command.  Fails if it is being edited or is already "
    "running.  The Hey command awkwardly uses \"do\" to mean execute, so it "
    "would have a double do like: hey " PROGRAM_NAME_WORD " let Command [0] "
    "do do Run",
    PN_RUN, {}, {}, {}},

  {g_PropertyNames[PN_RUN],
    {B_GET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Returns true if the Command is running, false otherwise.",
    PN_RUN, {}, {}, {}},

  {g_PropertyNames[PN_TEMPLATE],
    {B_GET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Get the Template for building the command line used by this Command.",
    PN_TEMPLATE, {}, {}, {}},

  {g_PropertyNames[PN_TEMPLATE],
    {B_SET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Set the Template for building the command line used by this Command.  "
    "Various magic words in the Template will be replaced with specific "
    "values.  The most basic magic word is a Field name, which will get "
    "replaced with the Field's Value.  Another is a Field name with \"AsDir\" "
    "appended, which uses the Value up to and including the last slash "
    "(either kind) character.  Similarly, \"AsName\" is all the text after "
    "the last slash.  \"ThisCommandName\" is replaced by the name of the "
    "Command.",
    PN_TEMPLATE, {}, {}, {}},

  {g_PropertyNames[PN_COMMAND_LINE],
    {B_GET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Get the current command line for this Command.  Will substitute "
    "Field names (and other related magic words) in the Template with their "
    "current Values to make the command line.",
    PN_COMMAND_LINE, {}, {}, {}},

  {g_PropertyNames[PN_TRIGGER],
    {B_GET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Get the Trigger string used by this Command.",
    PN_TRIGGER, {}, {}, {}},

  {g_PropertyNames[PN_TRIGGER],
    {B_SET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Set the Trigger string used by this Command.  A Trigger string can be "
    "several things, which control when this Command is started.  If it's an "
    "empty string, the Command can only be manually started.  If it's "
    "\"@Launch\" then the Command starts up when " PROGRAM_NAME_WORD " "
    "starts running (usually soon after the computer boots up, if it's "
    "configured that way).  If it's a Cron style string then the Command is "
    "started whenever the current time (local time zone) matches.  Finally, "
    "if it's the name of some other Command then this Command is started "
    "when that other Command finishes successfully (successful means with an "
    "exit code of zero).  I might as well explain Cron style time strings "
    "here.  Technically they're a modified version of the Unix crontab "
    "format (look it up online if you wish to read more).  It's a way of "
    "specifying dates and times, formatted as five fields separated by "
    "white space in this order (opposite of crontab order): DayOfWeek Month "
    "DayOfMonth Hour Minute.  Each field can contain a single value, a "
    "bunch of values separated by commas, a range of values using a hyphen, "
    "or be an asterisk to mean all possible values.  All fields must match "
    "to activate a time trigger.  So to specify the top of the hour at 2am "
    "on January 6th (years aren't mentioned, so it would happen every year "
    "at that time), it would be \"* 1 6 2 0\".  To specify 11:30pm on work "
    "days it would be \"Mon-Fri * * 23 30\".  For something happening every "
    "quarter hour from noon to just before 5pm (16:45 would be the last "
    "time) it would be \"* * * 12-16 0,15,30,45\".  The sneakiest may be "
    "for things like the first Friday of the month \"Fri * 1-7 0 0\" which "
    "activates at midnight if the day is a Friday and in the first 7 days "
    "of the month (one of them must be a Friday?).  Note that the standard "
    "Unix crontab interpretation does it differently when combining day of "
    "week and day of month.  The valid range for minutes is 0 to 59, hours "
    "0-23, days 1-31, months 1-12 (January to December), day of week 0-6 "
    "(Sunday, Monday, ..., Saturday).",
    PN_TRIGGER, {}, {}, {}},

  {g_PropertyNames[PN_LOG],
    {B_GET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Get the accumulated log output (currently Standard Output and Standard "
    "Error combined) from the execution of this Command.  The accumulated "
    "data (which is stored in memory) is also cleared at this point.  The "
    "serial number is also changed whenever new log data arrives.  If you are "
    "running a particularly verbose command, it's best to periodically get "
    "the log output data to avoid missing anything.",
    PN_LOG, {}, {}, {}},

  {g_PropertyNames[PN_LOG_FILE_ENABLE],
    {B_GET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Returns a boolean that is TRUE if the output from this Command is being "
    "logged to a file.  FALSE if it's not being filed away.  There's also a "
    "\"pathname\" string with the pathname of the log file for this Command.",
    PN_LOG_FILE_ENABLE, {}, {}, {}},

  {g_PropertyNames[PN_LOG_FILE_ENABLE],
    {B_SET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Turns on or off file logging of this Command's output.  Takes effect "
    "only when the Command starts running, not during.  The boolean \"data\" "
    "field needs to be set to TRUE to enable file logging, false to turn it "
    "off.  Regular in-memory logging proceeds irregardless.  Logged text "
    "will be appended to the file; it's up to you to delete it if it gets "
    "too big.  The log file will have the same name as the Command (minus "
    "slashes) and be stored in the Logs directory in the program's "
    "subdirectory in the user settings directory.  Typically it will be "
    "named something like /boot/home/config/settings/" PROGRAM_NAME_WORD
    "/Logs/CommandName.  If you don't want it there, replace the file with "
    "a symbolic link to a different file in a different place.",
    PN_LOG_FILE_ENABLE, {}, {}, {}},

  {g_PropertyNames[PN_FIELD],
    {B_CREATE_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Create a new Field inside the Command.  The Command needs to be in Edit "
    "mode.  The Field's default name is \"Unnamed\".  You can set a name for "
    "the new Field by adding a string named \"name\" to the BMessage used for "
    "scripting.  The Hey command would look like: hey " PROGRAM_NAME_WORD
    " let Command [0] do create Field with name=NewFieldNameHere",
    PN_FIELD, {}, {}, {}},

  {g_PropertyNames[PN_FIELD],
    {B_COUNT_PROPERTIES, 0}, {B_DIRECT_SPECIFIER, 0},
    "Counts the number of Fields in the Command.",
    PN_FIELD, {}, {}, {}},

  {g_PropertyNames[PN_FIELD],
    {0}, {B_NAME_SPECIFIER, B_INDEX_SPECIFIER, 0},
    "Specify a particular Field sub-object of a Command for further "
    "scripting.  It's better to refer to Fields by name than by index, since "
    "the array is kept in alphabetical order and adding or removing Fields "
    "will affect the indices of several other Fields.",
     PN_FIELD, {}, {}, {}},

  {0, {0}, {0}, 0, 0, {}, {}, {}} /* End of list of property commands. */
};


/* This array lists the scripting commands the Field object can handle. */

static struct property_info g_FieldScriptingPropertyList [] =
{
  /* *name; commands[10]; specifiers[10]; *usage; extra_data; ... */
  {g_PropertyNames[PN_NAME],
    {B_GET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Gets the name of this Field.",
    PN_NAME, {}, {}, {}},

  {g_PropertyNames[PN_NAME],
    {B_SET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Changes the name of this Field.  This will also affect the order of "
    "the Fields array (which is always in alphabetical order), so the index "
    "to this Field and others may change.  Fails if the new name is already "
    "in use or the parent Command isn't in edit mode.",
    PN_NAME, {}, {}, {}},

  {g_PropertyNames[PN_SERIAL],
    {B_GET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Returns the change serial number for this Field, increased when this "
    "object changes in any way (usually due to user editing).",
    PN_SERIAL, {}, {}, {}},

  {g_PropertyNames[PN_FIELD],
    {B_DELETE_PROPERTY, 0},
    {B_DIRECT_SPECIFIER, B_NAME_SPECIFIER, B_INDEX_SPECIFIER, 0},
    "Delete this Field.  Fails if the parent command is not in edit mode.",
    PN_FIELD, {}, {}, {}},

  {g_PropertyNames[PN_VALUE],
    {B_GET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Gets the Value of the Field.",
    PN_VALUE, {}, {}, {}},

  {g_PropertyNames[PN_VALUE],
    {B_SET_PROPERTY, 0}, {B_DIRECT_SPECIFIER, 0},
    "Sets the Value of the Field.  This is the user provided string that "
    "will be substituted in the command line everywhere the Field name is "
    "present.",
    PN_VALUE, {}, {}, {}},

  {0, {0}, {0}, 0, 0, {}, {}, {}} /* End of list of property commands. */
};


/* This array lists the scripting commands for manipulating UserData.  It's
implemented by both Commands and Fields.  Unfortunately, we can't put it in the
BApplication too due to class inheritance limitations. */

static struct property_info g_UserDataScriptingPropertyList [] =
{
  /* *name; commands[10]; specifiers[10]; *usage; extra_data; ... */

  {g_PropertyNames[PN_USER_DATA],
    {B_SET_PROPERTY, 0}, {B_NAME_SPECIFIER, B_INDEX_SPECIFIER, 0},
    "UserData is used for storing key and value pairs of user specified "
    "data.  It's often used by a GUI on top of this program to store things "
    "associated with Commands and Fields (such as the data type related UI "
    "to pop up to edit a Field).  The key (passed in the name specifier of "
    "the scripting operation) is a non-empty UTF-8 string, with case "
    "insensitive matching.  The value can be any of the BMessage field data "
    "types (including a BMessage) and can also be an array of them.  This "
    "scripting operation changes the value stored in the UserData entry for "
    "the specified name or at the specified index to the contents of your "
    "\"data\" argument(s).  If you don't specify any data, then the contents "
    "become NULL and when you get it next time, the \"result\" field will be "
    "missing.  If the entry doesn't exist and you used the name specifier, "
    "it will be created (thus the lack of a B_CREATE_PROPERTY operation).",
    PN_USER_DATA, {}, {}, {}},

  {g_PropertyNames[PN_USER_DATA],
    {B_DELETE_PROPERTY, 0}, {B_NAME_SPECIFIER, B_INDEX_SPECIFIER, 0},
    "Deletes the UserData entry for the specified name or at the specified "
    "index.",
    PN_USER_DATA, {}, {}, {}},

  {g_PropertyNames[PN_USER_DATA],
    {B_COUNT_PROPERTIES, 0}, {B_DIRECT_SPECIFIER, 0},
    "Counts the number of UserData entries currently stored.  "
    "Coincidentally and not surprisingly, valid index specifiers are from "
    "zero up to this number minus one.",
    PN_USER_DATA, {}, {}, {}},

  {g_PropertyNames[PN_USER_DATA],
    {B_GET_PROPERTY, 0}, {B_NAME_SPECIFIER, B_INDEX_SPECIFIER, 0},
    "Gets the value stored in the UserData entry for the specified name or "
    "at the specified index number (zero based, be wary of the shuffling "
    "which happens when items are added or removed - they're kept in "
    "alphabetical order).  Returned in the usual \"result\" output, which "
    "will be missing if the stored value is NULL.\n\n"

    "Here are some standard UserData keys/values:\n\n"

    "Description/string.  Supplies a verbose description of the Command.\n\n"

    "DisplayType/[\"LocalDir\"|\"String\"|\"URL\"] Specifies how to display "
    "a Field in the GUI version of the program.  If not specified, you get "
    "the usual String display.  The LocalDir one uses a file requester to "
    "select a directory, the resulting path always ends with a slash.  The "
    "URL one breaks apart the URL (only FTP and HTTP(S) ones) into all the "
    "parts and lets the user edit each part individually before reassembling "
    "the result into a new URL (with encoding of odd characters too).\n\n"

    "GUITitles/string.  Changes the title of various GUI elements used "
    "in displaying a Field like the LocalDir one from the default of \"To\", "
    "\"Browse\" and \"Save files to\" to be whatever the strings are.  The "
    "different elements separated by \"|\" characters and the order of the "
    "elements is Field DisplayType specific.\n\n"

    "OriginalModel/string.  If present then the user has not edited the "
    "template (they may have changed fields etc).  The original Model's name "
    "is given by the string.  Later on when doing software updates, we can "
    "replace the Template of the Command with the more recent Template from "
    "the Model of the same name.  When the user changes the Template, this "
    "entry is deleted.",
    PN_USER_DATA, {}, {}, {}},

  {0, {0}, {0}, 0, 0, {}, {}, {}} /* End of list of property commands. */
};



/* Utility function for searching a g_...ScriptingPropertyList array to see if
a given command is in it. */

static bool IsItOurScriptingCommand (struct property_info *InfoPntr,
  const char *PropertyName, uint32 SpecificationKind)
{
  int iKind;

  for (; InfoPntr->name != NULL; InfoPntr++)
  {
    if (strcasecmp (PropertyName, InfoPntr->name) == 0)
    {
      for (iKind = 0; iKind < (int) sizeof (InfoPntr->specifiers) /
      (int) sizeof (InfoPntr->specifiers[0]); iKind++)
      {
        if (InfoPntr->specifiers[iKind] == 0)
          break; // Zero terminated array of specifier kinds.
        if (InfoPntr->specifiers[iKind] == SpecificationKind)
          return true; // Found it.
      }
    }
  }
  return false;
}



/******************************************************************************
 * A base class for common stuff that our BeOS/Haiku system scriptable object
 * needs.  Also implements the change tracking serial number system and the
 * UserData (a key/value dictionary for user use) scripting functionality.  The
 * base BHandler's name is used as the object's name.
 *
 * Subclasses respond to scripting messages thus the BHandler parent class.
 * It's also archiveable into a BMessage, thus the _EXPORT.  For multitasking
 * simplicity and because there isn't any performance need for it, the
 * application is used as the BLooper.  That does mean the app's BLooper has
 * zillions of BHandlers in its list, hope it doesn't get too slow.
 */

class _EXPORT BaseChangeScriptable : public BHandler
{
public: // Constructor and destructor.
  BaseChangeScriptable (const char *NewName);
  BaseChangeScriptable (BMessage *ArchiveMessagePntr);
  virtual ~BaseChangeScriptable ();

public: // BeOS virtual functions and other BeOS callbacks.
  virtual status_t Archive (BMessage *ArchiveMessagePntr,
    bool Deep = true) const;
  virtual status_t GetSupportedSuites (BMessage *MessagePntr);
  virtual void MessageReceived (BMessage *MessagePntr);
  virtual BHandler *ResolveSpecifier (BMessage *MessagePntr, int32 Index,
    BMessage *SpecifierMsgPntr, int32 SpecificationKind,
    const char *PropertyNamePntr);

public: // Our functions.
  void AddToLooper (BLooper *LooperPntr);

  virtual void ChangeHappened (bool SettingsAffected = true);
    /* Update our serial number so that users of this object can quickly tell
    that it has changed by comparing serial numbers.  Also inform our parent
    object (whatever it is) that a change happened (which is why this is a
    virtual function).  If SettingsAffected is TRUE, it will also request a
    save of the settings file. */

  int GetSerial ();
  void RemoveFromLooper ();

  bool UserDataAddString (const char *KeyPntr, const char *ValueStringPntr);
    /* Creates a UserData entry under the given Key, with the given string
    value.  Returns TRUE if successful, FALSE if something is wrong.  If the
    key already exists the value will be added as a second or later string in
    the "result" BMessage field. */

  const char * UserDataGetString (const char *KeyPntr,
    const char *DefaultValue = NULL);
    /* Looks for the UserData under the given key.  If it exists and is a
    string then it's returned.  Otherwise the DefaultValue will be returned.
    If you don't specify DefaultValue or it's NULL, an empty string will be
    used for that purpose. */

protected: // Member variables, types, etc.
 typedef map<const char *, BMessage, LessThanCharStringStruct>
   UserDataMap; // Note key string is manually allocated and freed.

  int m_ChangeSerial;
    /* Incremented every time something is changed in this object, or in a
    child object. */

  UserDataMap m_UserData;
    /* A collection of key and value pairs for storing user specified data.
    Often used by a GUI on top of this program to store things associated with
    Commands (whether to use the basic or advanced GUI for a Command) and
    Fields (such as Field data type). */
};



/******************************************************************************
 * A Field is a holder for a chunk of information that the user specifies, such
 * as a file name or URL, which gets inserted in the command line at a certain
 * spot.  It can also specify the type of information, so we can use specific
 * ways of prompting the user for the Field's value (plain string input vs
 * popping up a file requestor if we know it is a file name).
 *
 * It responds to scripting messages for doing things to the Field.  It's also
 * archiveable into a BMessage, thus the _EXPORT.
 */

class _EXPORT Field : public BaseChangeScriptable
{
public: // Constructor and destructor.
  Field (const char *NewName);
  Field (BMessage *ArchiveMessagePntr);
  virtual ~Field ();

public: // BeOS virtual functions and other BeOS callbacks.
  virtual status_t Archive (BMessage *ArchiveMessagePntr,
    bool Deep = true) const;
  virtual status_t GetSupportedSuites (BMessage *MessagePntr);
  static BArchivable *Instantiate (BMessage *ArchiveMessagePntr);
  virtual void MessageReceived (BMessage *MessagePntr);
  virtual BHandler *ResolveSpecifier (BMessage *MessagePntr, int32 Index,
    BMessage *SpecifierMsgPntr, int32 SpecificationKind,
    const char *PropertyNamePntr);

public: // Member functions.
  virtual void ChangeHappened (bool SettingsAffected = true);

  void ExpandMagicWords (BString &CommandLine);
    /* Replaces all keywords related to this field in the given string. */

  void SetParentCommand (class Command *CommandPntr);

private: // Member variables.
  class Command *m_ParentCommand;
    /* Points to the Command that owns this field, or NULL.  Useful for
    propagating changes. */

  BString m_Value;
    /* The main value of this Field.  This string will replace every occurrence
    the Field's name in the command line. */
};



/******************************************************************************
 * A Command contains everything needed to run a command line program (values
 * for the Fields, the command line template, log output).  It also responds to
 * scripting commands for doing things with the command and is archiveable into
 * a BMessage, thus the _EXPORT.
 */

class _EXPORT Command : public BaseChangeScriptable
{
public: // Constructor and destructor.
  Command (const char *NewName);
  Command (BMessage *ArchiveMessagePntr);
  virtual ~Command ();

public: // BeOS virtual functions and other BeOS callbacks.
  virtual status_t Archive (BMessage *ArchiveMessagePntr,
    bool Deep = true) const;
  virtual status_t GetSupportedSuites (BMessage *MessagePntr);
  static BArchivable *Instantiate (BMessage *ArchiveMessagePntr);
  virtual void MessageReceived (BMessage *MessagePntr);
  virtual BHandler *ResolveSpecifier (BMessage *MessagePntr, int32 Index,
    BMessage *SpecifierMsgPntr, int32 SpecificationKind,
    const char *PropertyNamePntr);

public: // Class specific types, useable by anyone.
  static const int MAX_LOG_LENGTH = 20000000;
    /* We stop adding output text to m_LogAccumulated when it has gotten this
    big in bytes.  Given a maximum of 800 megabytes free memory for the
    application to use in 32 bit BeOS, don't make it too big. */

private: // More class specific types, for use only by this class.
 typedef map<const char *, class Field *, LessThanCharStringStruct>
   FieldMap;

public: // Member functions, some semi-private internal ones (used by Fields).
  void AppendLogText (const char *AddText, int32 AddLength = -1);
  virtual void ChangeHappened (bool SettingsAffected = true);
  bool DoesNamedFieldExist (const char *FieldNamePntr);
  void GetSanitisedName (BString &OutputName);
  void InternalAddField (Field *FieldPntr);
  void InternalRemoveField (Field *FieldPntr);
  bool IsEditing ();
  bool IsRunning ();
  void OtherCommandFinished (const char *CommandNamePntr);
  void PollAndUpdate (const struct tm *CurrentTime);
  void SetState (CommandState NewState); // Normally for internal use only.

private: // Completely private member functions.
  void ExpandTemplateIntoCommandLine ();
  void ParseTriggerString ();
  status_t RunCommand (char *ErrorTextPntr, int ErrorTextSize);
  status_t StopCommand (char *ErrorTextPntr, int ErrorTextSize);

private: // Member variables.
  int m_ChildExitCode;
    /* The exit code from the child process running the command line.  Well,
    the child is usually the /bin/sh command interpreter, but it passes on up
    the exit codes if asked nicely. */

  pid_t m_ChildPID;
    /* Process ID of the child (the /bin/sh program) executing the command
    line.  Set to -1 when the child isn't running.  Also used as the process
    group number for the /bin/sh and its children, so they can conveniently all
    be killed off when required. */

  BString m_CommandLine;
    /* The last used command line for this Command.  Generated by replacing
    Field names with the corresponding Field Values.  See also
    m_CommandLineSerial. */

  int m_CommandLineSerial;
    /* Serial number of this Command at the last time the m_CommandLine was
    generated from the m_Template and Field Values.  If it is different from
    the Command's current serial number, then the m_CommandLine needs to be
    regenerated the next time it is used. */

  CommandState m_CommandState;
    /* States to prevent various things from interfering, like trying to edit a
    running command.  Only should be changed through SetState (). */

  BMessage m_CopyOfCommandBeforeEditing;
    /* When editing a Command, a copy of its state (including fields) is saved
    here in BArchive format, in case the user wants to cancel editing and
    revert back to the original state. */

  FieldMap m_FieldMap;
    /* A collection of Fields, which are used to build the command line when
    this Command is executed.  The Fields are also scriptable objects and get
    saved inside the Command when it is Archived.  The name of the Field is the
    key. */

  struct tm m_LastRunStartTime;
    /* Time when the program was last run, broken down into seconds, minutes,
    etc.  Local time zone assumed.  Initialised to all zeroes, so you can
    detect the not-run-yet state, which happens after loading and maybe
    editing. */

  BString m_LogAccumulated;
    /* A string containing the accumulated standard output and standard error
    text from running the Command.  Gets cleared when someone gets the current
    log text.  Limited to at most MAX_LOG_LENGTH bytes stored. */

  bool m_LogFileEnabled;
    /* TRUE if logging to a file is enabled for this Command.  Will open (for
    append) the file when the Command starts running, and close it when
    finished running, and write the Commands output to it while running. */

  BFile m_LogFile;
    /* The BFile is usually opened on the log file.  Closed when the Command
    isn't running.  InitCheck() will tell you if the file is open. */

  int m_PipeStdOutFromChild[2];
    /* Pairs of file handles for the pipes for communicating with the child
    process running the actual Command.  [0] is the FD (file descriptor) number
    for the read end of the pipe, [1] is the write end.  Both are set to -1
    when the pipe is not open.  They're set to O_CLOEXEC (close on executing)
    mode in the main program thread so that copies (due to the way fork()
    works) of them in later Command child threads are automatically closed.
    They're also set to O_NONBLOCK mode so that reads from them will return
    immediately if there is no data available. */

  BString m_Template;
    /* The template used for building command lines.  Magic words, like Field
    names, get replaced with related strings, like Field Values.  Also see the
    OriginalModel key in UserData, which keeps track of whether the user has
    edited the Template. */

  BString m_Trigger;
    /* The time and other things trigger for starting this Command, as
    specified by the user.  "@Launch" starts it when the program starts.  A
    cron style string specifies that it starts at the given times (parsed value
    is stored in m_TriggerCronBits).  An empty string means it can only be
    manually started.  Otherwise it is the name of another Command (which may
    not exist yet) and specifies that it starts when the other command finishes
    successfully.  See m_TriggerType for which one it is. */

  CronBitset m_TriggerCronBits;
    /* If the Trigger is a cron style string, the decoded bitsets representing
    the time are stored here.  For other trigger types, this isn't used. */

  bool m_TriggerNamedCommandHasFinished;
    /* Set to TRUE when the Command named by m_Trigger has finished
    successfully, by that Command doing a search through all other Commands.
    Cleared when this Command changes state, so that it is only used when the
    other Command finishes while this one is in the ready to run state. */

  TriggerType m_TriggerType;
    /* When m_Trigger is changed, the string is parsed to determine what kind
    it is (manual, cron, reboot, name of a Command) and the result stored in
    this enum to save on having to figure it out every time it's used. */
};



/******************************************************************************
 * ScriptOCronApp is the BApplication class for this program.  Besides a bit of
 * bookkeeping (startup, shutdown), this main thread handles scripting messages
 * that manipulate the Commands and Fields.  Conveniently, this avoids
 * multiprocessing pitfalls.  All the operations are fairly fast, so there
 * shouldn't be a bottleneck with using a single message processing thread.  If
 * built with the GUI, the window thread also uses scripting messages sent to
 * this thread to do all its actions on Commands and Fields.
 */

class ScriptOCronApp : public BApplication
{
public: // Constructor and destructor.
  ScriptOCronApp ();
  ~ScriptOCronApp ();

public: // BeOS virtual functions.
  virtual void AboutRequested ();
  virtual void DispatchMessage (BMessage *MessagePntr, BHandler *Target);
  virtual status_t GetSupportedSuites (BMessage *MessagePntr);
  virtual void MessageReceived (BMessage *MessagePntr);
  virtual void Pulse ();
  virtual void ReadyToRun ();
  virtual BHandler *ResolveSpecifier (BMessage *MessagePntr, int32 Index,
    BMessage *SpecifierMsgPntr, int32 SpecificationKind,
    const char *PropertyNamePntr);


public: // Member functions.
  void ChangeHappened (bool SettingsAffected = true);
  void DeleteAllCommands ();
  void DeleteAllModels ();
  bool DoesNamedCommandExist (const char *CommandNamePntr);
  void DoNotificationsForFinishedCommand (const char *CommandNamePntr);
  const char *GetSettingsDirectory ();
  void InternalAddCommand (Command *CommandPntr);
  void InternalRemoveCommand (Command *CommandPntr);
  void LoadModels (BMessage *ArchivePntr, bool BuiltIn);
  status_t LoadSaveSettings (bool DoLoad);

private: // Internal member functions.
  void DefaultSettings ();

private: // Class specific types.
 typedef map<const char *, class Command *, LessThanCharStringStruct>
   CommandMap;

 typedef map<const char *, BMessage *, LessThanCharStringStruct> ModelMap;

private: // Member variables.
  int m_ChangeSerial;
    /* Incremented every time something is changed in this object, or in a
    child object or in the settings.  Settings will be saved on program exit if
    anything has changed. */

  CommandMap m_CommandMap;
    /* A collection of commands, which scripting messages can be sent to.  They
    also show up in the list of commands in the main window.  There may be a
    few outside this Map, such as ones being edited (the user works on a copy
    of the command when editing).  The name of the command is the key. */

  ModelMap m_ModelMap;
    /* A collection of Models (saved Commands), each stored in archived
    BMessage format with the name of the Model (referencing a string within the
    BMessage) being the key in the map.  Used when creating a new Command to
    start off with things set for common tasks, rather than a fully empty
    Command. */

  BMessage *m_QuitRequestedStashedMessage;
    /* To delay the quit until Commands have stopped running (need a few more
    Pulse() times), when we get the B_QUIT_REQUESTED message it gets detached
    from the message queue and stored here.  After we're finished, we reply to
    it with the appropriate boolean flag in "result" and a message type of
    B_REPLY. */

  int m_SerialWhenSettingsSaved;
    /* The m_ChangeSerial at the time when the settings were last saved or
    loaded, so we can detect if they get modified. */

  BPath m_SettingsDirectoryPath;
    /* The constructor initialises this to the settings directory path.  It
    never changes after that. */

  bigtime_t m_TimeSettingsWereLastSaved;
    /* Used for automatically saving the settings once in a while. */
};



/******************************************************************************
 * UserData is stored in BMessages, usually in the "result" field.  But it
 * arrives in a field named "data" and sometimes the key string is also stored
 * in the BMessage.  This utility function copies a named field from the source
 * BMessage to the destination BMessage, optionally renaming it in the process.
 * If there are multiple fields with that name (an array), they're all copied.
 * Note that the two messages must be different, copying from a message to
 * itself won't work properly.
 */

status_t CopyBMessageField (BMessage *SourceMessagePntr,
  BMessage *DestMessagePntr, const char *FieldName, const char *NewName)
{
  if (SourceMessagePntr == NULL || DestMessagePntr == NULL ||
  SourceMessagePntr == DestMessagePntr || FieldName == NULL ||
  FieldName[0] == 0 || (NewName != NULL && NewName[0] == 0))
    return B_BAD_VALUE; // Bad input arguments, do nothing.
  if (NewName == NULL)
    NewName = FieldName;

  int32 CountFound = 0;
  const void *DataPntr = NULL;
  status_t ErrorCode;
  bool FixedSize = false;
  type_code TypeFound = 0;

  ErrorCode = SourceMessagePntr->GetInfo (FieldName, &TypeFound, &FixedSize);
  if (ErrorCode != B_OK)
    return ErrorCode;
  ErrorCode = SourceMessagePntr->GetInfo (FieldName, &TypeFound, &CountFound);
  if (ErrorCode != B_OK)
    return ErrorCode;

  for (int32 iCount = 0; iCount < CountFound; iCount++)
  {
    ssize_t DataSize = 0;
    ErrorCode = SourceMessagePntr->FindData (FieldName, TypeFound, iCount,
      &DataPntr, &DataSize);
    if (ErrorCode != B_OK)
      return ErrorCode;

    ErrorCode = DestMessagePntr->AddData (NewName, TypeFound, DataPntr,
      DataSize, FixedSize, (iCount == 0) ? CountFound : 1 /* prealloc */);
    if (ErrorCode != B_OK)
      return ErrorCode;
  }
  return B_OK;
}



/******************************************************************************
 * Global utility function to return the current time as a readable string.
 * Not multithread safe - uses a global string array to return the result.
 */

static const char * CurrentTimeAsString ()
{
  static char TimeString[30];
  time_t TimeNow;
  struct tm BrokenUpTime;

  time (&TimeNow);
  localtime_r (&TimeNow, &BrokenUpTime);
  asctime_r (&BrokenUpTime, TimeString);

  // Strip off the end of line character.
  char *EOL = strchr (TimeString, '\n');
  if (EOL != NULL)
    *EOL = 0;

  return TimeString;
}



/******************************************************************************
 * Global utility function to display an error message and return.  The message
 * part describes the error, and if ErrorNumber is non-zero, gets the string
 * ", error code $X (standard description)." appended to it.  If the message
 * is NULL then it gets defaulted to "Something went wrong".  The title part
 * doesn't get displayed (no title bar in the dialog box, but you can see it in
 * the debugger as the window thread name), and defaults to "Error Message" if
 * you didn't specify one.  If running in command line mode, the error gets
 * printed to stderr rather than showing up in a dialog box.
 */

void DisplayErrorMessage (
  const char *MessageString = NULL,
  int ErrorNumber = 0,
  const char *TitleString = NULL)
{
  char ErrorBuffer [PATH_MAX + 1500];

  if (TitleString == NULL)
    TitleString = PROGRAM_NAME_WORD " Error Message";

  if (MessageString == NULL)
  {
    if (ErrorNumber == 0)
      MessageString = "No error, no message, why bother?";
    else
      MessageString = "Something went wrong";
  }

  if (ErrorNumber != 0)
  {
    sprintf (ErrorBuffer, "%s, error code $%X/%d (%s) has occurred.",
      MessageString, ErrorNumber, ErrorNumber, strerror (ErrorNumber));
    MessageString = ErrorBuffer;
  }

#if INCLUDE_GUI_CODE
  if (!g_ServerMode)
  {
    BAlert *AlertPntr;
    AlertPntr = new BAlert (TitleString, MessageString,
      "Acknowledge", NULL, NULL, B_WIDTH_AS_USUAL, B_STOP_ALERT);
    AlertPntr->SetShortcut (0, B_ESCAPE);
    if (AlertPntr != NULL)
      AlertPntr->Go (); // Synchronous mode, will wait until done.
  }
  else // Print message on standard error.
#endif //INCLUDE_GUI_CODE
  cerr << TitleString << ": " << MessageString << endl;
}



/******************************************************************************
 * A utility function to get the next globally defined (so that multiple
 * simultaneous users get unique values) serial number.
 */

static int GetNextSerialNumber ()
{
#if __HAIKU__
  static long g_NextSerial = 0;
#else // BeOS
  static vint32 g_NextSerial = 0;
#endif

  return (int) atomic_add (&g_NextSerial, 1);
}



/******************************************************************************
 * Word wrap a long line of text into shorter LineLength column lines and print
 * the result on the given output stream.  Appends a new line.  Zero LineLength
 * means don't wrap.
 */

static void WrapTextToStream (ostream& OutputStream, const char *TextPntr,
  const int LineLength = 79)
{
  if (LineLength <= 0)
  {
    OutputStream << TextPntr << endl;
    return;
  }

  char *StringPntr;
  char  TempString [LineLength+1]; // How modern, variable size locals!

  TempString[LineLength] = 0; /* Only needs to be done once. */

  while (*TextPntr != 0)
  {
    while (*TextPntr == ' ' || *TextPntr == '\t')
      TextPntr++; /* Skip leading spaces and tabs, but not newlines. */
    if (*TextPntr == 0)
      break; /* It was all spaces, don't print any more. */

    strncpy (TempString, TextPntr, LineLength);

    /* Advance StringPntr to the end of the temp string, partly to see how long
    it is (rather than doing strlen).  Stop at new line characters too, so they
    appear as a shorter line and don't wrap funny. */

    StringPntr = TempString;
    while (*StringPntr != 0 && *StringPntr != '\n')
      StringPntr++;

    if (StringPntr - TempString < LineLength)
    {
      /* This line fits completely. */

      if (*StringPntr == '\n')
        *StringPntr++ = 0; // Don't print line feed character + skip it too.

      OutputStream << TempString << endl;
      TextPntr += StringPntr - TempString;
      continue;
    }

    /* Advance StringPntr to the last space in the temp string. */

    while (StringPntr > TempString)
    {
      if (isspace (*StringPntr))
        break; /* Found the trailing space. */
      else /* Go backwards, looking for the trailing space. */
        StringPntr--;
    }

    /* Remove more trailing spaces at the end of the line, in case there were
    several spaces in a row. */

    while (StringPntr > TempString && isspace (StringPntr[-1]))
      StringPntr--;

    /* Print the line of text and advance the text pointer too. */

    if (StringPntr == TempString)
    {
      /* This line has no spaces, don't wrap it, just split off a chunk. */
      OutputStream << TempString << endl;
      TextPntr += strlen (TempString);
      continue;
    }

    *StringPntr = 0; /* Cut off after the first trailing space. */
    OutputStream << TempString << endl;
    TextPntr += StringPntr - TempString;
  }
}



/******************************************************************************
 * Print the usage info to the stream.  Optionally includes a list of all
 * commands and do word wrapping (zero to not wrap).
 */

void PrintUsage (ostream& OutputStream, int WrapWidth, bool ListCommands)
{
  struct property_info *PropInfoPntr;

#if INCLUDE_GUI_CODE
  WrapTextToStream (OutputStream,
    PROGRAM_NAME_FANCY " Downloader and command launcher.\n\n"
    "Copyright © 2015 TuneTracker Systems LLC.  Sold "
    "under the TuneTracker Systems Sales Terms and Conditions.", WrapWidth);
#else
  WrapTextToStream (OutputStream,
    PROGRAM_NAME_FANCY " - For editing and running commands.\n\n"
    PROGRAM_NAME_FANCY " copyright © 2015 by Alexander G. M. Smith.  "
    "Licensed under the MIT license.  Send e-mail to agmsmith@ncf.ca "
    "with questions or comments.", WrapWidth);
#endif // INCLUDE_GUI_CODE
  WrapTextToStream (OutputStream, "\n$Revision: 1.133 $ "
    __DATE__ ", " __TIME__ ".", WrapWidth);

#if INCLUDE_GUI_CODE
  if (!ListCommands)
    return; // Shorter about box for commercial GUI version.
#endif

  WrapTextToStream (OutputStream,
    "\nThis program maintains a library of commands which can be executed "
    "when requested or when a specified time of day arrives.  The commands "
    "can also be edited with a graphical user interface (available "
    "commercially) designed to keep the complexity down for ordinary "
    "users.\n\n"

    "A specialist sets up a template for each kind of command, specifying the "
    "full command line arguments and marking out key fields within those "
    "arguments for user editing, along with providing descriptions and "
    "default values for those key fields.  Then the ordinary user only needs "
    "to fill in the key fields, assisted by the descriptions.\n\n"

    "Commands are triggered either by a cron style time specification, "
    "by the user hitting the Run button, or by BeOS/Haiku scripting "
    "operations (which can be sent via the \"Hey\" utility program).\n\n"

    "Speaking of \"Hey\", scripting operations can be used for "
    "creating new Commands, running them and even modifying "
    "existing ones.  Send Hey's GetSuites command to the various different "
    "objects in " PROGRAM_NAME_WORD " (the Application itself, Command "
    "objects, Field objects) to see what scripting operations are available.  "
    "You can also see the list of scripting operations if you run "
    PROGRAM_NAME_WORD " from the command line.", WrapWidth);

  if (!ListCommands)
    return;

  OutputStream << "\n"
    "Scripting operations:\n\n"
    "Quit\n";
  WrapTextToStream (OutputStream,
    "Stop the program.  Useful if it's running as a server.", WrapWidth);
  OutputStream << endl;

  /* Go through all our scripting commands and add a description of each one to
  the usage text. */

  static property_info * ScriptingSuites[] =
  {
    g_AppScriptingPropertyList,
    g_CommandScriptingPropertyList,
    g_FieldScriptingPropertyList,
    g_UserDataScriptingPropertyList
  };

  int iSuite;
  for (iSuite = 0;
  iSuite < (int) sizeof (ScriptingSuites) / (int) sizeof (ScriptingSuites[0]);
  iSuite++)
  {
    for (PropInfoPntr = ScriptingSuites[iSuite];
    PropInfoPntr->name != 0; PropInfoPntr++)
    {
      switch (PropInfoPntr->commands[0])
      {
        case B_GET_PROPERTY:
          OutputStream << "Get " << PropInfoPntr->name << endl;
          break;

        case B_SET_PROPERTY:
          OutputStream << "Set " << PropInfoPntr->name << " NewValue" << endl;
          break;

        case B_COUNT_PROPERTIES:
          OutputStream << "Count " << PropInfoPntr->name << endl;
          break;

        case B_CREATE_PROPERTY:
          OutputStream << "Create " << PropInfoPntr->name << endl;
          break;

        case B_DELETE_PROPERTY:
          OutputStream << "Delete " << PropInfoPntr->name << endl;
          break;

        case B_EXECUTE_PROPERTY:
          OutputStream << "Execute " << PropInfoPntr->name << endl;
          break;

        default:
          OutputStream << "N/A " << PropInfoPntr->name << endl;
          break;
      }
      WrapTextToStream (OutputStream, PropInfoPntr->usage, WrapWidth);
      OutputStream << endl;
    }
  }
}



/******************************************************************************
 * Wrapper function for printing usage via stream commands, like
 * cout << PrintUsage;
 */

ostream& PrintUsage (ostream& OutputStream)
{
  PrintUsage (OutputStream, 79 /* WrapWidth */, true /* ListCommands */);
  OutputStream.flush ();
  return OutputStream;
}



/******************************************************************************
 * Implementation of the BaseChangeScriptable class, constructor, destructor
 * and the rest of the member functions in mostly alphabetical order.
 */

BaseChangeScriptable::BaseChangeScriptable (const char *NewName) :
  BHandler (NewName),
  m_ChangeSerial (GetNextSerialNumber ())
{
}


BaseChangeScriptable::BaseChangeScriptable (BMessage *ArchiveMessagePntr) :
  BHandler (ArchiveMessagePntr),
  m_ChangeSerial (GetNextSerialNumber ())
{
  int32 CountFound = 0;
  status_t ErrorCode;
  type_code TypeFound = 0;
  UserDataMap::iterator UserDataIter;

  // Load up the UserData key/value pairs.

  ErrorCode = ArchiveMessagePntr->
    GetInfo (g_UserDataName, &TypeFound, &CountFound);
  if (ErrorCode == B_OK && TypeFound == B_MESSAGE_TYPE && CountFound > 0)
  {
    int32 iUserData;
    for (iUserData = 0; iUserData < CountFound; iUserData++)
    {
      BMessage UserDataMessage;
      ErrorCode = ArchiveMessagePntr->
        FindMessage (g_UserDataName, iUserData, &UserDataMessage);
      if (ErrorCode != B_OK)
        continue; // Skip bad ones.
      const char *KeyPntr = NULL;
      if (UserDataMessage.FindString (g_NameKey, &KeyPntr) != B_OK)
        continue; // Always need a key.
      if (KeyPntr == NULL || KeyPntr[0] == 0)
        continue; // Don't allow empty string keys.

      UserDataIter = m_UserData.find (KeyPntr);
      if (UserDataIter != m_UserData.end ())
        continue; // Duplicate key, ignore redundant keys.

      KeyPntr = strdup (KeyPntr); // Allocate our own memory for string.
      if (KeyPntr == NULL)
        continue;

      BMessage *ValueMessagePntr = &(m_UserData[KeyPntr]); // Allocates BMsg.
      ValueMessagePntr->what = B_SIMPLE_DATA;

      ErrorCode = CopyBMessageField (&UserDataMessage, ValueMessagePntr,
        g_ResultName);
      if (ErrorCode != B_OK) // Can happen if no "result" field.
        ValueMessagePntr->MakeEmpty (); // Effectively NULL user data.
    }
  }
}


BaseChangeScriptable::~BaseChangeScriptable ()
{
  RemoveFromLooper ();

  // Need to explicitly deallocate the UserData key strings.  The BMessage
  // portion of the key/value pair is automatically deallocated by the map.

  UserDataMap::iterator Iter;
  for (Iter = m_UserData.begin (); Iter != m_UserData.end (); Iter++)
  {
    free (const_cast<char *> (Iter->first));
  }
  m_UserData.clear ();
}


/* Adds this scriptable object (or rather its BHandler ancestor) to the given
BLooper (usually the main BApplication).  Needed for receiving scripting
messages, even if ResolveSpecifier explicitly returns this object as the
BHandler for a particular message. */

void BaseChangeScriptable::AddToLooper (BLooper *LooperPntr)
{
  if (Looper () == NULL && LooperPntr != NULL && LooperPntr->Lock ())
  {
    LooperPntr->AddHandler (this);
    LooperPntr->Unlock ();
  }
}


status_t BaseChangeScriptable::Archive (
  BMessage *ArchiveMessagePntr, bool Deep) const
{
  status_t ErrorCode = BHandler::Archive (ArchiveMessagePntr, Deep);
  if (ErrorCode != B_OK)
    return ErrorCode;

  UserDataMap::const_iterator UserDataIter; // Const since in const function.
  for (UserDataIter = m_UserData.begin ();
  UserDataIter != m_UserData.end (); UserDataIter++)
  {
    // Copy the UserData BMessage (which just has a "result" field with the
    // user's data in it) and add the key to it.  That combo then goes in the
    // archive.

    BMessage TempMessage (B_SIMPLE_DATA);

    ErrorCode = TempMessage.AddString (g_NameKey, UserDataIter->first);
    if (ErrorCode != B_OK)
      return ErrorCode;

    ErrorCode = CopyBMessageField (const_cast<BMessage *> (
      &UserDataIter->second), &TempMessage, g_ResultName);
    if (ErrorCode != B_OK)
      return ErrorCode;

    ErrorCode = ArchiveMessagePntr->AddMessage (g_UserDataName, &TempMessage);
    if (ErrorCode != B_OK)
      return ErrorCode;
  }

  return B_OK;
}


void BaseChangeScriptable::ChangeHappened (bool SettingsAffected)
{
  m_ChangeSerial = GetNextSerialNumber ();
}


int BaseChangeScriptable::GetSerial ()
{
  return m_ChangeSerial;
}


status_t BaseChangeScriptable::GetSupportedSuites (BMessage *MessagePntr)
{
  BPropertyInfo TempPropInfo (g_UserDataScriptingPropertyList);
  MessagePntr->AddString ("suites", SUITE_BASE "userdata");
  MessagePntr->AddFlat ("messages", &TempPropInfo);
  return BHandler::GetSupportedSuites (MessagePntr);
}


void BaseChangeScriptable::MessageReceived (BMessage *MessagePntr)
{
  status_t              ErrorCode = B_ERROR;
  const char           *PropertyNamePntr;
  BMessage              ReplyMessage (B_REPLY);
  int32                 SpecifierIndex;
  int32                 SpecifierKind;
  BMessage              SpecifierMessage;
  char                  TempString [PATH_MAX + 1024];

  switch (MessagePntr->what)
  {
    case B_GET_PROPERTY:
    case B_SET_PROPERTY:
    case B_COUNT_PROPERTIES:
    case B_CREATE_PROPERTY:
    case B_DELETE_PROPERTY:
    case B_EXECUTE_PROPERTY:
      // Look at specifier[0] to find out what the scripting does.
      if (B_OK == MessagePntr->SetCurrentSpecifier (0))
        ErrorCode = MessagePntr->GetCurrentSpecifier (&SpecifierIndex,
        &SpecifierMessage, &SpecifierKind, &PropertyNamePntr);
      else // No specifiers, assume it is a direct operation on this UserData.
      {
        PropertyNamePntr = g_PropertyNames[PN_USER_DATA];
        SpecifierKind = B_DIRECT_SPECIFIER;
        ErrorCode = B_OK;
      }

      if (ErrorCode == B_OK &&
      IsItOurScriptingCommand (g_UserDataScriptingPropertyList,
      PropertyNamePntr, SpecifierKind))
        break; // It's one of ours, do it.
      // Else fall through, and punt the message to the parent class.

    default: // Not one of our messages, let the parent handle it.
      BHandler::MessageReceived (MessagePntr);
      return;
  }

  snprintf (TempString, sizeof (TempString),
    "Scripting operation code %c%c%c%c (get, set, count, etc) or access type "
    "%d (direct=1, index=2, reverse index=3, range=4, reverse range=5, "
    "name=6, ID=7) is unsupported for property \"%s\" of object \"%s\"",
    (char) (MessagePntr->what >> 24),
    (char) (MessagePntr->what >> 16),
    (char) (MessagePntr->what >>  8),
    (char) (MessagePntr->what >>  0),
    (int) SpecifierKind, PropertyNamePntr, Name ());
  ErrorCode = B_BAD_SCRIPT_SYNTAX;

  // Handle the UserData property.

  if (strcasecmp (PropertyNamePntr, g_PropertyNames[PN_USER_DATA]) == 0)
  {
    // Decode the specifier into a key for the UserData.

    const char *KeyPntr = NULL;

    if (SpecifierKind == B_NAME_SPECIFIER)
    {
      ErrorCode = SpecifierMessage.FindString (g_NameName, &KeyPntr);
      if (ErrorCode != B_OK || KeyPntr == NULL || KeyPntr[0] == 0)
      {
        ErrorCode = B_BAD_SCRIPT_SYNTAX;
        snprintf (TempString, sizeof (TempString),
          "\"%s\" string field is missing or empty in the scripting specifier "
          "for the \"%s\" property of object \"%s\"",
          g_NameName, PropertyNamePntr, Name ());
        goto ErrorExit;
      }
    }
    else if (SpecifierKind == B_INDEX_SPECIFIER)
    {
      // Look for the numbered (indexed) UserData key, the slow way.

      int32 DesiredIndex = 0;
      ErrorCode = SpecifierMessage.FindInt32 (g_NameIndex, &DesiredIndex);
      if (ErrorCode != B_OK)
      {
        ErrorCode = B_BAD_SCRIPT_SYNTAX;
        snprintf (TempString, sizeof (TempString), "\"%s\" field is "
          "missing in the scripting specifier for the \"%s\" property of "
          "object \"%s\"",
          g_NameIndex, PropertyNamePntr, Name ());
        goto ErrorExit;
      }

      if (DesiredIndex < 0 || DesiredIndex >= (int) m_UserData.size ())
      {
        ErrorCode = B_BAD_INDEX;
        snprintf (TempString, sizeof (TempString), "\"%s\" field for "
          "property \"%s\" is %d, which is not between 0 and %d, for "
          "object \"%s\"",
          g_NameIndex, PropertyNamePntr, (int) DesiredIndex,
          (int) m_UserData.size() - 1, Name ());
        goto ErrorExit;
      }

      UserDataMap::iterator UserDataIter = m_UserData.begin ();
      int32 i = 0;
      while (i < DesiredIndex && UserDataIter != m_UserData.end ())
      {
        UserDataIter++;
        i++;
      }
      if (UserDataIter == m_UserData.end ())
        goto ErrorExit; // Shouldn't happen.

      KeyPntr = UserDataIter->first;
    }

    if (MessagePntr->what == B_GET_PROPERTY)
    {
      if (KeyPntr == NULL)
        goto ErrorExit;

      UserDataMap::iterator UserDataIter = m_UserData.find (KeyPntr);
      if (UserDataIter == m_UserData.end ())
      {
        ErrorCode = B_NAME_NOT_FOUND;
        snprintf (TempString, sizeof (TempString), "\"%s\" key does not "
          "exist for UserData of object \"%s\"", KeyPntr, Name ());
        goto ErrorExit;
      }

      ErrorCode = CopyBMessageField (&UserDataIter->second,
        &ReplyMessage, g_ResultName);
      if (ErrorCode != B_OK)
      {
        snprintf (TempString, sizeof (TempString), "\"%s\" key copying data "
          "to output BMessage failed for object \"%s\"", KeyPntr, Name ());
        goto ErrorExit;
      }
      goto SuccessExit;
    }

    if (MessagePntr->what == B_SET_PROPERTY)
    {
      if (KeyPntr == NULL)
        goto ErrorExit;

      // Copy the "data" field(s) into the saved BMessage "result" field(s).
      // If the key doesn't exist, create a new key/BMessage.

      UserDataMap::iterator UserDataIter = m_UserData.find (KeyPntr);
      BMessage *ValueMessagePntr = NULL;
      if (UserDataIter == m_UserData.end ()) // New key needed.
      {
        // Need to explicitly allocate memory for the key string.

        char *NewKeyPntr = strdup (KeyPntr);
        if (NewKeyPntr == NULL)
        {
          ErrorCode = B_NO_MEMORY;
          snprintf (TempString, sizeof (TempString), "Out of memory when "
            "copying \"%s\" key string for %s of object \"%s\"",
            KeyPntr, PropertyNamePntr, Name ());
          goto ErrorExit; // Unlikely, and will blow up elsewhere anyway...
        }
        ValueMessagePntr = &(m_UserData[NewKeyPntr]);
        ValueMessagePntr->what = B_SIMPLE_DATA;
      }
      else // Key exists.
        ValueMessagePntr = &(UserDataIter->second);

      ValueMessagePntr->MakeEmpty ();
      ErrorCode = CopyBMessageField (MessagePntr, ValueMessagePntr,
        g_DataName, g_ResultName);
      if (ErrorCode != B_OK) // Field not found, or other errors okayish.
        ValueMessagePntr->MakeEmpty ();

      ChangeHappened ();
      goto SuccessExit;
    }

    // Count the number of UserData entries.

    if (MessagePntr->what == B_COUNT_PROPERTIES &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddInt32 (g_ResultName, m_UserData.size ());
      goto SuccessExit;
    }

    if (MessagePntr->what == B_DELETE_PROPERTY)
    {
      if (KeyPntr == NULL)
        goto ErrorExit;

      UserDataMap::iterator UserDataIter = m_UserData.find (KeyPntr);
      if (UserDataIter == m_UserData.end ())
      {
        ErrorCode = B_NAME_NOT_FOUND;
        snprintf (TempString, sizeof (TempString), "\"%s\" key does not "
          "exist for UserData of object \"%s\"", KeyPntr, Name ());
        goto ErrorExit;
      }

      char *OldKeyPntr = const_cast<char *> (UserDataIter->first);
      m_UserData.erase (UserDataIter);
      free (OldKeyPntr);

      ChangeHappened ();
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on UserData property.
  }

ErrorExit: /* Error message in TempString, return code in ErrorCode. */
  ReplyMessage.AddInt32 (g_ErrorName, ErrorCode);
  ReplyMessage.AddString (g_MessageName, TempString);
  ErrorCode = MessagePntr->SendReply (&ReplyMessage,
    g_App /* Reply's reply handler */, 500000 /* send timeout */);
  return;

SuccessExit: /* Success. */
  ReplyMessage.AddInt32 (g_ErrorName, B_OK);
  ErrorCode = MessagePntr->SendReply (&ReplyMessage,
    g_App /* Reply's reply handler */, 500000 /* send timeout */);
}


/* Removes this object from the BLooper it is assigned to.  Will no longer
receive scripting messages. */

void BaseChangeScriptable::RemoveFromLooper ()
{
  BLooper *MyLooperPntr = Looper ();
  if (MyLooperPntr != NULL && MyLooperPntr->Lock ())
  {
    MyLooperPntr->RemoveHandler (this);
    MyLooperPntr->Unlock ();
  }
}


BHandler *BaseChangeScriptable::ResolveSpecifier (BMessage *MessagePntr,
  int32 Index, BMessage *SpecifierMsgPntr, int32 SpecificationKind,
  const char *PropertyNamePntr)
{
#if 0
printf ("Resolving Specifier in BaseChangeScriptable object, Message is: ");
MessagePntr->PrintToStream();
printf ("Specifier index %d, Specifier kind %d, Property \"%s\"\n",
  (int) Index, (int) SpecificationKind, PropertyNamePntr);
#endif

  if (Index == 0 && IsItOurScriptingCommand (g_UserDataScriptingPropertyList,
  PropertyNamePntr, SpecificationKind))
  {
    return this; /* It's one of our properties, self will handle message. */
  }

  /* Handle an unrecognized scripting Field, let the parent figure it out. */

  return BHandler::ResolveSpecifier (
    MessagePntr, Index, SpecifierMsgPntr, SpecificationKind, PropertyNamePntr);
}


bool BaseChangeScriptable::UserDataAddString (
  const char *KeyPntr, const char *ValueStringPntr)
{
  if (KeyPntr == NULL || KeyPntr[0] == 0 || ValueStringPntr == NULL)
    return false; // Don't allow empty string keys.

  UserDataMap::iterator UserDataIter;
  BMessage *ValueMessagePntr = NULL;

  UserDataIter = m_UserData.find (KeyPntr);
  if (UserDataIter == m_UserData.end ())
  {
    // Need to make a new UserData entry, doesn't exist.

    KeyPntr = strdup (KeyPntr); // Allocate our own memory for string.
    if (KeyPntr == NULL)
      return false;

    ValueMessagePntr = &(m_UserData[KeyPntr]); // Allocates BMsg.
    ValueMessagePntr->what = B_SIMPLE_DATA;
  }
  else
    ValueMessagePntr = &(UserDataIter->second);

  ValueMessagePntr->AddString (g_ResultName, ValueStringPntr);
  return true;
}


const char * BaseChangeScriptable::UserDataGetString (
  const char *KeyPntr, const char *DefaultValue)
{
  if (DefaultValue == NULL)
    DefaultValue = "";

  if (KeyPntr == NULL || KeyPntr[0] == 0)
    return DefaultValue; // Don't allow empty string keys.

  UserDataMap::iterator UserDataIter;

  UserDataIter = m_UserData.find (KeyPntr);
  if (UserDataIter == m_UserData.end ())
    return DefaultValue; // Key not found.

  BMessage *ValueMessagePntr = &(UserDataIter->second);
  const char *ReturnValue = NULL;
  if (B_OK != ValueMessagePntr->FindString (g_ResultName, &ReturnValue))
    return DefaultValue;
  return ReturnValue;
}



/******************************************************************************
 * Implementation of the Command class, constructor, destructor and the rest of
 * the member functions in mostly alphabetical order.
 */

Command::Command (const char *NewName) :
  BaseChangeScriptable (NewName),
  m_ChildExitCode (0),
  m_ChildPID (-1),
  m_CommandLineSerial (-1),
  m_CommandState (CS_EDITING),
  m_LogFileEnabled (true),
  m_TriggerNamedCommandHasFinished (false),
  m_TriggerType (TRG_MANUAL)
{
  memset (&m_LastRunStartTime, 0, sizeof (m_LastRunStartTime));

  int i;
  for (i = 0; i < 2; i++)
  {
    m_PipeStdOutFromChild [i] = -1;
  }
}


Command::Command (BMessage *ArchiveMessagePntr) :
  BaseChangeScriptable (ArchiveMessagePntr),
  m_ChildExitCode (0),
  m_ChildPID (-1),
  m_CommandLineSerial (-1),
  m_CommandState (CS_EDITING),
  m_LogFileEnabled (true),
  m_TriggerNamedCommandHasFinished (false),
  m_TriggerType (TRG_MANUAL)
{
  int32 CountFound = 0;
  status_t ErrorCode;
  type_code TypeFound = 0;

  memset (&m_LastRunStartTime, 0, sizeof (m_LastRunStartTime));

  int i;
  for (i = 0; i < 2; i++)
  {
    m_PipeStdOutFromChild [i] = -1;
  }

  ErrorCode = ArchiveMessagePntr->FindBool (g_LogFileEnabledName,
  &m_LogFileEnabled);
  if (ErrorCode != B_OK)
    m_LogFileEnabled = true;

  ErrorCode = ArchiveMessagePntr->FindString (g_TemplateName, &m_Template);
  if (ErrorCode != B_OK)
    m_Template = "";

  ErrorCode = ArchiveMessagePntr->FindString (g_TriggerName, &m_Trigger);
  if (ErrorCode != B_OK)
    m_Trigger = "";
  ParseTriggerString ();

  ErrorCode = ArchiveMessagePntr->
    GetInfo (g_FieldListName, &TypeFound, &CountFound);
  if (ErrorCode == B_OK && TypeFound == B_MESSAGE_TYPE && CountFound > 0)
  {
    int32 iField;
    for (iField = 0; iField < CountFound; iField++)
    {
      BMessage ArchiveMsg;
      ErrorCode = ArchiveMessagePntr->
        FindMessage (g_FieldListName, iField, &ArchiveMsg);
      if (ErrorCode != B_OK)
        continue; // Skip bad ones.
      BArchivable *DearchivedPntr = instantiate_object (&ArchiveMsg);
      if (DearchivedPntr == NULL)
        continue; // Unable to find C++ class of object to rebuild it, etc.

      Field *FieldPntr = cast_as (DearchivedPntr, Field);
      if (FieldPntr != NULL)
        InternalAddField (FieldPntr);
      else // Some other kind of object, someone's been hacking the settings?
        delete DearchivedPntr;
    }
  }
}


Command::~Command ()
{
  // Delete all Fields in this Command.

  FieldMap::iterator FieldIter;
  for (FieldIter = m_FieldMap.begin ();
  FieldIter != m_FieldMap.end (); FieldIter++)
  {
    Field *FieldPntr = FieldIter->second;
    delete FieldPntr;
  }
  m_FieldMap.clear ();
}


/* Archives the Command object into a BMessage, for storage (usually in the
settings file). */

status_t Command::Archive (BMessage *ArchiveMessagePntr, bool Deep) const
{
  status_t ErrorCode;

  ErrorCode = BaseChangeScriptable::Archive (ArchiveMessagePntr, Deep);
  if (ErrorCode != B_OK)
    return ErrorCode;

  // BHandler's Archive function adds the class name of "Command" too rather
  // than "BHandler", so this isn't really needed, at least in BeOS R5.03 and
  // Haiku OS.
/*
  ErrorCode = ArchiveMessagePntr->AddString ("class", "Command");
  if (ErrorCode != B_OK)
    return ErrorCode;
*/
  ErrorCode = ArchiveMessagePntr->AddBool (g_LogFileEnabledName,
    m_LogFileEnabled);
  if (ErrorCode != B_OK)
    return ErrorCode;

  ErrorCode = ArchiveMessagePntr->AddString (g_TemplateName, m_Template);
  if (ErrorCode != B_OK)
    return ErrorCode;

  ErrorCode = ArchiveMessagePntr->AddString (g_TriggerName, m_Trigger);
  if (ErrorCode != B_OK)
    return ErrorCode;

  if (Deep)
  {
    FieldMap::const_iterator FieldIter; // Since this function is const too.
    for (FieldIter = m_FieldMap.begin ();
    FieldIter != m_FieldMap.end (); FieldIter++)
    {
      BMessage ArchiveMsg (B_ARCHIVED_OBJECT);

      Field *FieldPntr = FieldIter->second;
      ErrorCode = FieldPntr->Archive (&ArchiveMsg, Deep);
      if (ErrorCode != B_OK)
        return ErrorCode;
      ErrorCode =
        ArchiveMessagePntr->AddMessage (g_FieldListName, &ArchiveMsg);
      if (ErrorCode != B_OK)
        return ErrorCode;
    }
  }

  return ErrorCode;
}


/* Appends text to the log string, and to the log file too if it is open. */

void Command::AppendLogText (const char *AddText, int32 AddLength)
{
  if (AddText == NULL || AddText[0] == 0 || AddLength == 0)
    return;

  if (AddLength < 0)
    AddLength = strlen (AddText);

  // Write to the log file.

  if (m_LogFile.InitCheck () == B_OK)
  {
    ssize_t AmountWritten = m_LogFile.Write (AddText, AddLength);
    if (AmountWritten != AddLength)
      m_LogFile.Unset (); // Stop writing on error.
  }

  // Extra text past the max length of our in-memory buffer gets dropped.

  int32 OldLength = m_LogAccumulated.Length ();
  if (OldLength + AddLength >= MAX_LOG_LENGTH)
    AddLength = MAX_LOG_LENGTH - OldLength;
  if (AddLength > 0)
    m_LogAccumulated.Append (AddText, AddLength);
}


/* This Command has been changed, increment the serial number and let the
parent object know about the change. */

void Command::ChangeHappened (bool SettingsAffected)
{
  BaseChangeScriptable::ChangeHappened (SettingsAffected);
  g_App->ChangeHappened (SettingsAffected);
}


/* Like the function name says... */

bool Command::DoesNamedFieldExist (const char *FieldNamePntr)
{
  FieldMap::iterator FieldIter;
  FieldIter = m_FieldMap.find (FieldNamePntr);
  return FieldIter != m_FieldMap.end ();
}


void Command::ExpandTemplateIntoCommandLine ()
{
  // Is the previously stored command line out of date?

  int CurrentSerialNumber = GetSerial ();
  if (m_CommandLineSerial != CurrentSerialNumber)
  {
    // Yes it is, generate command line by substituting keywords in the
    // Template text.  Each field does its own substitution.

    m_CommandLine = m_Template;

    FieldMap::iterator FieldIter;
    FieldIter = m_FieldMap.begin ();
    while (FieldIter != m_FieldMap.end ())
    {
      FieldIter->second->ExpandMagicWords (m_CommandLine);
      FieldIter++;
    }

    m_CommandLine.ReplaceAll ("ThisCommandName", Name ());

    m_CommandLineSerial = CurrentSerialNumber;
  }
}


void Command::GetSanitisedName (BString &OutputName)
{
  OutputName = Name ();
  if (OutputName == "" || OutputName == "." || OutputName == ".." ||
  OutputName.ByteAt (0) == '#')
  {
    OutputName.Prepend ("#");
  }
  OutputName.ReplaceAll ('/', '-'); // No slashes allowed in file names.
}


/* Returns a list of the scripting commands a Command understands. */

status_t Command::GetSupportedSuites (BMessage *MessagePntr)
{
  BPropertyInfo TempPropInfo (g_CommandScriptingPropertyList);
  MessagePntr->AddString ("suites", SUITE_BASE "command");
  MessagePntr->AddFlat ("messages", &TempPropInfo);
  return BaseChangeScriptable::GetSupportedSuites (MessagePntr);
}


/* Class global function for recreating objects from a BMessage archive. */

BArchivable *Command::Instantiate (BMessage *ArchiveMessagePntr)
{
  if (validate_instantiation (ArchiveMessagePntr, "Command"))
    return new Command (ArchiveMessagePntr);
  return NULL;
}


/* Add the given Field to the Map of all Fields.  Used by the Field object when
it is renaming itself.  Also used when adding a newly dearchived Field.
Replaces any existing Field with the same name.  Also updates the BLooper and
parent pointer. */

void Command::InternalAddField (Field *FieldPntr)
{
  FieldMap::iterator FieldIter;
  FieldIter = m_FieldMap.find (FieldPntr->Name ());
  if (FieldIter != m_FieldMap.end ())
  {
    Field *OldFieldPntr = FieldIter->second;

    if (OldFieldPntr == FieldPntr)
      return; // It's already added to our Map of Fields.  Do nothing.

    m_FieldMap.erase (FieldIter);
    delete OldFieldPntr; // Will deLooper it as part of destructor.
  }

  m_FieldMap[FieldPntr->Name ()] = FieldPntr;
  FieldPntr->AddToLooper (g_App);
  FieldPntr->SetParentCommand (this);
}


/* Remove the given Field from the Map of all Fields (but don't delete it).
Used by the Field object when it is renaming itself.  Also updates the BLooper
and parent pointer. */

void Command::InternalRemoveField (Field *FieldPntr)
{
  FieldMap::iterator FieldIter;
  FieldIter = m_FieldMap.find (FieldPntr->Name ());
  if (FieldIter != m_FieldMap.end ())
  {
    Field *OldFieldPntr = FieldIter->second;
    if (OldFieldPntr == FieldPntr)
    {
      m_FieldMap.erase (FieldIter);
      OldFieldPntr->RemoveFromLooper ();
      OldFieldPntr->SetParentCommand (NULL);
    }
  }
}


/* Getter that returns TRUE if in edit mode. */

bool Command::IsEditing ()
{
  return m_CommandState == CS_EDITING;
}


/* Getter that returns TRUE if in run mode. */

bool Command::IsRunning ()
{
  return m_CommandState == CS_RUNNING;
}


/* Process mostly scripting messages. */

void Command::MessageReceived (BMessage *MessagePntr)
{
  const char           *ArgumentString = NULL;
  status_t              ErrorCode = B_ERROR;
  const char           *PropertyName;
  BMessage              ReplyMessage (B_REPLY);
  int32                 SpecifierIndex;
  int32                 SpecifierKind;
  BMessage              SpecifierMessage;
  char                  TempString [PATH_MAX + 1024];

  switch (MessagePntr->what)
  {
    case B_GET_PROPERTY:
    case B_SET_PROPERTY:
    case B_COUNT_PROPERTIES:
    case B_CREATE_PROPERTY:
    case B_DELETE_PROPERTY:
    case B_EXECUTE_PROPERTY:
      // Look at specifier[0] to find out what the scripting does.
      if (B_OK == MessagePntr->SetCurrentSpecifier (0))
        ErrorCode = MessagePntr->GetCurrentSpecifier (&SpecifierIndex,
        &SpecifierMessage, &SpecifierKind, &PropertyName);
      else // No specifiers, assume it is a direct operation on this Command.
      {
        PropertyName = g_PropertyNames[PN_COMMAND];
        SpecifierKind = B_DIRECT_SPECIFIER;
        ErrorCode = B_OK;
      }

      if (ErrorCode == B_OK &&
      IsItOurScriptingCommand (g_CommandScriptingPropertyList,
      PropertyName, SpecifierKind))
        break; // It's one of ours, do it.
      // Else fall through, and punt the message to the parent class.

    default: // Not one of our messages, let the parent handle it.
      BaseChangeScriptable::MessageReceived (MessagePntr);
      return;
  }

  snprintf (TempString, sizeof (TempString),
    "Scripting operation code %c%c%c%c (get, set, count, etc) or access type "
    "%d (direct=1, index=2, reverse index=3, range=4, reverse range=5, "
    "name=6, ID=7) is unsupported for property \"%s\" of the Command "
    "object named \"%s\".",
    (char) (MessagePntr->what >> 24),
    (char) (MessagePntr->what >> 16),
    (char) (MessagePntr->what >>  8),
    (char) (MessagePntr->what >>  0),
    (int) SpecifierKind, PropertyName, Name ());
  ErrorCode = B_BAD_SCRIPT_SYNTAX;

  // Handle the Name property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_NAME]) == 0)
  {
    // Get the current name.

    if (MessagePntr->what == B_GET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddString (g_ResultName, Name ());
      goto SuccessExit;
    }

    // Set the new Name.  This also means changing the global Map of Commands
    // since they are in alphabetical order.

    if (MessagePntr->what == B_SET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ErrorCode = MessagePntr->FindString (g_DataName, &ArgumentString);
      if (ErrorCode == B_OK)
      {
        if (g_App->DoesNamedCommandExist (ArgumentString))
        {
          snprintf (TempString, sizeof (TempString),
            "Command named \"%s\" already exists, can't rename \"%s\" to it",
            ArgumentString, Name ());
          ErrorCode = B_NAME_IN_USE;
          goto ErrorExit;
        }

        g_App->InternalRemoveCommand (this);
        SetName (ArgumentString);
        g_App->InternalAddCommand (this);
        ChangeHappened ();
        goto SuccessExit;
      }
      else
      {
        snprintf (TempString, sizeof (TempString),
          "Need to specify a \"%s\" string value to set the Name property "
          "from.  Was for Command \"%s\"",
          g_DataName, Name ());
        goto ErrorExit;
      }
    }

    goto ErrorExit; // Unknown operation on Name property.
  }

  // Handle the Serial property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_SERIAL]) == 0)
  {
    // Get the current serial number.

    if (MessagePntr->what == B_GET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddInt32 (g_ResultName, GetSerial ());
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on Serial property.
  }

  // Handle the Command property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_COMMAND]) == 0)
  {
    if (MessagePntr->what == B_DELETE_PROPERTY /* any kind of specifier OK */)
    {
      if (m_CommandState != CS_EDITING)
      {
        snprintf (TempString, sizeof (TempString),
          "Can't delete Command named \"%s\" because it is %s "
          "instead of %s", Name (), g_CommandStateNames [m_CommandState],
          g_CommandStateNames [CS_EDITING]);
        ErrorCode = B_BUSY;
        goto ErrorExit;
      }

      // Delete the corresponding log file, by moving it to trash.

      BString PathName;
      GetSanitisedName (PathName);
      PathName.Prepend ("/");
      PathName.Prepend (g_LogDirName);
      PathName.Prepend ("/");
      PathName.Prepend (g_App->GetSettingsDirectory ());

      // Don't really care if trashing fails, just leaves a log file, and the
      // Command is deleted as far as we are concerned, so it's not an error.
      ErrorCode = trash (PathName.String ());

      g_App->InternalRemoveCommand (this); // DeLoopers this too.
      g_App->ChangeHappened ();
      delete this; // Careful, can't reference member variables after here.
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on Command property.
  }

  // Handle the Duplicate property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_DUPLICATE]) == 0)
  {
    if (MessagePntr->what == B_CREATE_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ErrorCode = MessagePntr->FindString (g_NameName, &ArgumentString);
      if (ErrorCode != B_OK)
        ArgumentString = g_NameUnnamed;

      if (g_App->DoesNamedCommandExist (ArgumentString))
      {
        ErrorCode = B_NAME_IN_USE;
        snprintf (TempString, sizeof (TempString),
          "Can't create a duplicate Command named \"%s\" because it "
          "already exists", ArgumentString);
        goto ErrorExit;
      }

      // Make a copy of the command in BMessage format, including contained
      // fields.

      BMessage MsgCopyOfCommand (B_ARCHIVED_OBJECT);
      ErrorCode = Archive (&MsgCopyOfCommand, true /* deep */);
      if (ErrorCode != B_OK)
      {
        snprintf (TempString, sizeof (TempString),
          "Pre-duplication failed to archive Command \"%s\" into a BMessage",
          Name ());
        goto ErrorExit;
      }

      BArchivable *DearchivedPntr = instantiate_object (&MsgCopyOfCommand);
      Command *DuplicateCommandPntr = NULL;
      if (DearchivedPntr == NULL ||
      (DuplicateCommandPntr = cast_as (DearchivedPntr, Command)) == NULL)
      {
        delete DuplicateCommandPntr; // NULL or wrong kind of object.
        ErrorCode = B_ERROR;
        snprintf (TempString, sizeof (TempString),
          "Failed to instantiate a duplicate of Command \"%s\"", Name ());
        goto ErrorExit;
      }

      DuplicateCommandPntr->SetState (CS_EDITING);
      DuplicateCommandPntr->SetName (ArgumentString);
      g_App->InternalAddCommand (DuplicateCommandPntr);
      DuplicateCommandPntr->ChangeHappened ();

      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on Duplicate property.
  }

  // Handle the Template property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_TEMPLATE]) == 0)
  {
    // Get the current Template string.

    if (MessagePntr->what == B_GET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddString (g_ResultName, m_Template);
      goto SuccessExit;
    }

    // Set the Template string, but only in Edit mode.

    if (MessagePntr->what == B_SET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ErrorCode = MessagePntr->FindString (g_DataName, &ArgumentString);
      if (ErrorCode != B_OK)
      {
        snprintf (TempString, sizeof (TempString),
          "Need to specify a \"%s\" string value to set the Template property "
          "from.  Was for Command \"%s\"",
          g_DataName, Name ());
        goto ErrorExit;
      }

      if (m_CommandState != CS_EDITING)
      {
        snprintf (TempString, sizeof (TempString),
          "Can't set Template of Command named \"%s\" because it is %s "
          "instead of %s", Name (), g_CommandStateNames [m_CommandState],
          g_CommandStateNames [CS_EDITING]);
        ErrorCode = B_BUSY;
        goto ErrorExit;
      }

      m_Template = ArgumentString;

      // Delete the OriginalModel UserData entry now that the user has edited
      // the template.

      UserDataMap::iterator UserDataIter =
        m_UserData.find (g_KeyOriginalModel);
      if (UserDataIter != m_UserData.end ())
      {
        char *OldKeyPntr = const_cast<char *> (UserDataIter->first);
        m_UserData.erase (UserDataIter);
        free (OldKeyPntr);
      }

      ChangeHappened ();
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on Template property.
  }

  // Handle the Trigger property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_TRIGGER]) == 0)
  {
    // Get the current Trigger string.

    if (MessagePntr->what == B_GET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddString (g_ResultName, m_Trigger);
      goto SuccessExit;
    }

    // Set the Trigger string, but only in Edit mode.

    if (MessagePntr->what == B_SET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ErrorCode = MessagePntr->FindString (g_DataName, &ArgumentString);
      if (ErrorCode != B_OK)
      {
        snprintf (TempString, sizeof (TempString),
          "Need to specify a \"%s\" string value to set the Trigger property "
          "from.  Was for Command \"%s\"",
          g_DataName, Name ());
        goto ErrorExit;
      }

      if (m_CommandState != CS_EDITING)
      {
        snprintf (TempString, sizeof (TempString),
          "Can't set Trigger of Command named \"%s\" because it is %s "
          "instead of %s", Name (), g_CommandStateNames [m_CommandState],
          g_CommandStateNames [CS_EDITING]);
        ErrorCode = B_BUSY;
        goto ErrorExit;
      }

      m_Trigger = ArgumentString;
      ParseTriggerString ();
      ChangeHappened ();
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on Trigger property.
  }

  // Handle the CommandLine property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_COMMAND_LINE]) == 0)
  {
    // Get the current command line string.

    if (MessagePntr->what == B_GET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ExpandTemplateIntoCommandLine ();
      ReplyMessage.AddString (g_ResultName, m_CommandLine);
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on CommandLine property.
  }

  // Handle the Edit property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_EDIT]) == 0)
  {
    // Start editing.

    if (MessagePntr->what == B_CREATE_PROPERTY) // All kinds of specifiers OK.
    {
      if (m_CommandState != CS_READY_TO_RUN)
      {
        snprintf (TempString, sizeof (TempString),
          "Can't edit Command named \"%s\" because it is busy %s",
          Name (), g_CommandStateNames [m_CommandState]);
        ErrorCode = B_BUSY;
        goto ErrorExit;
      }

      // Make a copy of the current state of the command, including contained
      // fields.

      m_CopyOfCommandBeforeEditing.MakeEmpty ();
      m_CopyOfCommandBeforeEditing.what = B_ARCHIVED_OBJECT;
      ErrorCode = Archive (&m_CopyOfCommandBeforeEditing, true /* deep */);
      if (ErrorCode != B_OK)
      {
        snprintf (TempString, sizeof (TempString),
          "Failed to make a backup copy of Command named \"%s\" before "
          "starting editing.  Editing operation aborted", Name ());
        goto ErrorExit;
      }

      SetState (CS_EDITING);
      ChangeHappened (false /* Settings file doesn't need saving */);
      goto SuccessExit;
    }

    // Cancel editing.

    if (MessagePntr->what == B_DELETE_PROPERTY) // All kinds of specifiers OK.
    {
      if (m_CommandState != CS_EDITING)
      {
        snprintf (TempString, sizeof (TempString),
          "Can't cancel editing of Command named \"%s\" because it is %s "
          "instead of %s", Name (), g_CommandStateNames [m_CommandState],
          g_CommandStateNames [CS_EDITING]);
        ErrorCode = B_BUSY;
        goto ErrorExit;
      }

      // Restore the orginal version of the command by dearchiving the copy,
      // deleting the old Command and inserting the newly dearchived copy.  If
      // no previous copy exists (such as from totally new Commands which start
      // in Edit state), the Command disappears silently.

      BArchivable *DearchivedPntr =
        instantiate_object (&m_CopyOfCommandBeforeEditing);
      Command *OriginalCommandPntr = NULL;
      if (DearchivedPntr == NULL ||
      (OriginalCommandPntr = cast_as (DearchivedPntr, Command)) == NULL)
        delete DearchivedPntr; // NULL or wrong kind of object.

      g_App->InternalRemoveCommand (this);
      if (OriginalCommandPntr != NULL)
      {
        OriginalCommandPntr->SetState (CS_READY_TO_RUN);
        g_App->InternalAddCommand (OriginalCommandPntr);
        OriginalCommandPntr->ChangeHappened ();
      }
      else // At least let parent know we're going away.
        g_App->ChangeHappened ();
      delete this; // No more references to member variables after this!
      goto SuccessExit;
    }

    // Finish editing (Execute Edit).

    if (MessagePntr->what == B_EXECUTE_PROPERTY) // All kinds of specifiers OK.
    {
      if (m_CommandState != CS_EDITING)
      {
        snprintf (TempString, sizeof (TempString),
          "Can't finish editing of Command named \"%s\" because it is %s "
          "instead of %s", Name (), g_CommandStateNames [m_CommandState],
          g_CommandStateNames [CS_EDITING]);
        ErrorCode = B_BUSY;
        goto ErrorExit;
      }

      // Delete the backup copy of the command as it was before editing.

      m_CopyOfCommandBeforeEditing.MakeEmpty ();
      m_CopyOfCommandBeforeEditing.what = 0;
      SetState (CS_READY_TO_RUN);
      ChangeHappened (false /* Settings file doesn't need saving */);
      goto SuccessExit;
    }

    // Get editing state.

    if (MessagePntr->what == B_GET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddBool (g_ResultName, m_CommandState == CS_EDITING);
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on Edit property.
  }

  // Handle the Run property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_RUN]) == 0)
  {
    // Start running (Execute Run).

    if (MessagePntr->what == B_EXECUTE_PROPERTY) // All kinds of specifiers OK.
    {
      if (g_QuitCountdown >= 0)
      {
        snprintf (TempString, sizeof (TempString),
          "Can't run Command named \"%s\" because " PROGRAM_NAME_WORD
          " is in the process of shutting down", Name ());
        ErrorCode = B_BUSY;
        goto ErrorExit;
      }

      if (m_CommandState != CS_READY_TO_RUN)
      {
        snprintf (TempString, sizeof (TempString),
          "Can't run Command named \"%s\" because it is %s "
          "instead of %s", Name (), g_CommandStateNames [m_CommandState],
          g_CommandStateNames [CS_READY_TO_RUN]);
        ErrorCode = B_BUSY;
        goto ErrorExit;
      }

      ErrorCode = RunCommand (TempString, sizeof (TempString));
      if (ErrorCode != B_OK)
        goto ErrorExit;

      snprintf (TempString, sizeof (TempString),
        "\n=================================================================="
        "==============\n\nCommand \"%s\" started on %s.\n%s\n",
        Name (), CurrentTimeAsString (), m_CommandLine.String ());
      AppendLogText (TempString);

      SetState (CS_RUNNING);
      ChangeHappened (false /* Settings file doesn't need saving */);
      goto SuccessExit;
    }

    // Stop running, abort, abort!

    if (MessagePntr->what == B_DELETE_PROPERTY) // All kinds of specifiers OK.
    {
      if (m_CommandState != CS_RUNNING)
      {
        snprintf (TempString, sizeof (TempString),
          "Can't stop Command named \"%s\" because it is %s "
          "instead of %s", Name (), g_CommandStateNames [m_CommandState],
          g_CommandStateNames [CS_RUNNING]);
        ErrorCode = B_BUSY;
        goto ErrorExit;
      }

      ErrorCode = StopCommand (TempString, sizeof (TempString));
      if (ErrorCode != B_OK)
        goto ErrorExit;

      goto SuccessExit;
    }

    // Get the current run state.

    if (MessagePntr->what == B_GET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddBool (g_ResultName, m_CommandState == CS_RUNNING);
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on Run property.
  }

  // Handle the Log property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_LOG]) == 0)
  {
    // Get the accumulated log output.

    if (MessagePntr->what == B_GET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddString (g_ResultName, m_LogAccumulated);
      m_LogAccumulated = "";
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on Log property.
  }

  // Handle the LogFile property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_LOG_FILE_ENABLE]) == 0)
  {
    // Get the current logging enabled status.

    if (MessagePntr->what == B_GET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddBool (g_ResultName, m_LogFileEnabled);

      BString PathName;
      GetSanitisedName (PathName);
      PathName.Prepend ("/");
      PathName.Prepend (g_LogDirName);
      PathName.Prepend ("/");
      PathName.Prepend (g_App->GetSettingsDirectory ());
      ReplyMessage.AddString (g_PathName, PathName);
      goto SuccessExit;
    }

    // Set the logging enabled status.

    if (MessagePntr->what == B_SET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      bool TempBool = false;
      ErrorCode = MessagePntr->FindBool (g_DataName, &TempBool);
      if (ErrorCode != B_OK)
      {
        snprintf (TempString, sizeof (TempString),
          "Need to specify a \"%s\" boolean value to set the \"%s\" "
          "property.  Was for Command \"%s\"",
          g_DataName, PropertyName, Name ());
        goto ErrorExit;
      }

      m_LogFileEnabled = TempBool;
      ChangeHappened ();
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on LogFile property.
  }

  // Handle the Field property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_FIELD]) == 0)
  {
    // Count the number of Fields.

    if (MessagePntr->what == B_COUNT_PROPERTIES &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddInt32 (g_ResultName, m_FieldMap.size ());
      goto SuccessExit;
    }

    // Create a new Field, optionally with a name.

    if (MessagePntr->what == B_CREATE_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ErrorCode = MessagePntr->FindString (g_NameName, &ArgumentString);
      if (ErrorCode != B_OK)
        ArgumentString = g_NameUnnamed;

      if (!IsEditing ())
      {
        snprintf (TempString, sizeof (TempString),
          "Can't create a Field named \"%s\" because the related "
          "Command named \"%s\" isn't in edit mode",
          ArgumentString, Name ());
        ErrorCode = B_BUSY;
        goto ErrorExit;
      }

      if (DoesNamedFieldExist (ArgumentString))
      {
        snprintf (TempString, sizeof (TempString),
          "Can't create a Field named \"%s\" because it already exists "
          "inside Command \"%s\"",
          ArgumentString, Name ());
        ErrorCode = B_NAME_IN_USE;
        goto ErrorExit;
      }

      Field *NewFieldPntr = new Field (ArgumentString);
      InternalAddField (NewFieldPntr);
      ChangeHappened ();
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on a Field property.
  }

ErrorExit: /* Error message in TempString, return code in ErrorCode. */
  ReplyMessage.AddInt32 (g_ErrorName, ErrorCode);
  ReplyMessage.AddString (g_MessageName, TempString);
  ErrorCode = MessagePntr->SendReply (&ReplyMessage,
    g_App /* Reply's reply handler */, 500000 /* send timeout */);
  return;

SuccessExit: /* Success. */
  ReplyMessage.AddInt32 (g_ErrorName, B_OK);
  ErrorCode = MessagePntr->SendReply (&ReplyMessage,
    g_App /* Reply's reply handler */, 500000 /* send timeout */);
}


/* If we were waiting for another Command to finish, and it did finish, then
get ready to start this Command. */

void Command::OtherCommandFinished (const char *CommandNamePntr)
{
  if (m_TriggerType != TRG_NAMED_COMMAND)
    return;

  if (strcasecmp (CommandNamePntr, m_Trigger.String ()) != 0)
    return;

  if (m_CommandState != CS_READY_TO_RUN)
    return;

  m_TriggerNamedCommandHasFinished = true;
}


/* Examine the m_Trigger string to see what kind it is, and update related
variables with decoded values. */

void Command::ParseTriggerString ()
{
  if (m_Trigger.Length () == 0)
    m_TriggerType = TRG_MANUAL;
  else if (strcasecmp (m_Trigger.String (), g_NameReboot) == 0)
    m_TriggerType = TRG_REBOOT;
  else if (m_TriggerCronBits.FromString (m_Trigger.String ()) == NULL)
    m_TriggerType = TRG_CRON;
  else
    m_TriggerType = TRG_NAMED_COMMAND;
}


/* Do periodic polling operations for the command.  If it's running, see if
there is any output text and log it.  Also see if it has finished running and
get the final status code.  If it's ready to run, see if it wants to start
running. */

void Command::PollAndUpdate (const struct tm *CurrentTime)
{
  if (m_CommandState == CS_RUNNING)
  {
    // Try reading the pipe.  If successful, return and don't bother checking
    // to see if the child program is dead (want to empty out the pipe first).

    status_t ErrorCode = B_OK;

    if (m_PipeStdOutFromChild [0] != -1)
    {
      char ReadBuffer [1024 * 64]; // BeOS gives us 4K, Haiku 64K at most.
      ssize_t AmountRead = -1;

      AmountRead = read (m_PipeStdOutFromChild [0], ReadBuffer,
        sizeof (ReadBuffer));
      if (AmountRead < 0)
      {
        ErrorCode = errno;
        if (ErrorCode == EAGAIN)
          return; // No new data available yet, a normal situation.
        // Read error of some sort, fall through and see if the program is
        // dead.
      }
      else if (AmountRead == 0)
      {
        // End of file, means the other end of the pipe has been closed.  Fall
        // through and see if the child program is dead.
      }
      else // AmountRead > 0
      {
        AppendLogText (ReadBuffer, AmountRead);
        ChangeHappened (false /* Settings file doesn't need saving */);
        return;
      }
    }

    // No more data is readable, close the pipe.

    if (m_PipeStdOutFromChild [0] != -1)
    {
      close (m_PipeStdOutFromChild [0]);
      m_PipeStdOutFromChild [0] = -1;
    }

    // See if the child program has exited.  Retrieve the return code if so.
    // This also "reaps" a zombie child (final step in the process life cycle).

    m_ChildExitCode = -1;
    pid_t ProcessIDAffected = waitpid (m_ChildPID, &m_ChildExitCode, WNOHANG);
    if (ProcessIDAffected == 0)
      return; // Child hasn't stopped.

    // Child has stopped, or an error happened (-1 in that case).

    char TempString [PATH_MAX + 1024];
    snprintf (TempString, sizeof (TempString),
      "Command \"%s\" finished on %s.\nIt provided an exit code of %d (by "
      "convention zero means OK).\n", Name (), CurrentTimeAsString (),
      m_ChildExitCode);
    AppendLogText (TempString);
    m_LogFile.Unset (); // Stop logging, closes the file.

    m_ChildPID = -1;

    // Notify other Commands that this one has finished, in case they were
    // waiting for this one to start up.

    if (m_ChildExitCode == 0)
      g_App->DoNotificationsForFinishedCommand (Name ());

    SetState (CS_READY_TO_RUN); // State change after, avoids notify self.
    ChangeHappened (false /* Settings file doesn't need saving */);
    return;
  }

  if (m_CommandState == CS_READY_TO_RUN)
  {
    // See if this Command should start running now.

    bool bStartRunning = false;
    switch (m_TriggerType)
    {
      case TRG_CRON:
        if (m_LastRunStartTime.tm_min != CurrentTime->tm_min ||
        m_LastRunStartTime.tm_hour != CurrentTime->tm_hour ||
        m_LastRunStartTime.tm_mday != CurrentTime->tm_mday ||
        m_LastRunStartTime.tm_mon != CurrentTime->tm_mon ||
        m_LastRunStartTime.tm_year != CurrentTime->tm_year ||
        m_LastRunStartTime.tm_isdst != CurrentTime->tm_isdst)
        {
          // Last start time is in a different minute than current time (only
          // start something once per minute).

          if (m_TriggerCronBits.TimeMatches (CurrentTime))
            bStartRunning = true;
        }
        break;

      case TRG_NAMED_COMMAND:
        if (m_TriggerNamedCommandHasFinished)
        {
          bStartRunning = true;
          m_TriggerNamedCommandHasFinished = false;
        }
        break;

      case TRG_REBOOT:
        if (m_LastRunStartTime.tm_year == 0 &&
        m_LastRunStartTime.tm_min == 0 &&
        m_LastRunStartTime.tm_hour == 0 &&
        m_LastRunStartTime.tm_mday == 0 &&
        m_LastRunStartTime.tm_mon == 0)
          bStartRunning = true;
        break;

      default:
        break;
    }

    if (bStartRunning)
    {
      BMessage RunScriptingMessage (B_EXECUTE_PROPERTY);
      RunScriptingMessage.AddSpecifier (g_PropertyNames[PN_RUN]);
      RunScriptingMessage.AddSpecifier (g_PropertyNames[PN_COMMAND], Name ());
      g_App->PostMessage (&RunScriptingMessage);
    }
  }
}


/* A Command is looking for some object to execute a scripting command. */

BHandler * Command::ResolveSpecifier (
  BMessage *MessagePntr,
  int32 Index,
  BMessage *SpecifierMsgPntr,
  int32 SpecificationKind,
  const char *PropertyNamePntr)
{
  status_t ErrorCode = B_OK;
  char TempString [PATH_MAX + 1024];

#if 0
printf ("Resolving Specifier in Command object, Message is: ");
MessagePntr->PrintToStream();
printf ("Specifier index %d, Specifier kind %d, Property \"%s\"\n",
  (int) Index, (int) SpecificationKind, PropertyNamePntr);
#endif

  /* See if it is one of our sub-object selection properties. */

  if (strcasecmp (PropertyNamePntr, g_PropertyNames[PN_FIELD]) == 0)
  {
    // Looking for a named Field?

    if (SpecificationKind == B_NAME_SPECIFIER)
    {
      const char *NamePntr = NULL;
      ErrorCode = SpecifierMsgPntr->FindString (g_NameName, &NamePntr);
      if (ErrorCode != B_OK)
      {
        ErrorCode = B_BAD_SCRIPT_SYNTAX;
        snprintf (TempString, sizeof (TempString),
          "\"%s\" string field is missing in the scripting specifier for "
          "the \"%s\" property", g_NameName, PropertyNamePntr);
        goto ErrorExit;
      }

      FieldMap::iterator FieldIter;
      FieldIter = m_FieldMap.find (NamePntr);
      if (FieldIter != m_FieldMap.end ())
      {
        // Found the named Field, pass on the rest of the scripting command
        // to the Field object.

        MessagePntr->PopSpecifier ();
        return FieldIter->second;
      }
      ErrorCode = B_NAME_NOT_FOUND;
      snprintf (TempString, sizeof (TempString),
        "Property \"%s\" sub-object specified by name \"%s\" not found inside "
        "Command named \"%s\"",
        PropertyNamePntr, NamePntr, Name ());
      goto ErrorExit;
    }
    else if (SpecificationKind == B_INDEX_SPECIFIER)
    {
      // Look for the numbered (indexed) Field, the slow way.

      int32 DesiredIndex = 0;
      ErrorCode = SpecifierMsgPntr->FindInt32 (g_NameIndex, &DesiredIndex);
      if (ErrorCode != B_OK)
      {
        ErrorCode = B_BAD_SCRIPT_SYNTAX;
        snprintf (TempString, sizeof (TempString), "\"%s\" field is "
          "missing in the scripting specifier for the \"%s\" property of "
          "Command \"%s\"",
          g_NameIndex, PropertyNamePntr, Name ());
        goto ErrorExit;
      }

      if (DesiredIndex < 0 || DesiredIndex >= (int) m_FieldMap.size ())
      {
        ErrorCode = B_BAD_INDEX;
        snprintf (TempString, sizeof (TempString), "\"%s\" field for "
          "property \"%s\" is %d, which is not between 0 and %d, for "
          "Command \"%s\"",
          g_NameIndex, PropertyNamePntr, (int) DesiredIndex,
          (int) m_FieldMap.size() - 1, Name ());
        goto ErrorExit;
      }

      FieldMap::iterator FieldIter;
      FieldIter = m_FieldMap.begin ();
      int32 i = 0;
      while (i < DesiredIndex && FieldIter != m_FieldMap.end ())
      {
        FieldIter++;
        i++;
      }
      if (FieldIter != m_FieldMap.end ())
      {
        MessagePntr->PopSpecifier ();
        return FieldIter->second;
      }
      ErrorCode = B_ERROR;
      strcpy (TempString, "Command::ResolveSpecifier shouldn't happen");
      goto ErrorExit;
    }
  }

  if (Index == 0 && IsItOurScriptingCommand (g_CommandScriptingPropertyList,
  PropertyNamePntr, SpecificationKind))
  {
    return this; /* It's one of our properties, self will handle message. */
  }

  /* Handle an unrecognized scripting command, let the parent figure it out. */

  return BaseChangeScriptable::ResolveSpecifier (
    MessagePntr, Index, SpecifierMsgPntr, SpecificationKind, PropertyNamePntr);

ErrorExit: /* Error message in TempString, return code in ErrorCode. */
  BMessage ReplyMessage (/* B_REPLY or maybe */ B_MESSAGE_NOT_UNDERSTOOD);
  ReplyMessage.AddInt32 (g_ErrorName, ErrorCode);
  ReplyMessage.AddString (g_MessageName, TempString);
  ErrorCode = MessagePntr->SendReply (&ReplyMessage,
    g_App /* Reply's reply handler */, 500000 /* send timeout */);
  return NULL; // Stop trying to resolve the message.
}


status_t Command::RunCommand (char *ErrorTextPntr, int ErrorTextSize)
{
  status_t ErrorCode = B_ERROR;
  pid_t ChildThreadID = -1;

  // Time stamp the start time, even if we fail, so that it doesn't try to
  // start it repeatedly.

  time_t CurrentTime;
  time (&CurrentTime);
  localtime_r (&CurrentTime, &m_LastRunStartTime);

  ExpandTemplateIntoCommandLine ();
  BString CommandLine (m_CommandLine); // Local copy for child thread to use.

  // Create the pipe for transfering standard output data from the child.

  snprintf (ErrorTextPntr, ErrorTextSize,
    "RunCommand: Trouble setting up StdOut pipe for Command \"%s\".", Name ());

  if (m_PipeStdOutFromChild [0] != -1 || m_PipeStdOutFromChild [1] != -1)
    return B_NO_INIT; // Should have started at -1 set by the constructor.

  if (pipe (m_PipeStdOutFromChild) != 0) // Returns -1 if failed, 0 if okay.
    return errno; // Pipe creation failed.

  // Make the reading end [0] of StdOut non-blocking to the main program, so it
  // can continue running when there is no data available.  Writing end [1]
  // left unchanged so child will block when the pipe is full.  Hopefully the
  // program loop is running often enough to empty the pipe into the log file
  // and output text display, so that the child doesn't pause too often.

  int PipeFlags = fcntl (m_PipeStdOutFromChild [0], F_GETFL);
  PipeFlags |= O_NONBLOCK;
  fcntl (m_PipeStdOutFromChild [0], F_SETFL, PipeFlags);

  // Flush file buffers, before forking, so that delayed contents don't show up
  // in the child.

  fflush (stdout);
  fflush (stderr);

  // Fork into a second thread, which will later run the command.

  ChildThreadID = fork ();
  if (ChildThreadID == -1)
  {
    ErrorCode = errno;
    snprintf (ErrorTextPntr, ErrorTextSize,
      "RunCommand: Fork failed, can't start executing Command \"%s\".",
      Name ());
    close (m_PipeStdOutFromChild[0]);
    m_PipeStdOutFromChild [0] = -1;
    close (m_PipeStdOutFromChild[1]);
    m_PipeStdOutFromChild [1] = -1;
    return ErrorCode;
  }

  if (ChildThreadID == 0)
  {
    // We are the child thread.  To abort, call _exit(), which will close our
    // file descriptors, but not run atexit() functions (so C++ objects that
    // the child has a copied from the main program, like BApplication, won't
    // get destroyed, confusing the OS).

    // Only writing to the StdOut pipe, close the read end of it.

    close (m_PipeStdOutFromChild[0]);

    // Switch our StdOut and StdErr to write to the pipe instead.

    if (dup2 (m_PipeStdOutFromChild[1], STDOUT_FILENO) == -1)
    {
      perror ("RunCommand: dup2 for StdOut failed in child thread");
      _exit (EXIT_FAILURE);
    }
    if (dup2 (m_PipeStdOutFromChild[1], STDERR_FILENO) == -1)
    {
      perror ("RunCommand: dup2 for StdErr failed in child thread");
      _exit (EXIT_FAILURE);
    }
    close (m_PipeStdOutFromChild[1]);

    // Switch to a new process group (actual group number is copied from the
    // thread ID), so signals sent the the main program (like control-C
    // SIGTERM) don't also affect the child.  Also makes the new /bin/sh the
    // head of a process group, so if we kill the group, we also kill all the
    // commands fired off by the shell (otherwise we'd just kill the /bin/sh
    // interpreter program, not the programs invoked by the command line).

    setpgid (0, 0);

    // Run the command line, by passing it to /bin/sh to be executed.

    char * Arguments [4];
    Arguments [0] = "sh";
    Arguments [1] = "-c";
    Arguments [2] = const_cast<char *> (CommandLine.String ());
    Arguments [3] = NULL;

    execve ("/bin/sh", Arguments, environ); // Doesn't return if successful.
    perror ("RunCommand: Failed to start /bin/sh to run the command line");
    _exit (EXIT_FAILURE);
  }
  else // We are the parent thread.  Safe to access member variables etc.
  {
    m_ChildPID = ChildThreadID;

    // We shouldn't touch the write end of the StdOut pipe.  Close our copy of
    // the file descriptor (it gets duplicated as part of the fork ()
    // operation), leaving the child as the only one using it.

    close (m_PipeStdOutFromChild[1]);
    m_PipeStdOutFromChild[1] = -1;

    // Mark the read end of the StdOut pipe as close-on-exec, so that it only
    // exists in the main program (child tasks shouldn't access it).  Then
    // future forked children will automatically close their copy of it.

    fcntl (m_PipeStdOutFromChild [0], F_SETFD, FD_CLOEXEC);

    // Open the log file if requested.  Create log directory if needed.

    if (m_LogFileEnabled)
    {
      m_LogFile.Unset ();

      BString FileName;
      GetSanitisedName (FileName);

      BString LogPath (g_App->GetSettingsDirectory ());
      LogPath.Append ("/");
      LogPath.Append (g_LogDirName);
      BDirectory LogDirectory;
      ErrorCode = LogDirectory.SetTo (LogPath.String ());
      if (ErrorCode != B_OK)
      {
        create_directory (LogPath.String (), 0755);
        ErrorCode = LogDirectory.SetTo (LogPath.String ());
      }
      if (ErrorCode != B_OK)
      {
        m_LogAccumulated << "\nUnable to access log directory of \"" <<
          LogPath << "\", error number " << ErrorCode << " (" <<
          strerror (ErrorCode) << ").  Logging to file won't be done for "
          "this run.\n";
      }
      else
      {
        ErrorCode = m_LogFile.SetTo (&LogDirectory, FileName.String (),
          B_WRITE_ONLY | B_CREATE_FILE | B_OPEN_AT_END | O_CLOEXEC);
        if (ErrorCode != B_OK)
        {
          m_LogFile.Unset ();
          m_LogAccumulated << "\nUnable to write to log file \"" <<
            LogPath << "/" << FileName << "\", error number " <<
            ErrorCode << " (" << strerror (ErrorCode) <<
            ").  Logging to file won't be done for this run.\n";
        }
      }
    }
  }
  return B_OK;
}


void Command::SetState (CommandState NewState)
{
  m_CommandState = NewState;

  // Clear pending other Command has finished flag so that when changing from
  // editing or some other state to ready-to-run, it doesn't suddenly start
  // running.  It should only start up if the other Command finishes while this
  // Command is in the ready-to-run state.

  m_TriggerNamedCommandHasFinished = false;
}


status_t Command::StopCommand (char *ErrorTextPntr, int ErrorTextSize)
{
  snprintf (ErrorTextPntr, ErrorTextSize,
    "StopCommand: Unable to stop Command \"%s\".", Name ());
  if (m_ChildPID == -1)
    return B_NO_INIT;

  // Use negative thread ID to kill the child process group (the /bin/sh
  // process is at the head of it) rather than just the /bin/sh thread,
  // otherwise children of the shell (generally everything invoked by the
  // command line) won't get killed.

  if (kill (-m_ChildPID, SIGTERM) == -1)
    return errno; // Failed to kill it.

  return B_OK;
}



/******************************************************************************
 * Implementation of the Field class, constructor, destructor and the rest of
 * the member functions in mostly alphabetical order.
 */

Field::Field (const char *NewName) :
  BaseChangeScriptable (NewName),
  m_ParentCommand (NULL)
{
}


Field::Field (BMessage *ArchiveMessagePntr) :
  BaseChangeScriptable (ArchiveMessagePntr),
  m_ParentCommand (NULL)
{
  // Load our subclass specific data from the archived Field data.

  ArchiveMessagePntr->FindString (g_ValueName, &m_Value);
}


Field::~Field ()
{
}


/* Archives the Field object into a BMessage, for storage (usually in the
settings file). */

status_t Field::Archive (BMessage *ArchiveMessagePntr, bool Deep) const
{
  status_t ErrorCode;

  // Do the parent class archiving first.

  ErrorCode = BaseChangeScriptable::Archive (ArchiveMessagePntr, Deep);
  if (ErrorCode != B_OK)
    return ErrorCode;
/*
  ArchiveMessagePntr->AddString ("class", "Field");
    // BHandler's Archive function adds the class name of "Field" too (rather
    // than "BHandler"), so this isn't really needed, at least in BeOS R5.03.
*/
  // Add our subclass specific data to the archive.

  ArchiveMessagePntr->AddString (g_ValueName, m_Value);
  return ErrorCode;
}


/* This Field has been changed, increment the serial number and let the
parent object know about the change. */

void Field::ChangeHappened (bool SettingsAffected)
{
  BaseChangeScriptable::ChangeHappened (SettingsAffected);

  if (m_ParentCommand != NULL)
    m_ParentCommand->ChangeHappened (SettingsAffected);
}


/* Replaces all keywords related to this field in the given CommandLine string.
The result is in the modified CommandLine. */

void Field::ExpandMagicWords (BString &CommandLine)
{
  // Replace "FieldNameAsDir" with the Value up to and including the last slash
  // (forwards or backwards slash; / or \).  Empty string if no slash.

  BString MagicWord (Name ());
  MagicWord << "AsDir";

  BString ValueAsDir;
  int iLastSlash = m_Value.FindLast ('/');
  if (iLastSlash < 0)
    iLastSlash = m_Value.FindLast ('\\');
  if (iLastSlash >= 0)
    m_Value.CopyInto (ValueAsDir, 0, iLastSlash + 1);

  CommandLine.ReplaceAll (MagicWord.String (), ValueAsDir.String (), 0);

  // Replace "FieldNameAsName" with the Value after the last slash.

  MagicWord = Name ();
  MagicWord << "AsName";

  BString ValueAsName (m_Value);
  if (iLastSlash >= 0)
    m_Value.CopyInto (ValueAsName, iLastSlash + 1,
    m_Value.Length () - iLastSlash - 1);
  CommandLine.ReplaceAll (MagicWord.String (), ValueAsName.String (), 0);

  // Replace the plain Field name with the Field's value.

  CommandLine.ReplaceAll (Name (), m_Value.String (), 0);
}


/* Returns a list of the scripting Fields a Field understands. */

status_t Field::GetSupportedSuites (BMessage *MessagePntr)
{
  BPropertyInfo TempPropInfo (g_FieldScriptingPropertyList);
  MessagePntr->AddString ("suites", SUITE_BASE "field");
  MessagePntr->AddFlat ("messages", &TempPropInfo);
  return BaseChangeScriptable::GetSupportedSuites (MessagePntr);
}


/* Class global function for recreating objects from a BMessage archive.  Most
of the data loading work is done in the constructor, which takes an Archive
argument. */

BArchivable *Field::Instantiate (BMessage *ArchiveMessagePntr)
{
  if (validate_instantiation (ArchiveMessagePntr, "Field"))
    return new Field (ArchiveMessagePntr);
  return NULL;
}


/* Process mostly scripting messages. */

void Field::MessageReceived (BMessage *MessagePntr)
{
  const char           *ArgumentString = NULL;
  status_t              ErrorCode = B_ERROR;
  const char           *PropertyName;
  BMessage              ReplyMessage (B_REPLY);
  int32                 SpecifierIndex;
  int32                 SpecifierKind;
  BMessage              SpecifierMessage;
  char                  TempString [PATH_MAX + 1024];

  switch (MessagePntr->what)
  {
    case B_GET_PROPERTY:
    case B_SET_PROPERTY:
    case B_COUNT_PROPERTIES:
    case B_CREATE_PROPERTY:
    case B_DELETE_PROPERTY:
    case B_EXECUTE_PROPERTY:
      // Look at specifier[0] to find out what the scripting does.
      if (B_OK == MessagePntr->SetCurrentSpecifier (0))
        ErrorCode = MessagePntr->GetCurrentSpecifier (&SpecifierIndex,
        &SpecifierMessage, &SpecifierKind, &PropertyName);
      else // No specifiers, assume it is a direct operation on this Field.
      {
        PropertyName = g_PropertyNames[PN_FIELD];
        SpecifierKind = B_DIRECT_SPECIFIER;
        ErrorCode = B_OK;
      }

      if (ErrorCode == B_OK &&
      IsItOurScriptingCommand (g_FieldScriptingPropertyList,
      PropertyName, SpecifierKind))
        break; // It's one of ours, do it.
      // Else fall through, and punt the message to the parent class.

    default: // Not one of our messages, let the parent handle it.
      BaseChangeScriptable::MessageReceived (MessagePntr);
      return;
  }

  if (m_ParentCommand == NULL)
  {
    snprintf (TempString, sizeof (TempString),
      "Can't do anything to Field named \"%s\" because it isn't part of a "
      "Command.  How did you even send a scripting message to it?", Name ());
    ErrorCode = B_ERROR;
    goto ErrorExit;
  }

  snprintf (TempString, sizeof (TempString),
    "Scripting operation code %c%c%c%c (get, set, count, etc) or access type "
    "%d (direct=1, index=2, reverse index=3, range=4, reverse range=5, "
    "name=6, ID=7) is unsupported for property \"%s\" of the Field "
    "object named \"%s\" which is part of Command named \"%s\"",
    (char) (MessagePntr->what >> 24),
    (char) (MessagePntr->what >> 16),
    (char) (MessagePntr->what >>  8),
    (char) (MessagePntr->what >>  0),
    (int) SpecifierKind, PropertyName, Name (), m_ParentCommand->Name ());
  ErrorCode = B_BAD_SCRIPT_SYNTAX;

  // Handle the Name property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_NAME]) == 0)
  {
    // Get the current name.

    if (MessagePntr->what == B_GET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddString (g_ResultName, Name ());
      goto SuccessExit;
    }

    // Set the new Name.  This also means changing the Map of Fields inside the
    // parent Command since they are in alphabetical order.

    if (MessagePntr->what == B_SET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ErrorCode = MessagePntr->FindString (g_DataName, &ArgumentString);
      if (ErrorCode != B_OK)
      {
        snprintf (TempString, sizeof (TempString),
          "Need to specify a \"%s\" string value to set the Name property "
          "from.  Was for Field \"%s\" of Command \"%s\"",
          g_DataName, Name (), m_ParentCommand->Name ());
        goto ErrorExit;
      }

      if (!m_ParentCommand->IsEditing ())
      {
        snprintf (TempString, sizeof (TempString),
          "Can't change name of Field named \"%s\" to \"%s\" because the "
          "related Command named \"%s\" isn't in edit mode",
          Name (), ArgumentString, m_ParentCommand->Name ());
        ErrorCode = B_BUSY;
        goto ErrorExit;
      }

      if (m_ParentCommand->DoesNamedFieldExist (ArgumentString))
      {
        snprintf (TempString, sizeof (TempString),
          "Field named \"%s\" already exists in Command \"%s\", can't "
          "rename \"%s\" to it",
          ArgumentString, m_ParentCommand->Name (), Name ());
        ErrorCode = B_NAME_IN_USE;
        goto ErrorExit;
      }

      Command *ParentCommandPntr = m_ParentCommand;
        // Need a copy of the parent, since removing the field sets
        // m_ParentCommand to NULL, and adding it sets it to the Command it
        // is added to.

      ParentCommandPntr->InternalRemoveField (this);
      SetName (ArgumentString);
      ParentCommandPntr->InternalAddField (this);
      ChangeHappened ();
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on Name property.
  }

  // Handle the Serial property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_SERIAL]) == 0)
  {
    // Get the current serial number.

    if (MessagePntr->what == B_GET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddInt32 (g_ResultName, GetSerial ());
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on Serial property.
  }

  // Handle the Field property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_FIELD]) == 0)
  {
    if (MessagePntr->what == B_DELETE_PROPERTY /* any kind of specifier OK */)
    {
      if (!m_ParentCommand->IsEditing ())
      {
        snprintf (TempString, sizeof (TempString),
          "Can't delete Field named \"%s\" because the related "
          "Command named \"%s\" isn't in edit mode",
          Name (), m_ParentCommand->Name ());
        ErrorCode = B_BUSY;
        goto ErrorExit;
      }

      Command *ParentCommandPntr = m_ParentCommand; // Copy before NULLified
      ParentCommandPntr->InternalRemoveField (this); // by the removal.
      ParentCommandPntr->ChangeHappened ();
      delete this; // Careful, can't reference member variables after here.
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on Field property.
  }

  // Handle the Value property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_VALUE]) == 0)
  {
    // Get the current Value.

    if (MessagePntr->what == B_GET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddString (g_ResultName, m_Value);
      goto SuccessExit;
    }

    // Set the new Value.

    if (MessagePntr->what == B_SET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ErrorCode = MessagePntr->FindString (g_DataName, &ArgumentString);
      if (ErrorCode != B_OK)
      {
        snprintf (TempString, sizeof (TempString),
          "Need to specify a \"%s\" string value to set the Value property "
          "from.  Was for Field \"%s\" of Command \"%s\"",
          g_DataName, Name (), m_ParentCommand->Name ());
        goto ErrorExit;
      }

      if (!m_ParentCommand->IsEditing ())
      {
        snprintf (TempString, sizeof (TempString),
          "Can't change Value of Field named \"%s\" to \"%s\" because the "
          "related Command named \"%s\" isn't in edit mode",
          Name (), ArgumentString, m_ParentCommand->Name ());
        ErrorCode = B_BUSY;
        goto ErrorExit;
      }

      m_Value.SetTo (ArgumentString);
      ChangeHappened ();
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on Value property.
  }

ErrorExit: /* Error message in TempString, return code in ErrorCode. */
  ReplyMessage.AddInt32 (g_ErrorName, ErrorCode);
  ReplyMessage.AddString (g_MessageName, TempString);
  ErrorCode = MessagePntr->SendReply (&ReplyMessage,
    g_App /* Reply's reply handler */, 500000 /* send timeout */);
  return;

SuccessExit: /* Success. */
  ReplyMessage.AddInt32 (g_ErrorName, B_OK);
  ErrorCode = MessagePntr->SendReply (&ReplyMessage,
    g_App /* Reply's reply handler */, 500000 /* send timeout */);
}


/* A Field is looking for the object to execute a scripting command. */

BHandler * Field::ResolveSpecifier (
  BMessage *MessagePntr,
  int32 Index,
  BMessage *SpecifierMsgPntr,
  int32 SpecificationKind,
  const char *PropertyNamePntr)
{
#if 0
printf ("Resolving Specifier in Field object, Message is: ");
MessagePntr->PrintToStream();
printf ("Specifier index %d, Specifier kind %d, Property \"%s\"\n",
  (int) Index, (int) SpecificationKind, PropertyNamePntr);
#endif

  if (Index == 0 && IsItOurScriptingCommand (g_FieldScriptingPropertyList,
  PropertyNamePntr, SpecificationKind))
  {
    return this; /* It's one of our properties, self will handle message. */
  }

  /* Handle an unrecognized scripting Field, let the parent figure it out. */

  return BaseChangeScriptable::ResolveSpecifier (
    MessagePntr, Index, SpecifierMsgPntr, SpecificationKind, PropertyNamePntr);
}


void Field::SetParentCommand (class Command *CommandPntr)
{
  m_ParentCommand = CommandPntr;
}



/******************************************************************************
 * Implementation of the ScriptOCronApp class, constructor, destructor and the
 * rest of the member functions in mostly alphabetical order.
 */

ScriptOCronApp::ScriptOCronApp ()
: BApplication (g_AppSignature),
  m_ChangeSerial (GetNextSerialNumber ()),
  m_QuitRequestedStashedMessage (NULL),
  m_SerialWhenSettingsSaved (-1),
  m_TimeSettingsWereLastSaved (0)
{
  status_t    ErrorCode;

  /* Set up the pathname which identifies our settings directory.  Note that
  the actual settings are loaded later on (or set to defaults) by the main()
  function, before this BApplication starts running.  So we don't bother
  initialising the other setting related variables here. */

  ErrorCode =
    find_directory (B_USER_SETTINGS_DIRECTORY, &m_SettingsDirectoryPath);
  if (ErrorCode == B_OK)
    ErrorCode = m_SettingsDirectoryPath.Append (PROGRAM_NAME_WORD);
  if (ErrorCode != B_OK)
    m_SettingsDirectoryPath.SetTo (".");
}


ScriptOCronApp::~ScriptOCronApp ()
{
  // Save the settings if needed.  Also may save the Commands, so don't delete
  // them just yet!

  if (g_SettingsNeedSaving)
    LoadSaveSettings (false /* DoLoad */);

  DeleteAllCommands ();
  DeleteAllModels ();

  delete m_QuitRequestedStashedMessage;
  m_QuitRequestedStashedMessage = NULL;
}


/* Display a box showing information about this program. */

void ScriptOCronApp::AboutRequested ()
{
  BAlert *AboutAlertPntr;

  std::ostringstream TempStream;

  PrintUsage (TempStream, 0 /* WrapWidth */, false /* ListCommands */);
  TempStream.flush ();

  // Remove trailing linefeeds and spaces.

  BString AboutText (TempStream.str().c_str());
  while (isspace (AboutText.ByteAt (AboutText.Length () - 1)))
    AboutText.Truncate (AboutText.Length () - 1);

  AboutAlertPntr = new BAlert ("About", AboutText.String (), "Done");
  if (AboutAlertPntr != NULL)
  {
    AboutAlertPntr->SetShortcut (0, B_ESCAPE);

    // Do a synchronous alert so that the string still exists while the box is
    // up, otherwise it gets deleted and the BAlert sometimes crashes
    // (particularly in Haiku) while trying to reference it.  Though it does
    // mean all Command processing stops while the About is showing.

    AboutAlertPntr->Go ();
  }
}


/* The Application has been changed, increment the serial number.  There is no
parent to notify about the change so that's all that needs to be done.  Other
interested parties can use scripting to check the serial number for changes,
and then update themselves as needed. */

void ScriptOCronApp::ChangeHappened (bool SettingsAffected)
{
  m_ChangeSerial = GetNextSerialNumber ();

  if (SettingsAffected)
    g_SettingsNeedSaving = true;
}


/* Set the settings to the defaults.  Needed in case there isn't a settings
file or it is obsolete. */

void ScriptOCronApp::DefaultSettings ()
{
#if INCLUDE_GUI_CODE
  g_AdvancedMode = false; // Basic the first time.
  g_InitialWindowFrame = BRect (1, 2, 3, 4); // Meaning not initialised.
  g_ServerMode = false; // Shows the GUI window on the first run.
#endif

  // Load built-in Models from a resource named "Models", which contains a
  // flattened BMessage made from drag and dropping an AGMSScriptOCron
  // settings file into QuickRes, with a "deres" side trip to a text file
  // resource to give it a name which sticks (seems to be a bug in QuickRes).

  BMessage    ModelArchive;
  const char *ResourceData;
  size_t      ResourceSize;
  BResources *ResourcesPntr;

  if ((ResourcesPntr = AppResources ()) != NULL &&
  (ResourceData = (const char *) ResourcesPntr->LoadResource ('MSGG',
  "Models", &ResourceSize)) != NULL)
  {
    if (B_OK == ModelArchive.Unflatten (ResourceData))
      LoadModels (&ModelArchive, true /* BuiltIn */);
  }

  m_SerialWhenSettingsSaved = m_ChangeSerial;
}


/* Delete all the commands, taking the time to deallocate their memory too. */

void ScriptOCronApp::DeleteAllCommands ()
{
  CommandMap::iterator CommandIter;

  for (CommandIter = m_CommandMap.begin ();
  CommandIter != m_CommandMap.end (); CommandIter++)
  {
    // Just need to delete the value (Command object).  The key is the name of
    // the command as a const char *, which is a pointer into the Command
    // object's data (actually the BHandler's name field) so it implicitly gets
    // deleted when the Command is deleted.

    Command *CommandToBeDeletedPntr = CommandIter->second;
    delete CommandToBeDeletedPntr; // Will also remove it from Looper.
  }
  m_CommandMap.clear ();
}


void ScriptOCronApp::DeleteAllModels ()
{
  ModelMap::iterator ModelIter;

  for (ModelIter = m_ModelMap.begin ();
  ModelIter != m_ModelMap.end (); ModelIter++)
  {
    // Just need to delete the value (Model object).  The key is the name of
    // the Model as a const char *, which is a pointer into the Model's
    // BMessage data so it implicitly gets deleted when the Model is deleted.

    BMessage *ModelToBeDeletedPntr = ModelIter->second;
    delete ModelToBeDeletedPntr;
  }
  m_ModelMap.clear ();
}


/* Intercept B_QUIT_REQUESTED.  Otherwise this BLooper/BApplication may be
killed and a reply message automatically sent.  We do both of those ourselves
in an effort to delay death. */

void ScriptOCronApp::DispatchMessage (BMessage *MessagePntr, BHandler *Target)
{
  if (MessagePntr->what != B_QUIT_REQUESTED)
  {
    BApplication::DispatchMessage (MessagePntr, Target);
    return;
  }

  if (m_QuitRequestedStashedMessage != NULL)
  {
    // Reject an unexpected second request to quit.  Can happen if someone
    // closes the window while the operating system is shutting down.

    BMessage ReplyMessage (B_REPLY);
    ReplyMessage.AddBool (g_ResultName, false);
    int32 ThreadID = Thread (); // Haiku needs the thread ID too.
    ReplyMessage.AddInt32 ("thread", ThreadID);
    MessagePntr->SendReply (&ReplyMessage,
      (BHandler *) NULL /* where to send the reply to the reply */,
      500000 /* send timeout */);
    return;
  }

  m_QuitRequestedStashedMessage = DetachCurrentMessage ();

  if (g_QuitCountdown < 0)
  {
    g_QuitCountdown = 5; /* Start the countdown. */
    cerr << PROGRAM_NAME_WORD " shutdown sequence started.\n";

    // Ask all Commands to stop running.  Need the App message loop to
    // keep on running a bit longer so that the zombie processes can be
    // reaped, logs closed, etc.

    CommandMap::iterator CommandIter;
    for (CommandIter = m_CommandMap.begin ();
    CommandIter != m_CommandMap.end (); CommandIter++)
    {
      Command *CommandPntr = CommandIter->second;
      if (CommandPntr->IsRunning ())
      {
        cerr << "Stopping running Command \"" <<
          CommandPntr->Name () << "\".\n";

        BMessage StopScriptingMessage (B_DELETE_PROPERTY);
        StopScriptingMessage.AddSpecifier (g_PropertyNames[PN_RUN]);
        StopScriptingMessage.AddSpecifier (
          g_PropertyNames[PN_COMMAND], CommandPntr->Name ());
        PostMessage (&StopScriptingMessage);
      }
    }
  }
}


/* Like the function name says... */

bool ScriptOCronApp::DoesNamedCommandExist (const char *CommandNamePntr)
{
  CommandMap::iterator CommandIter;
  CommandIter = m_CommandMap.find (CommandNamePntr);
  return CommandIter != m_CommandMap.end ();
}


/* Called by a Command when it finishes running, to let other Commands which
may be waiting for it get ready to start running. */

void ScriptOCronApp::DoNotificationsForFinishedCommand (
  const char *CommandNamePntr)
{
  CommandMap::iterator CommandIter;

  for (CommandIter = m_CommandMap.begin ();
  CommandIter != m_CommandMap.end (); CommandIter++)
  {
    Command *CommandPntr = CommandIter->second;
    CommandPntr->OtherCommandFinished (CommandNamePntr);
  }
}


/* Tell other programs about the top Application level scripting commands we
support.  Try this command: "hey application/x-vnd.agmsmith.scriptocron
getsuites" to see it in action (this program has to be already running for it
to work).  Scripting commands for subobjects are handled in those objects. */

status_t ScriptOCronApp::GetSupportedSuites (BMessage *MessagePntr)
{
  BPropertyInfo TempPropInfo (g_AppScriptingPropertyList);
  MessagePntr->AddString ("suites", SUITE_BASE "application");
  MessagePntr->AddFlat ("messages", &TempPropInfo);
  return BApplication::GetSupportedSuites (MessagePntr);
}


/* Load the Models from an archive stored in a BMessage, which could be from
the settings file or from the built-in Models resource.  Skips ones which have
a duplicate name. */

void ScriptOCronApp::LoadModels (BMessage *ArchivePntr, bool BuiltIn)
{
  type_code TypeFound = 0;
  int32 CountFound = 0;
  if (B_OK == ArchivePntr->GetInfo(g_ModelListName, &TypeFound, &CountFound) &&
  TypeFound == B_MESSAGE_TYPE && CountFound > 0)
  {
    int32 iModel;
    for (iModel = 0; iModel < CountFound; iModel++)
    {
      // Load the next candidate Model.

      BMessage *ModelMessagePntr = new BMessage;
      if (B_OK != ArchivePntr->FindMessage (
      g_ModelListName, iModel, ModelMessagePntr))
      {
        delete ModelMessagePntr;
        continue; // Skip bad ones.
      }

      // What is the name of the Model?

      const char *ModelNamePntr = NULL;
      if (B_OK != ModelMessagePntr->FindString (g_InternalName,
      &ModelNamePntr))
      {
        delete ModelMessagePntr;
        continue; // It doesn't have a name, probably not a Model.
      }

      // Make sure a Model with the same name does not already exist.  If
      // another does exist, leave it alone, since it may be one of the
      // built-in Models, and we don't want the user to override them
      // (leads to support confusion).

      ModelMap::iterator ModelIter = m_ModelMap.find (ModelNamePntr);
      if (ModelIter != m_ModelMap.end ())
      {
        delete ModelMessagePntr;
        continue;
      }

      // Mark built-in ones with a flag so they don't get saved.  Flag only
      // exists in memory, will never be in a saved file (or leak into a
      // resource), so we don't have to worry about adding it too many times.

      if (BuiltIn)
        ModelMessagePntr->AddBool (g_NameBuiltIn, true);

      // Reload the name string.  Modifying the message by adding the bool for
      // BuiltIn may have changed its address (happens in Haiku, not BeOS).

      ModelMessagePntr->FindString (g_InternalName, &ModelNamePntr);

      // Have a valid enough Model, add it to the collection.

      m_ModelMap [ModelNamePntr] = ModelMessagePntr;
    }
  }
}


/* Either load the settings (DoLoad is TRUE) from the configuration file or
write them (DoLoad is FALSE) to it.  The configuration file is a flattened
BMessage containing the various program settings.  If it doesn't exist (and its
parent directories don't exist) then it will be created when saving.  If it
doesn't exist when loading, the settings will be set to default values. */

status_t ScriptOCronApp::LoadSaveSettings (bool DoLoad)
{
  status_t    ErrorCode;
  const char *NamePntr;
  BMessage    Settings;
  BDirectory  SettingsDirectory;
  BFile       SettingsFile;
  char        TempString [PATH_MAX + 1024];

  // Preset things to default values if loading, in case of an error or it's an
  // older version of the settings file which doesn't have every field defined.

  if (DoLoad)
  {
    DeleteAllModels ();
    DeleteAllCommands ();
    DefaultSettings ();
  }

  // Look for our settings directory.  When saving we can try to create it,
  // though it should be there since we're using the user's settings directory
  // directly now (no point in a subdirectory; we've only got one file).

  ErrorCode = SettingsDirectory.SetTo (m_SettingsDirectoryPath.Path ());
  if (ErrorCode != B_OK)
  {
    ErrorCode = create_directory (m_SettingsDirectoryPath.Path (), 0755);
    if (ErrorCode == B_OK)
      ErrorCode = SettingsDirectory.SetTo (m_SettingsDirectoryPath.Path ());
    if (ErrorCode != B_OK)
    {
      sprintf (TempString, "Can't create settings directory \"%s\"",
        m_SettingsDirectoryPath.Path ());
      goto ErrorExit;
    }
  }

  ErrorCode = SettingsFile.SetTo (&SettingsDirectory, g_SettingsFileName,
    DoLoad ? B_READ_ONLY : B_READ_WRITE | B_CREATE_FILE | B_ERASE_FILE);
  if (ErrorCode != B_OK)
  {
    sprintf (TempString,
      "Can't open settings file \"%s\" in directory \"%s\" for %s",
      g_SettingsFileName, m_SettingsDirectoryPath.Path(),
      DoLoad ? "reading" : "writing");

    if (ErrorCode == B_ENTRY_NOT_FOUND)
    { // No error message for expected error that happens on the first run.
      g_FirstTimeAppRun = true;
      return ErrorCode;
    }
    goto ErrorExit;
  }

  if (DoLoad)
  {
    ErrorCode = Settings.Unflatten (&SettingsFile);
    if (ErrorCode != 0 || Settings.what != g_SettingsWhatCode)
    {
      sprintf (TempString, "Corrupt data detected while reading settings "
        "file \"%s\" in directory \"%s\", will revert to defaults",
        g_SettingsFileName, m_SettingsDirectoryPath.Path());
      goto ErrorExit;
    }
  }

  // Transfer the settings between the BMessage and our various global
  // variables.  For loading, if the setting isn't present, leave it at the
  // default value.  Note that loading and saving are intermingled here to make
  // code maintenance easier (less chance of forgetting to update it if load
  // and save were separate functions).

  ErrorCode = B_OK; /* So that saving settings can record an error. */

#if INCLUDE_GUI_CODE
  NamePntr = "AdvancedMode";
  if (DoLoad)
  {
    bool TempBool;
    if (Settings.FindBool (NamePntr, &TempBool) == B_OK)
      g_AdvancedMode = TempBool;
  }
  else if (ErrorCode == B_OK)
    ErrorCode = Settings.AddBool (NamePntr, g_AdvancedMode);

  NamePntr = "InitialWindowFrame";
  if (DoLoad)
  {
    BRect TempRect;
    if (Settings.FindRect (NamePntr, &TempRect) == B_OK)
      g_InitialWindowFrame = TempRect;
  }
  else if (ErrorCode == B_OK)
    ErrorCode = Settings.AddRect (NamePntr, g_InitialWindowFrame);

  NamePntr = "ServerMode";
  if (DoLoad)
  {
    bool TempBool;
    if (Settings.FindBool (NamePntr, &TempBool) == B_OK)
      g_ServerMode = TempBool;
  }
  else if (ErrorCode == B_OK)
    ErrorCode = Settings.AddBool (NamePntr, g_ServerMode);
#endif // INCLUDE_GUI_CODE

  // Load/save the Commands, which are stored as archived BMessages.

  NamePntr = g_CommandListName;
  if (DoLoad)
  {
    type_code TypeFound = 0;
    int32 CountFound = 0;
    ErrorCode = Settings.GetInfo (NamePntr, &TypeFound, &CountFound);
    if (ErrorCode == B_OK && TypeFound == B_MESSAGE_TYPE && CountFound > 0)
    {
      int32 iCommand;
      for (iCommand = 0; iCommand < CountFound; iCommand++)
      {
        BMessage ArchiveMsg;
        ErrorCode = Settings.FindMessage (NamePntr, iCommand, &ArchiveMsg);
        if (ErrorCode != B_OK)
          continue; // Skip bad ones.
        BArchivable *DearchivedPntr = instantiate_object (&ArchiveMsg);
        if (DearchivedPntr == NULL)
          continue; // Unable to find C++ class of object to rebuild it, etc.

        Command *CommandPntr = cast_as (DearchivedPntr, Command);
        if (CommandPntr != NULL)
        {
          CommandPntr->SetState (CS_READY_TO_RUN); // Default is Editing.
          InternalAddCommand (CommandPntr);
        }
        else // Some other kind of object, someone's been hacking the settings?
          delete DearchivedPntr;
      }
    }
    ChangeHappened (false /* Settings file doesn't need saving */);
  }
  else if (ErrorCode == B_OK) // Saving the Command list.
  {
    CommandMap::iterator CommandIter;
    for (CommandIter = m_CommandMap.begin ();
    CommandIter != m_CommandMap.end (); CommandIter++)
    {
      BMessage ArchiveMsg (B_ARCHIVED_OBJECT);

      Command *CommandPntr = CommandIter->second;
      ErrorCode = CommandPntr->Archive (&ArchiveMsg, true /* deep */);
      if (ErrorCode != B_OK)
        break;
      ErrorCode = Settings.AddMessage (NamePntr, &ArchiveMsg);
      if (ErrorCode != B_OK)
        break;
    }
  }

  // Load/save the user created Models (don't save the built-in ones), also
  // stored as archived objects in a BMessage.

  NamePntr = g_ModelListName;
  if (DoLoad)
    LoadModels (&Settings, false /* BuiltIn */);
  else if (ErrorCode == B_OK) // Saving the Model list.
  {
    ModelMap::iterator ModelIter;
    for (ModelIter = m_ModelMap.begin ();
    ModelIter != m_ModelMap.end (); ModelIter++)
    {
      BMessage *ModelPntr = ModelIter->second;
      bool IsBuiltIn = false;
      if (B_OK != ModelPntr->FindBool (g_NameBuiltIn, &IsBuiltIn))
        IsBuiltIn = false;
      if (!IsBuiltIn)
      {
        ErrorCode = Settings.AddMessage (NamePntr, ModelPntr);
        if (ErrorCode != B_OK)
          break;
      }
    }
  }

  // Loading/saving from/to the temporary BMessage is complete.  Did it work?

  if (ErrorCode != B_OK && !DoLoad)
  {
    strcpy (TempString, "Unable to stuff the program settings into a "
      "temporary BMessage, settings not saved");
    goto ErrorExit;
  }

  // Save the settings BMessage to the settings file.

  if (!DoLoad)
  {
    Settings.what = g_SettingsWhatCode;
    ErrorCode = Settings.Flatten (&SettingsFile);
    if (ErrorCode != 0)
    {
      sprintf (TempString, "Problems while writing settings file \"%s\" in "
        "directory \"%s\"", g_SettingsFileName,
        m_SettingsDirectoryPath.Path ());
      goto ErrorExit;
    }
  }

  m_SerialWhenSettingsSaved = m_ChangeSerial;
  return B_OK;

ErrorExit: /* Error message in TempString, code in ErrorCode. */
  DisplayErrorMessage (TempString, ErrorCode, DoLoad ?
    "Loading Settings Error" : "Saving Settings Error");
  return ErrorCode;
}


const char *ScriptOCronApp::GetSettingsDirectory ()
{
  return m_SettingsDirectoryPath.Path ();
}


/* Add the given Command to the Map of all Commands.  Used by the Command
object when it is renaming itself.  Replaces any existing Command with the same
name.  Updates Loopers too. */

void ScriptOCronApp::InternalAddCommand (Command *CommandPntr)
{
  CommandMap::iterator CommandIter;
  CommandIter = m_CommandMap.find (CommandPntr->Name ());
  if (CommandIter != m_CommandMap.end ())
  {
    Command *OldCommandPntr = CommandIter->second;

    if (OldCommandPntr == CommandPntr)
      return; // It's already added to our Map of Commands.  Do nothing.

    m_CommandMap.erase (CommandIter);
    delete OldCommandPntr; // Will deLooper it as part of destructor.
  }

  m_CommandMap[CommandPntr->Name ()] = CommandPntr;
  CommandPntr->AddToLooper (this);
}


/* Remove the given Command from the Map of all Commands (but don't delete it).
Used by the Command object when it is renaming itself.  Updates Looper too. */

void ScriptOCronApp::InternalRemoveCommand (Command *CommandPntr)
{
  CommandMap::iterator CommandIter;
  CommandIter = m_CommandMap.find (CommandPntr->Name ());
  if (CommandIter != m_CommandMap.end ())
  {
    Command *OldCommandPntr = CommandIter->second;
    if (OldCommandPntr == CommandPntr)
    {
      m_CommandMap.erase (CommandIter);
      OldCommandPntr->RemoveFromLooper ();
    }
  }
}


/* Process a message directed to the Application level. */

void ScriptOCronApp::MessageReceived (BMessage *MessagePntr)
{
  const char           *ArgumentString = NULL;
  status_t              ErrorCode = B_BAD_SCRIPT_SYNTAX;
  const char           *PropertyName;
  BMessage              ReplyMessage (B_REPLY);
  int32                 SpecifierIndex;
  int32                 SpecifierKind;
  BMessage              SpecifierMessage;
  char                  TempString [PATH_MAX + 1024];

  /* See if it is a scripting message that applies to one of the operations
  this program supports.  Pass on other kinds of messages to the inherited
  parent MessageReceived () function. */

  switch (MessagePntr->what)
  {
    case B_GET_PROPERTY:
    case B_SET_PROPERTY:
    case B_COUNT_PROPERTIES:
    case B_CREATE_PROPERTY:
    case B_DELETE_PROPERTY:
    case B_EXECUTE_PROPERTY:
      // Look at specifier[0] to find out what the scripting does.
      if (B_OK == MessagePntr->SetCurrentSpecifier (0))
        ErrorCode = MessagePntr->GetCurrentSpecifier (&SpecifierIndex,
        &SpecifierMessage, &SpecifierKind, &PropertyName);
      else // No specifiers, no default target for our BApplication.
        ErrorCode = B_ERROR;

      if (ErrorCode == B_OK &&
      IsItOurScriptingCommand (g_AppScriptingPropertyList,
      PropertyName, SpecifierKind))
        break; // It's one of ours, do it.
      // Else fall through, and punt the message to the parent class.

    default: // Not one of our messages, let the parent handle it.
      BApplication::MessageReceived (MessagePntr);
      return;
  }

  snprintf (TempString, sizeof (TempString),
    "Scripting operation code %c%c%c%c (get, set, count, etc) or access type "
    "%d (direct=1, index=2, reverse index=3, range=4, reverse range=5, "
    "name=6, ID=7) is unsupported for property \"%s\" of the "
    PROGRAM_NAME_WORD " Application level object",
    (char) (MessagePntr->what >> 24),
    (char) (MessagePntr->what >> 16),
    (char) (MessagePntr->what >>  8),
    (char) (MessagePntr->what >>  0),
    (int) SpecifierKind, PropertyName);
  ErrorCode = B_BAD_SCRIPT_SYNTAX;

  // Handle the Serial property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_SERIAL]) == 0)
  {
    // Get the current serial number.

    if (MessagePntr->what == B_GET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddInt32 (g_ResultName, m_ChangeSerial);
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on Serial property.
  }

  // Handle the Command property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_COMMAND]) == 0)
  {
    // Count the number of Commands.

    if (MessagePntr->what == B_COUNT_PROPERTIES &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ReplyMessage.AddInt32 (g_ResultName, m_CommandMap.size ());
      goto SuccessExit;
    }

    // Create a new command, optionally with a name and using a Model.

    if (MessagePntr->what == B_CREATE_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ErrorCode = MessagePntr->FindString (g_NameName, &ArgumentString);
      if (ErrorCode != B_OK)
        ArgumentString = g_NameUnnamed;

      if (DoesNamedCommandExist (ArgumentString))
      {
        ErrorCode = B_NAME_IN_USE;
        snprintf (TempString, sizeof (TempString),
          "Can't create a Command named \"%s\" because it already exists",
          ArgumentString);
        goto ErrorExit;
      }

      const char *ModelNamePntr = NULL;
      ErrorCode = MessagePntr->FindString (g_ModelName, &ModelNamePntr);
      if (ErrorCode != B_OK)
        ModelNamePntr = NULL;

      Command *NewCommandPntr = NULL;
      if (ModelNamePntr == NULL)
        NewCommandPntr = new Command (ArgumentString);
      else // Make the new Command from a saved Model.
      {
        ModelMap::iterator ModelIter = m_ModelMap.find (ModelNamePntr);
        if (ModelIter == m_ModelMap.end ())
        {
          ErrorCode = B_NAME_NOT_FOUND;
          snprintf (TempString, sizeof (TempString),
            "Can't create a Command named \"%s\" because the Model "
            "named \"%s\" doesn't exist", ArgumentString, ModelNamePntr);
          goto ErrorExit;
        }

        BArchivable *DearchivedPntr = instantiate_object (ModelIter->second);
        if (DearchivedPntr == NULL ||
        (NewCommandPntr = cast_as (DearchivedPntr, Command)) == NULL)
        {
          delete DearchivedPntr; // NULL or wrong kind of object.
          snprintf (TempString, sizeof (TempString), "Something went wrong "
            "when instantiating a Command named \"%s\" from a Model named "
            "\"%s\"", ArgumentString, ModelNamePntr);
          ErrorCode = B_ERROR;
          goto ErrorExit;
        }

        // Save the cannonical name of the Model (the user's model may have
        // letters in a different case) in the UserData/OriginalModel key, so
        // that we can tell if the user has changed the Template so we don't
        // overwrite it while upgrading to a newer Template from a newer Model.

        NewCommandPntr->UserDataAddString (g_KeyOriginalModel,
          NewCommandPntr->Name ());

        NewCommandPntr->SetName (ArgumentString);
      }

      InternalAddCommand (NewCommandPntr);
      ChangeHappened ();
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on a Command property.
  }

  // Handle the Model property.

  if (strcasecmp (PropertyName, g_PropertyNames[PN_MODEL]) == 0)
  {
    // Get a list of all Model names, returned as an array of strings.

    if (MessagePntr->what == B_GET_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ModelMap::iterator ModelIter;
      for (ModelIter = m_ModelMap.begin ();
      ModelIter != m_ModelMap.end (); ModelIter++)
      {
        const char *ModelName = ModelIter->first;
        ReplyMessage.AddString (g_ResultName, ModelName);
      }
      goto SuccessExit;
    }

    // Create a new Model from a named Command.

    if (MessagePntr->what == B_CREATE_PROPERTY &&
    SpecifierKind == B_DIRECT_SPECIFIER)
    {
      ArgumentString = NULL;
      ErrorCode = MessagePntr->FindString (g_NameName, &ArgumentString);
      if (ErrorCode != B_OK || ArgumentString == NULL ||
      ArgumentString[0] == 0)
      {
        snprintf (TempString, sizeof (TempString),
        "Need an additional \"%s\" string to specify the name of the "
        "Command to use for the new Model", g_NameName);
        goto ErrorExit;
      }

      CommandMap::iterator CommandIter;
      CommandIter = m_CommandMap.find (ArgumentString);
      if (CommandIter == m_CommandMap.end ())
      {
        ErrorCode = B_NAME_NOT_FOUND;
        snprintf (TempString, sizeof (TempString), "Unable to create a Model "
          "for Command named \"%s\" because no Command with that name exists",
          ArgumentString);
        goto ErrorExit;
      }

      ModelMap::iterator ModelIter;
      ModelIter = m_ModelMap.find (ArgumentString);
      if (ModelIter != m_ModelMap.end ())
      {
        ErrorCode = B_NAME_IN_USE;
        snprintf (TempString, sizeof (TempString), "Unable to create a Model "
          "for Command named \"%s\" because a Model with that name already "
          "exists.  Delete it first if you want to replace it",
          ArgumentString);
        goto ErrorExit;
      }

      Command *CommandPntr = CommandIter->second;
      BMessage *ModelMessagePntr = new BMessage (B_ARCHIVED_OBJECT);
      ErrorCode = CommandPntr->Archive (ModelMessagePntr, true /* deep */);
      if (ErrorCode != B_OK)
      {
        snprintf (TempString, sizeof (TempString),
          "Failed to make an Archive copy of Command \"%s\"", ArgumentString);
        delete ModelMessagePntr;
        goto ErrorExit;
      }

      const char *PersistentNameString;
      ErrorCode = ModelMessagePntr->FindString (g_InternalName,
        &PersistentNameString);
      if (ErrorCode != B_OK ||
      strcmp (PersistentNameString, ArgumentString) != 0)
      {
        snprintf (TempString, sizeof (TempString),
          "Something wrong with internal name in Archived copy of "
          "Command \"%s\".  Doesn't exist or doesn't match Command name",
          ArgumentString);
        delete ModelMessagePntr;
        goto ErrorExit;
      }

      m_ModelMap [PersistentNameString] = ModelMessagePntr;

      ChangeHappened ();
      goto SuccessExit;
    }

    // Delete a Model, specified by a name.

    if (MessagePntr->what == B_DELETE_PROPERTY &&
    SpecifierKind == B_NAME_SPECIFIER)
    {
      ArgumentString = NULL;
      ErrorCode = SpecifierMessage.FindString (g_NameName, &ArgumentString);
      if (ErrorCode != B_OK || ArgumentString == NULL ||
      ArgumentString[0] == 0)
      {
        snprintf (TempString, sizeof (TempString),
          "Need to specify the name of the Model to be deleted");
        goto ErrorExit;
      }

      ModelMap::iterator ModelIter;
      ModelIter = m_ModelMap.find (ArgumentString);
      if (ModelIter == m_ModelMap.end ())
      {
        ErrorCode = B_NAME_NOT_FOUND;
        snprintf (TempString, sizeof (TempString),
          "Unable to delete Model named \"%s\" because it doesn't exist",
          ArgumentString);
        goto ErrorExit;
      }

      BMessage *DeletedModelPntr = ModelIter->second;
      m_ModelMap.erase (ModelIter);
      delete DeletedModelPntr;

      ChangeHappened ();
      goto SuccessExit;
    }

    goto ErrorExit; // Unknown operation on a Model property.
  }

ErrorExit: /* Error message in TempString, return code in ErrorCode. */
  ReplyMessage.AddInt32 (g_ErrorName, ErrorCode);
  ReplyMessage.AddString (g_MessageName, TempString);
  ErrorCode = MessagePntr->SendReply (&ReplyMessage,
    g_App /* Reply's reply handler */, 500000 /* send timeout */);
  return;

SuccessExit: /* Success. */
  ReplyMessage.AddInt32 (g_ErrorName, B_OK);
  ErrorCode = MessagePntr->SendReply (&ReplyMessage,
    g_App /* Reply's reply handler */, 500000 /* send timeout */);
}


/* Do periodic polling, mostly checking for output from the running Commands.
Since quitting stops the program before the results of a script command are
received, we use a time delay to do the quit.  Also, we have a quit countdown
which can be interrupted by an incoming scripting message in case one client
tells us to quit while another one is still using us. */

void ScriptOCronApp::Pulse ()
{
  if (g_QuitCountdown == 0)
  {
    cerr << PROGRAM_NAME_WORD " quitting time.\n";

    // There's a "_shutdown_" boolean field set to TRUE in the original
    // B_QUIT_REQUESTED message if the system is shutting down, it's in BeOS
    // too.

    bool OSIsShuttingDown = false;
    if (m_QuitRequestedStashedMessage != NULL &&
    m_QuitRequestedStashedMessage->FindBool ("_shutdown_", &OSIsShuttingDown)
    != B_OK)
      OSIsShuttingDown = false;

    // Quitting time.  First try closing windows, and ask each window if it's
    // okay to quit.  QuitRequested returns TRUE if all windows closed.  Skip
    // if shutting down, windows will be closed in app destructor instead.

    bool QuitOK = true;
    if (!OSIsShuttingDown)
      QuitOK = BApplication::QuitRequested ();

    // Reply to a delayed quit request (usually from the system when it is
    // trying to shut down).

    if (m_QuitRequestedStashedMessage != NULL)
    {
      BMessage ReplyMessage (B_REPLY);
      ReplyMessage.AddBool (g_ResultName, QuitOK);
      // Haiku's BApplication quit message handling code returns a "thread"
      // field with the BApplication's thread identifier, presumably the
      // BRoster requesting the shutdown needs it to figure out which pending
      // shutdown has failed or succeeded.
      int32 ThreadID = Thread ();
      ReplyMessage.AddInt32 ("thread", ThreadID);
      m_QuitRequestedStashedMessage->SendReply (&ReplyMessage,
        (BHandler *) NULL /* where to send the reply to the reply */,
        500000 /* send timeout */);

      delete m_QuitRequestedStashedMessage;
      m_QuitRequestedStashedMessage = NULL;
    }

    if (QuitOK)
      Quit (); // Shut down our message processing loop.
    else // Go back to normal running.
      g_QuitCountdown = -1;
  }
  else if (g_QuitCountdown > 0)
  {
    cerr << PROGRAM_NAME_WORD " quitting in " << g_QuitCountdown << ".\n";
    g_QuitCountdown--;
  }

  // Poll Commands.  Collects and stores output of running commands and checks
  // for commands that have finished and ones which should start running.  We
  // can assume no Command will be deleted during this polling (any delete
  // requests would have to go through the message system, which won't do
  // anything until this function has returned).

  time_t CurrentTime;
  time (&CurrentTime);
  struct tm CurrentBrokenUpTime;
  localtime_r (&CurrentTime, &CurrentBrokenUpTime);

  CommandMap::iterator CommandIter;
  for (CommandIter = m_CommandMap.begin ();
  CommandIter != m_CommandMap.end (); CommandIter++)
  {
    Command *CommandPntr = CommandIter->second;
    CommandPntr->PollAndUpdate (&CurrentBrokenUpTime);
  }

  // Save settings once in a while.  Don't want to save it on every change
  // while the user is editing things.

  if (g_SettingsNeedSaving && system_time () - m_TimeSettingsWereLastSaved >
  60 * 1000000 /* microsecond units */)
  {
    LoadSaveSettings (false /* DoLoad */);
    g_SettingsNeedSaving = false;
    m_TimeSettingsWereLastSaved = system_time ();
  }
}


void ScriptOCronApp::ReadyToRun ()
{
#if INCLUDE_GUI_CODE
  SetSizesAndOpenWindow ();
#endif

  SetPulseRate (200000); // Do after GUI windows opened, to get our rate.
}


/* A scripting command is looking for some object to execute it. */

BHandler * ScriptOCronApp::ResolveSpecifier (
  BMessage *MessagePntr,
  int32 Index,
  BMessage *SpecifierMsgPntr,
  int32 SpecificationKind,
  const char *PropertyNamePntr)
{
  status_t ErrorCode = B_OK;
  char TempString [PATH_MAX + 1024];

#if 0
printf ("Resolving Specifier in ScriptOCronApp, Message is: ");
MessagePntr->PrintToStream();
printf ("Specifier index %d, Specifier kind %d, Property \"%s\"\n",
  (int) Index, (int) SpecificationKind, PropertyNamePntr);
#endif

  /* See if it is one of our sub-object selection properties. */

  if (strcasecmp (PropertyNamePntr, g_PropertyNames[PN_COMMAND]) == 0)
  {
    // Looking for a named Command?

    if (SpecificationKind == B_NAME_SPECIFIER)
    {
      const char *NamePntr = NULL;
      ErrorCode = SpecifierMsgPntr->FindString (g_NameName, &NamePntr);
      if (ErrorCode != B_OK)
      {
        ErrorCode = B_BAD_SCRIPT_SYNTAX;
        snprintf (TempString, sizeof (TempString),
          "\"%s\" string field is missing in the scripting specifier for "
          "the \"%s\" property", g_NameName, PropertyNamePntr);
        goto ErrorExit;
      }

      CommandMap::iterator CommandIter;
      CommandIter = m_CommandMap.find (NamePntr);
      if (CommandIter != m_CommandMap.end ())
      {
        // Found the named Command, pass on the rest of the scripting command
        // to the Command object.

        MessagePntr->PopSpecifier ();
        return CommandIter->second;
      }
      ErrorCode = B_NAME_NOT_FOUND;
      snprintf (TempString, sizeof (TempString),
        "Property \"%s\" sub-object specified by name \"%s\" not found",
        PropertyNamePntr, NamePntr);
      goto ErrorExit;
    }
    else if (SpecificationKind == B_INDEX_SPECIFIER)
    {
      // Look for the numbered (indexed) Command, the slow way.

      int32 DesiredIndex = 0;
      ErrorCode = SpecifierMsgPntr->FindInt32 (g_NameIndex, &DesiredIndex);
      if (ErrorCode != B_OK)
      {
        ErrorCode = B_BAD_SCRIPT_SYNTAX;
        snprintf (TempString, sizeof (TempString), "\"%s\" field is "
          "missing in the scripting specifier for the \"%s\" property",
          g_NameIndex, PropertyNamePntr);
        goto ErrorExit;
      }

      if (DesiredIndex < 0 || DesiredIndex >= (int) m_CommandMap.size ())
      {
        ErrorCode = B_BAD_INDEX;
        snprintf (TempString, sizeof (TempString), "\"%s\" field for "
          "property \"%s\" is %d, which is not between 0 and %d",
          g_NameIndex, PropertyNamePntr, (int) DesiredIndex,
          (int) m_CommandMap.size() - 1);
        goto ErrorExit;
      }

      CommandMap::iterator CommandIter;
      CommandIter = m_CommandMap.begin ();
      int32 i = 0;
      while (i < DesiredIndex && CommandIter != m_CommandMap.end ())
      {
        CommandIter++;
        i++;
      }
      if (CommandIter != m_CommandMap.end ())
      {
        MessagePntr->PopSpecifier ();
        return CommandIter->second;
      }
      ErrorCode = B_ERROR;
      strcpy (TempString, "ScriptOCronApp::ResolveSpecifier shouldn't happen");
      goto ErrorExit;
    }
  }

  if (0 == Index && IsItOurScriptingCommand (g_AppScriptingPropertyList,
  PropertyNamePntr, SpecificationKind))
  {
    return this; /* It's one of our properties, self will handle message. */
  }

  /* Handle an unrecognized scripting command, let the parent figure it out. */

  return BApplication::ResolveSpecifier (
    MessagePntr, Index, SpecifierMsgPntr, SpecificationKind, PropertyNamePntr);

ErrorExit: /* Error message in TempString, return code in ErrorCode. */
  BMessage ReplyMessage (/* B_REPLY or maybe */ B_MESSAGE_NOT_UNDERSTOOD);
  ReplyMessage.AddInt32 (g_ErrorName, ErrorCode);
  ReplyMessage.AddString (g_MessageName, TempString);
  ErrorCode = MessagePntr->SendReply (&ReplyMessage,
    g_App /* Reply's reply handler */, 500000 /* send timeout */);
  return NULL; // Stop trying to resolve the message.
}



/******************************************************************************
 * Finally, the main program which drives it all.
 */

int main (int argc, char **argv)
{
  cout << endl << PrintUsage;

#if TEST_CRON_BITSET
  TestCronBitset ();
#endif

#if TEST_URL_CONVERTER
  TestURLConverter ();
#endif

#if INCLUDE_GUI_CODE
  { // Magic command line argument of "meta" turns on meta mode.
    int iArg;
    for (iArg = 1; iArg < argc; iArg++)
    {
      if (strcasecmp (argv[iArg], "meta") == 0)
        g_MetaMode = true;
    }
    cerr << "Meta-advanced editing is o" <<
      (g_MetaMode ? "n." : "ff.") << endl;
  }
#endif // INCLUDE_GUI_CODE

  {
    ScriptOCronApp MyApp;

    if (MyApp.InitCheck () == B_OK)
    {
      g_App = &MyApp;
      MyApp.LoadSaveSettings (true /* DoLoad */);
      MyApp.Run ();
    }
    cerr << PROGRAM_NAME_WORD " shutting down." << endl;
  } // Implicitly destroys MyApp at the end of this block.
  g_App = NULL;

  return 0;
}


/******************************************************************************
 * $Log: ScriptOCron.cpp,v $
 * Revision 1.133  2015/12/30 22:23:12  agmsmith
 * Some scripting operations usage wording improved.
 *
 * Revision 1.132  2015/12/26 17:14:09  agmsmith
 * Add the close on exec O_CLOEXEC flag to log files, don't want to leave
 * zillions of file handles open in child processes.
 *
 * Revision 1.131  2015/12/21 00:22:53  agmsmith
 * Move log files of deleted Commands to the system trash directory.
 *
 * Revision 1.130  2015/12/20 17:30:50  agmsmith
 * Bump version number.
 *
 * Revision 1.129  2015/12/19 22:22:30  agmsmith
 * Bump version number.
 *
 * Revision 1.128  2015/12/19 04:41:06  agmsmith
 * Change @Reboot to be @Launch, and the button label to be Run at Startup.
 *
 * Revision 1.127  2015/12/17 20:02:52  agmsmith
 * Added magic word ThisCommandName, which is what it says.
 *
 * Revision 1.126  2015/12/17 18:48:22  agmsmith
 * Use the canonical name of the Model in the Original Model saved data.
 *
 * Revision 1.125  2015/12/16 16:18:29  agmsmith
 * Downloads renamed, now "FTP Download" is the default for new ones in Basic
 * mode.  Now saves and loads the initial window size, so your last window size
 * and position is remembered.  Goes back to default size if you make the
 * window too small or move it off screen.
 *
 * Revision 1.124  2015/12/15 22:46:55  agmsmith
 * Bumped version.
 *
 * Revision 1.123  2015/12/13 20:13:56  agmsmith
 * Added Duplicate command and menu items, reworked new name generation too.
 *
 * Revision 1.122  2015/12/12 20:24:01  agmsmith
 * Added Meta mode and command line argument to trigger it.
 *
 * Revision 1.121  2015/12/11 19:09:53  agmsmith
 * Copyright notice shortened.
 *
 * Revision 1.120  2015/12/11 18:56:15  agmsmith
 * About box usage text vastly reduced for GUI version.
 *
 * Revision 1.119  2015/12/11 01:59:17  agmsmith
 * Bump version number to reflect changes in other files.
 *
 * Revision 1.118  2015/12/09 22:46:05  agmsmith
 * Export the CopyBMessageField function.
 *
 * Revision 1.117  2015/12/09 16:26:09  agmsmith
 * Remove welcome error message on first run (no settings file error).
 *
 * Revision 1.116  2015/12/09 03:34:45  agmsmith
 * Wording different for first time Fetchit vs AGMSScriptOCron.
 *
 * Revision 1.115  2015/12/08 21:44:01  agmsmith
 * By default, enable the log file on new Commands.  Make start in Windowed
 * mode the default, and rephrase the welcome text error message.
 *
 * Revision 1.114  2015/12/08 21:09:22  agmsmith
 * Back to launch with window showing as the initial default.
 *
 * Revision 1.113  2015/12/08 11:42:28  agmsmith
 * Don't confirm quit with user if it's a system shutdown kind of quit.
 *
 * Revision 1.112  2015/12/08 00:16:39  agmsmith
 * Resume running if the quit was cancelled.
 *
 * Revision 1.111  2015/12/07 14:58:38  agmsmith
 * Added some more UserData key descriptions.  Functions to access string type
 * UserData conveniently.  Add the OriginalModel UserData to new Commands
 * created from a Model, and remove it if the Template is edited by the user.
 *
 * Revision 1.110  2015/12/06 00:12:40  agmsmith
 * Fix QuitRequested behaviour so it delays replying to the B_QUIT_REQUESTED
 * message until a few updates have gone by (time to close threads and fetch
 * last log file entries).
 *
 * Revision 1.109  2015/12/04 22:33:15  agmsmith
 * Default to server mode on for new installs, as Dane requested.
 *
 * Revision 1.108  2015/12/04 22:25:10  agmsmith
 * Check for changes that affect settings, and save once a minute.
 *
 * Revision 1.107  2015/12/04 21:18:01  agmsmith
 * Added the ability to detect the first use of the program.
 *
 * Revision 1.106  2015/12/03 22:23:51  agmsmith
 * Log file GUI now working, enable/disable plus view log file button.
 *
 * Revision 1.105  2015/12/03 20:50:35  agmsmith
 * Also return the log path name when asked for log info.
 *
 * Revision 1.104  2015/12/03 19:14:34  agmsmith
 * Logging to a file feature added, no GUI yet.
 *
 * Revision 1.103  2015/12/02 00:03:05  agmsmith
 * Tune log receiving buffer and speed, Haiku uses bigger blocks too.
 *
 * Revision 1.102  2015/12/01 00:38:48  agmsmith
 * Move the proprietary GUI code into a separate file, change program name for
 * GUI mode to Fetchit!TM.
 *
 * Revision 1.101  2015/11/29 17:37:07  agmsmith
 * File requestor and local directory text input now working.
 *
 * Revision 1.100  2015/11/29 14:26:55  agmsmith
 * Work in progress adding FieldViewLocalDir, also rearranged base class
 * FieldViewGeneric and fixed some enabling bugs in FieldViewURL and some width
 * spacing everywhere.
 *
 * Revision 1.99  2015/11/28 17:27:13  agmsmith
 * Wording changes from Dane.
 *
 * Revision 1.98  2015/11/28 03:10:53  agmsmith
 * Trigger time classes reorganised under a new base class.  Now auto-edit-mode
 * works for the full trigger time editor GUI.  Added Disabled trigger time
 * checkbox.  Template text now has auto-edit.
 *
 * Revision 1.97  2015/11/27 02:35:41  agmsmith
 * Added basic mode time display and editing.  Still need to backport the
 * disabled feature and auto-edit to the full time edit view.
 *
 * Revision 1.96  2015/11/26 01:11:21  agmsmith
 * Process user input from the URL field GUI.  Now fully working.
 *
 * Revision 1.95  2015/11/25 22:49:58  agmsmith
 * URL parsing now displayed, but the reverse direction (editing URL) isn't
 * implemented yet.  Auto-save in basic mode when Enter is typed.  Change short
 * instruction text to reflect current situation.
 *
 * Revision 1.94  2015/11/25 00:36:04  agmsmith
 * Working on the URL display, added URLConverter class and test code.
 *
 * Revision 1.93  2015/11/23 23:04:12  agmsmith
 * Work in progress - moved URL form into a FieldViewURL class, no longer built
 * into the Basic view (so can be used in Advanced too).
 *
 * Revision 1.92  2015/11/23 21:53:19  agmsmith
 * Renaming FieldView classes, getting ready to work on FieldViewURL.
 *
 * Revision 1.91  2015/11/23 20:49:40  agmsmith
 * Make a common parent BView class for the advanced and basic Command details
 * views, so they have a common virtual polling function and can then be used
 * interchangably.  Turns out we didn't need it, can force an extra poll by
 * using the SelectionChanged flag when switching between advanced and basic.
 *
 * Revision 1.90  2015/11/22 22:42:13  agmsmith
 * Split FieldGenericView into a base abstract class and one for text Fields.
 *
 * Revision 1.89  2015/11/22 02:40:05  agmsmith
 * Added the UserData facility, now we can store stuff in Commands and Fields
 * without having to add more scripting code.
 *
 * Revision 1.88  2015/11/20 20:31:47  agmsmith
 * Fix bug with detecting if a name is used by a Model.
 *
 * Revision 1.87  2015/11/20 20:03:22  agmsmith
 * Added "What" elements for picking a file name.  Also don't start Commands
 * when shutting down.
 *
 * Revision 1.86  2015/11/20 18:27:41  agmsmith
 * Removed g_ServerMode from non-GUI build and from scripting system (though
 * still saved in settings), renamed non-server as Windowed mode.  Added
 * g_AdvancedMode and a menu to switch between basic and advanced mode Command
 * details user interfaces.  Also New Command button now only creates file
 * download Commands in basic mode (no pop-up menu).
 *
 * Revision 1.85  2015/11/19 21:51:00  agmsmith
 * Got an initial Basic command details view working, just with name and a
 * radio button.
 *
 * Revision 1.84  2015/11/18 22:57:06  agmsmith
 * Now working with advanced stuff in a separate sub-view.
 *
 * Revision 1.83  2015/11/17 22:51:45  agmsmith
 * Splitting out code into AdvancedCommandDetailsView so that we can have both
 * basic and advanced modes.  Work in progress - have to hook up messaging
 * again.
 *
 * Revision 1.82  2015/11/17 19:02:25  agmsmith
 * x-vnd rather than vnd, for app signature.
 *
 * Revision 1.81  2015/11/17 19:00:48  agmsmith
 * Move some GUI variables to their own section, make signature
 * usage more consistently lower case for scripting sigs etc.
 *
 * Revision 1.80  2015/11/16 22:42:22  agmsmith
 * Added the company logo badge graphic, which pulls up the docs too.
 *
 * Revision 1.79  2015/11/12 19:57:22  agmsmith
 * Added a menu bar and moved the About and Server Mode gadgets there.
 *
 * Revision 1.78  2015/11/11 23:07:47  agmsmith
 * Fix a bug in loading models, name pointer changes when manipulating a
 * BMessage.
 *
 * Revision 1.77  2015/11/05 23:02:35  agmsmith
 * Now loads built-in Models from a resource, and only saves user created
 * Models.
 *
 * Revision 1.76  2015/11/05 21:01:49  agmsmith
 * Use Model name when making new Commands in the GUI, also add a sequence
 * number at the end so you can add a lot of Commands in a batch.
 *
 * Revision 1.75  2015/11/04 22:58:50  agmsmith
 * Add a text box so you can edit the Template of a Command.
 *
 * Revision 1.74  2015/11/04 21:11:10  agmsmith
 * Create new Commands in Edit state, so they don't start running immediately.
 * Also only allow delete of a Command in edit state.  Punctuation of error
 * messages removed from the end (the display adds a comma and the error
 * codes).
 *
 * Revision 1.73  2015/11/03 22:38:48  agmsmith
 * Show a pop-up menu listing available Models when New button used.
 *
 * Revision 1.72  2015/11/03 20:54:51  agmsmith
 * Added buttons for creating and deleting Models.
 *
 * Revision 1.71  2015/11/01 23:03:48  agmsmith
 * Finished Model functionality, can now create, list, delete and use Models.
 *
 * Revision 1.70  2015/10/31 21:42:24  agmsmith
 * Now loading and saving the Models.
 *
 * Revision 1.69  2015/10/31 19:11:46  agmsmith
 * Fix up a few small things, start on adding Models.
 *
 * Revision 1.68  2015/10/30 21:34:26  agmsmith
 * Move the CommandState into a global space enum, also fix space for text
 * labels to be a bit bigger so it shows up unclipped in Haiku.
 *
 * Revision 1.67  2015/10/29 17:01:16  agmsmith
 * Added the radio buttons for setting Trigger type.
 *
 * Revision 1.66  2015/10/28 19:14:58  agmsmith
 * Added the trigger string text box and connected up the checkboxes so that
 * they change the trigger time too.
 *
 * Revision 1.65  2015/10/27 21:56:47  agmsmith
 * Now updates the time checkboxes with data from the Command's trigger string.
 *
 * Revision 1.64  2015/10/26 17:00:35  agmsmith
 * Added a scroll bar to the Start Times tab, also moved the scrolling stuff to
 * an internal class so it can be reused for subclasses.
 *
 * Revision 1.63  2015/10/25 22:49:48  agmsmith
 * Add the All or None checkbox after the other cron time ones.
 *
 * Revision 1.62  2015/10/25 21:42:19  agmsmith
 * Reverse the order of crontab fields to be more intuitive, with minutes last.
 *
 * Revision 1.61  2015/10/24 21:09:14  agmsmith
 * Working on TriggerTimeView to let you edit the cron string, currently shows
 * the checkboxes but they aren't hooked up to data yet.
 *
 * Revision 1.60  2015/10/21 23:23:29  agmsmith
 * Fixed tabbed view resizing so that when switching back to a tab, it gets
 * resized to the current size of the parent view, compensating for missed
 * resize messages.  Also move creation of child GUI elements out of
 * AttachedToWindow so you don't get duplicates every time a tab is selected
 * (tabs work by detaching the unused tabs, and then attaching the visible tab
 * to the main view).
 *
 * Revision 1.59.1.1  2015/10/21 19:53:36  agmsmith
 * Tried fixing missing size updates by using a custom BScrollView that resizes
 * whenever it is reattached to window.  Unfortunately, parent view is an
 * internal one in the BTab that is the wrong size.
 *
 * Revision 1.59  2015/10/21 18:51:16  agmsmith
 * Move Command Line display to the miscellaneous tab.
 *
 * Revision 1.58  2015/10/19 19:57:31  agmsmith
 * Moved various GUI elements into a BTabView, with tabs for Fields, Log and
 * Miscellaneous.  Mostly working, but with some redraw problems in Fields.
 *
 * Revision 1.57  2015/10/15 18:52:13  agmsmith
 * List of Field data entry fields is now scrollable, with a scroll bar that
 * responds to the size of the window, etc.  Also have the log and Field
 * display splitting the available space in half, dynamically updated as the
 * window is resized.
 *
 * Revision 1.56  2015/10/05 21:44:31  agmsmith
 * Started moving the display of Fields into its own BView so that we can later
 * do scrolling of the whole thing.
 *
 * Revision 1.55  2015/10/04 20:13:10  agmsmith
 * Now with working Cron type command triggers, and @ReBoot works too.
 *
 * Revision 1.54  2015/10/02 03:03:22  agmsmith
 * Work in progress on trigger functionality.  Works for named Command, still
 * have to do Cron and reboot.
 *
 * Revision 1.53  2015/10/01 23:25:45  agmsmith
 * Added Trigger string to the Command scripting, doesn't yet do anything.
 *
 * Revision 1.52  2015/09/28 22:05:55  agmsmith
 * Work in progress on CronBitset.
 *
 * Revision 1.51  2015/08/24 23:16:22  agmsmith
 * Now keeps log text separately for each Command in the GUI, so selecting
 * commands will change the log text to the accumulation for that Command.
 *
 * Revision 1.50  2015/08/24 21:09:36  agmsmith
 * Moved change log to the end of the file, was getting too long.
 *
 * Revision 1.49  2015/08/21 22:35:06  agmsmith
 * Added a text box to display the log output.  Not quite right, need to
 * preserve logs for each Command separately, and show the appropriate one when
 * selecting different Commands.
 *
 * Revision 1.48  2015/08/21 18:05:31  agmsmith
 * Added polling to fetch standard output and store it in a log string, added a
 * scripting operation to read the log string, figured out how to kill the
 * child shell and its children using process groups.
 *
 * Revision 1.47  2015/08/20 19:22:32  agmsmith
 * Can now run a command, using fork() and execve().  Can't see the output yet.
 *
 * Revision 1.46  2015/08/13 18:25:39  agmsmith
 * Now does keyword substitution to generate the command line.
 *
 * Revision 1.45  2015/08/11 21:54:46  agmsmith
 * Added scripting to get the current command line, GUI to display it, just
 * need some code to calculate it.
 *
 * Revision 1.44  2015/08/10 21:47:51  agmsmith
 * Added the Template string to Commands, including scripting and archiving.
 *
 * Revision 1.43  2015/08/10 16:06:56  agmsmith
 * Too many buttons, make two rows.  Also Save button now forces text updates
 * for text boxes so that the last bit of editing by the user gets transfered
 * too.
 *
 * Revision 1.42  2015/08/10 15:28:59  agmsmith
 * Added a Stop button, scripting for Run command, enable and disable buttons
 * depending on state.
 *
 * Revision 1.41  2015/08/10 00:08:21  agmsmith
 * Also select the new Command when using the New button.
 *
 * Revision 1.40  2015/08/09 23:48:12  agmsmith
 * Switch to using messages to change the selected Command after a rename,
 * finally figured out how to send a message to self with a BInvoker
 * (PostMessage doesn't work for some reason).
 *
 * Revision 1.39  2015/08/09 21:53:38  agmsmith
 * Added Delete button, reselection after name changed (using global vars).
 *
 * Revision 1.38  2015/07/28 21:50:47  agmsmith
 * Can now edit the name of a Command in the GUI.
 *
 * Revision 1.37  2015/07/28 19:25:12  agmsmith
 * Field BViews now are inside the Command Details box.
 *
 * Revision 1.36  2015/07/28 15:27:34  agmsmith
 * Trying to stuff the scrolling list view inside a BBox, but the coordinates
 * of the box's interior space aren't quite right.
 *
 * Revision 1.35  2015/07/24 19:50:32  agmsmith
 * Can now see and edit the Field values.
 *
 * Revision 1.34  2015/07/24 18:56:06  agmsmith
 * Consistently format class parts, using // comments after public/private.
 *
 * Revision 1.33  2015/07/22 21:27:04  agmsmith
 * Added buttons for Edit, Save, Cancel (all now working) and an unimplemented
 * Run button.
 *
 * Revision 1.32  2015/07/15 22:26:58  agmsmith
 * Can now set and get Field's Value, also stored in settings.
 *
 * Revision 1.31  2015/07/15 15:07:22  agmsmith
 * Switch to BTextControl for displaying field values.
 *
 * Revision 1.30  2015/07/08 01:41:23  agmsmith
 * Field value as a BTextView.  Not too good, should try BTextControl instead.
 *
 * Revision 1.29  2015/07/05 00:24:50  agmsmith
 * Made fonts global, rather than having a copy for each displayed thing.
 * Display the Field names (temporary code using DrawString) and update them as
 * the fields change.
 *
 * Revision 1.28  2015/06/15 21:59:01  agmsmith
 * Adding FieldGenericView, also put implementation of classes in alphabetical
 * order, so Command stuff comes before Field stuff.
 *
 * Revision 1.27  2015/06/12 19:39:16  agmsmith
 * Can now create, delete, rename, count Field objects.
 *
 * Revision 1.26  2015/06/11 17:19:09  agmsmith
 * In the middle of adding Field objects.
 *
 * Revision 1.25  2015/06/10 01:08:54  agmsmith
 * Finished the Edit property implementation, changed the list
 * items to update their state if the Command changed.
 *
 * Revision 1.24  2015/06/09 20:34:08  agmsmith
 * Working on Edit property of Commands.
 *
 * Revision 1.23  2015/05/26 02:04:34  agmsmith
 * List items now have their own large font, and the progress bar
 * is now being drawn.
 *
 * Revision 1.22  2015/05/25 22:00:03  agmsmith
 * More colourful list items, with proper antialiasing of text too.
 *
 * Revision 1.21  2015/05/22 21:57:08  agmsmith
 * Switch to using a custom ListItem to draw our Commands.
 *
 * Revision 1.20  2015/05/22 16:19:12  agmsmith
 * Force use of Specifier[0] when examining scripting messages, and other
 * changes to get Scripting also working in Haiku.
 *
 * Revision 1.19  2015/05/21 21:08:30  agmsmith
 * Now saves and loads Commands, also use modification detection system to
 * trigger saving of the settings file.
 *
 * Revision 1.18  2015/05/21 19:23:37  agmsmith
 * Use the change detection mechanism to avoid some updates.
 *
 * Revision 1.17  2015/05/21 18:39:37  agmsmith
 * Window list of commands now tracks the server's list of commands.
 *
 * Revision 1.16  2015/05/18 20:49:12  agmsmith
 * Work in progress in doing scripting and displaying commands.
 *
 * Revision 1.15  2015/05/15 21:57:26  agmsmith
 * Now with a scrolling list box that will later contain task names.
 *
 * Revision 1.14  2015/05/14 03:01:52  agmsmith
 * Added change serial numbers, and the ability to delete a Command.
 *
 * Revision 1.13  2015/05/13 00:18:34  agmsmith
 * Can now change the name of a Command, and read it, and the scripting now
 * gets down to the Command object using the specifier system.
 *
 * Revision 1.12  2015/05/12 22:52:54  agmsmith
 * Shuffling order of classes and implementations, more alphabetical now.
 *
 * Revision 1.11  2015/05/12 21:59:44  agmsmith
 * Working on adding a Command operation, also removed useless busy cursor and
 * added some locking utility methods.
 *
 * Revision 1.10  2015/05/12 15:50:17  agmsmith
 * Add scripting for an edit mode for Commands, with the ability to cancel an
 * edit and restore the Command to the original state.  Also usage wording
 * updated.
 *
 * Revision 1.9  2015/05/12 02:15:10  agmsmith
 * Better error messages for resolving specifiers.
 *
 * Revision 1.8  2015/05/11 21:45:45  agmsmith
 * Can now create Commands using scripting.
 *
 * Revision 1.7  2015/05/08 21:52:48  agmsmith
 * Resolving scripting specifiers still in progress, but closer now!
 *
 * Revision 1.6  2015/05/07 21:58:48  agmsmith
 * Work in progress on scripting.
 *
 * Revision 1.5  2015/05/07 18:48:03  agmsmith
 * Remove CommanderLooper since we no longer take command line arguments, also
 * update the scripting documentation and usage.
 *
 * Revision 1.4  2015/05/07 15:53:48  agmsmith
 * Mostly cosmetics - program name as a macro.
 *
 * Revision 1.3  2015/05/01 21:58:43  agmsmith
 * Use the same About text in the printed usage and in the about box, which
 * means being able to turn off word wrap, and capturing the usage output to an
 * in-memory string.
 *
 * Revision 1.2  2015/05/01 21:04:39  agmsmith
 * Got ServerMode set, get, save, load, display working.
 *
 * Revision 1.1  2015/05/01 19:46:46  agmsmith
 * Initial revision.  Main program for Script-o-Cron, initially a stripped down
 * copy of AGMSBaysianSpamServer.
 */
