/******************************************************************************
 * $Header: /CommonBe/agmsmith/Programming/ScriptOCron/RCS/ScriptOCron.h,v 1.12 2015/12/19 04:41:06 agmsmith Exp $
 *
 * This file contains stuff that's common to both the command line version and
 * the graphical user interface version of AGMSScriptOCron.
 *
 * This is a BeOS / Haiku OS program for editing and running commands without
 * needing to know about using the command line.  There's a graphical user
 * interface for editing the commands, letting you specify each argument
 * separately.  Running of commands can be triggered by "cron" style time based
 * events, the BeOS scripting system (see the Hey command), as well as a GUI
 * start button.
 *
 * Note that this uses the AGMS coding style.  That means no tabs, indents are
 * two spaces, m_ is the prefix for member variables, g_ is the prefix for
 * global names, C style comments, constants are in all capital letters and
 * most other things are mixed case, it's word wrapped to fit in 79 characters
 * per line to make proofreading on paper easier, and functions are listed in
 * reverse dependency order so that forward declarations (function prototypes
 * with no code) aren't needed.
 *
 * The command line only (no graphical user interface) AGMSScriptOCron software
 * is licensed under the MIT license.  The commercial Fetchit! software (the
 * graphical user interface layer on top of AGMSSCriptOCron) is sold under the
 * TuneTracker Systems Sales Terms and Conditions (summary: licensed to use a
 * single copy on a single computer, no reselling, no source code).  The source
 * code for the commercial version is proprietary, owned by TuneTracker Systems
 * LLC under a royalty agreement with Alexander G. M. Smith.
 *
 * The contents of this source code file are MIT licensed.
 *
 * Copyright (c) 2015 by Alexander G. M. Smith.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#ifndef INCLUDE_GUI_CODE
  #define INCLUDE_GUI_CODE 0
#endif
  // Controls building with or without the graphical user interface.  The
  // commercial version has the GUI and the public version does not (you have
  // to use "hey" commands to achieve similar functionality).  The source code
  // included with the public version also has the GUI code removed).


#if INCLUDE_GUI_CODE
  #define PROGRAM_NAME_WORD "Fetchit"
    // Whenever we need to concatenate the program name into some text, this
    // string is used.  Thus it needs to be a macro.  And not have spaces or
    // weird characters since it gets appended to various file and other names.
  #define PROGRAM_NAME_FANCY "Fetchit!" B_UTF8_TRADEMARK
    // The program name as displayed to the user, with odd characters like the
    // trademark symbol.
#else // Building command line only version.
  #define PROGRAM_NAME_WORD "AGMSScriptOCron"
  #define PROGRAM_NAME_FANCY "AGMSScriptOCron"
#endif

#define SUITE_BASE "suite/vnd.agmsmith.scriptocron."
  // The various command suites all start with this phrase.  It's followed by a
  // name of a type of scriptable object, like "command" or "field" or
  // "application".  Needs to be the same for both commercial and public
  // versions of the program, so third party scripting works without changes.


/******************************************************************************
 * Global variables, and not-so-variable things too.  Grouped by functionality.
 */
extern const char *g_DataName; // "data" stored in a BMessage.
extern const char *g_ErrorName; // "error" code stored in a BMessage.
extern const char *g_MessageName; // "message" describes error in a BMessage.
extern const char *g_ModelName; // "model" scripting field for Model's name.
extern const char *g_NameName; // "name" BMessage field with new name.
extern const char *g_NameReboot; // "@Launch" for the trigger string.
extern const char *g_NameUnnamed; // "Unnamed" default name for new things.
extern const char *g_PathName; // "pathname" element returned in a BMessage.
extern const char *g_ResultName; // "result" returned in a BMessage.

extern bool g_FirstTimeAppRun;
  /* TRUE if this is the very first time this program has been run (detected by
  the lack of a settings file).  FALSE for later times.  Useful for doing very
  first time initialisation, like setting up auto-run on system boot. */

extern bool g_SettingsNeedSaving;
  /* Set to TRUE by anything that changes something which may affect the
  settings file.  Once a minute the main application will check this flag and
  save the settings if needed, clearing the flag after it's done. */


/* These ones are only used in the GUI version of the program, but are common
so that they can be saved in the settings file. */

#if INCLUDE_GUI_CODE
extern volatile bool g_AdvancedMode;
  /* When TRUE the GUI is running in advanced mode, otherwise basic mode. */

extern BRect g_InitialWindowFrame;
  /* A saved setting that specifies the initial window size and position. */

extern bool g_MetaMode;
  /* When TRUE the GUI has some of the meta things shown in advanced mode,
  currently just the Template text entry and expanded command line. */

extern volatile bool g_ServerMode;
  /* When TRUE the program runs in server mode - error messages don't result in
  pop-up dialog boxes, but you can still see them in stderr.  Also the window
  is minimized, if it exists.  Don't need to lock the Application or go through
  scripting since this is simple enough to be read directly from multiple
  threads. */
#endif


typedef enum TriggerTypeEnum {
  TRG_CRON = 0,
  TRG_NAMED_COMMAND,
  TRG_REBOOT,
  TRG_MANUAL,
  TRG_MAX} TriggerType;

typedef enum CommandStateEnum {
  CS_READY_TO_RUN = 0,
  CS_RUNNING,
  CS_EDITING,
  CS_MAX} CommandState;

/* The various properties you can use in the scripting system for all the kinds
of objects we support. */

typedef enum PropertyNumbersEnum
{
  PN_COMMAND = 0,
  PN_EDIT,
  PN_RUN,
  PN_SERIAL,
  PN_TEMPLATE,
  PN_COMMAND_LINE,
  PN_TRIGGER,
  PN_LOG,
  PN_MODEL,
  PN_FIELD,
  PN_NAME,
  PN_VALUE,
  PN_USER_DATA,
  PN_LOG_FILE_ENABLE,
  PN_DUPLICATE,
  PN_MAX
} PropertyNumbers;

extern char * g_PropertyNames [PN_MAX];
  // Strings for the property names, useful for building scripting BMessages.
  // Or for parsing them.


/******************************************************************************
 * Scriptable objects like Commands or Fields are stored in a Map, with the key
 * being the case insenitive name of the object (actually the pointer obtained
 * from BHandler::Name) and the value being a pointer to the BHandler based
 * object.
 *
 * typedef map<const char *, BHandler *, LessThanCharStringStruct> ObjectMap;
 * Though to be more type safe, we'll define maps for specific object types
 * separately, as needed.  Maybe using the object itself rather than a pointer.
 */

struct LessThanCharStringStruct
{
  bool operator() ( /* Our less-than comparison function for sorting. */
    const char * const &ItemA,
    const char * const &ItemB) const
  {
    int Delta = strcasecmp (ItemA, ItemB);
    return Delta < 0;
  };
};


/******************************************************************************
 * This utility function copies a named field from the source BMessage to the
 * destination BMessage, optionally renaming it in the process.  If there are
 * multiple fields with that name (an array), they're all copied.  Note that
 * the two messages must be different, copying from a message to itself won't
 * work properly.
 */

extern status_t CopyBMessageField (
  BMessage *SourceMessagePntr, BMessage *DestMessagePntr,
  const char *FieldName, const char *NewName = NULL);


/******************************************************************************
 * Global utility function to display an error message and return.  The message
 * part describes the error, and if ErrorNumber is non-zero, gets the string
 * ", error code $X (standard description)." appended to it.  If the message
 * is NULL then it gets defaulted to "Something went wrong".  The title part
 * doesn't get displayed (no title bar in the dialog box, but you can see it in
 * the debugger as the window thread name), and defaults to "Error Message" if
 * you didn't specify one.  If running in command line mode, the error gets
 * printed to stderr rather than showing up in a dialog box.
 */

extern void DisplayErrorMessage (
  const char *MessageString = NULL,
  int ErrorNumber = 0,
  const char *TitleString = NULL);


/******************************************************************************
 * Print the usage info to the stream.  Optionally includes a list of all
 * commands and do word wrapping (zero to not wrap).
 */

extern void PrintUsage (ostream& OutputStream, int WrapWidth,
  bool ListCommands);


#if INCLUDE_GUI_CODE
/******************************************************************************
 * Start up the GUI user interface.  Called by ScriptOCronApp::ReadyToRun.
 */

extern void SetSizesAndOpenWindow ();
#endif // INCLUDE_GUI_CODE


/******************************************************************************
 * $Log: ScriptOCron.h,v $
 * Revision 1.12  2015/12/19 04:41:06  agmsmith
 * Change @Reboot to be @Launch, and the button label to be Run at Startup.
 *
 * Revision 1.11  2015/12/16 16:18:29  agmsmith
 * Downloads renamed, now "FTP Download" is the default for new ones in Basic
 * mode.  Now saves and loads the initial window size, so your last window size
 * and position is remembered.  Goes back to default size if you make the
 * window too small or move it off screen.
 *
 * Revision 1.10  2015/12/13 20:13:56  agmsmith
 * Added Duplicate command and menu items, reworked new name generation too.
 *
 * Revision 1.9  2015/12/12 20:24:01  agmsmith
 * Added Meta mode and command line argument to trigger it.
 *
 * Revision 1.8  2015/12/09 22:46:05  agmsmith
 * Export the CopyBMessageField function.
 *
 * Revision 1.7  2015/12/04 22:25:10  agmsmith
 * Check for changes that affect settings, and save once a minute.
 *
 * Revision 1.6  2015/12/04 21:17:44  agmsmith
 * Added the ability to detect the first use of the program.
 *
 * Revision 1.5  2015/12/03 22:23:51  agmsmith
 * Log file GUI now working, enable/disable plus view log file button.
 *
 * Revision 1.4  2015/12/03 19:14:34  agmsmith
 * Logging to a file feature added, no GUI yet.
 *
 * Revision 1.3  2015/12/02 00:02:47  agmsmith
 * Wording.
 *
 * Revision 1.2  2015/12/01 00:38:48  agmsmith
 * Move the proprietary GUI code into a separate file, change program name for
 * GUI mode to Fetchit!TM.
 *
 * Revision 1.1  2015/11/30 13:27:09  agmsmith
 * Initial revision.  Common things shared by GUI and command line builds.
 */
